/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LAPACK
#define COMMON_LAPACK

#ifndef ASSEMBLER

/* Lapack Library */

int sgetf2_k(long, long, float  *a, long, blasint *, long, float *);
int dgetf2_k(long, long, double *a, long, blasint *, long, double *);
int cgetf2_k(long, long, float  *a, long, blasint *, long, float *);
int zgetf2_k(long, long, double *a, long, blasint *, long, double *);

int sgetrf_single(long, long, float  *a, long, blasint *, long, float *);
int dgetrf_single(long, long, double *a, long, blasint *, long, double *);
int cgetrf_single(long, long, float  *a, long, blasint *, long, float *);
int zgetrf_single(long, long, double *a, long, blasint *, long, double *);

int sgetrf_parallel(long, long, float  *a, long, blasint *, long, float *);
int dgetrf_parallel(long, long, double *a, long, blasint *, long, double *);
int cgetrf_parallel(long, long, float  *a, long, blasint *, long, float *);
int zgetrf_parallel(long, long, double *a, long, blasint *, long, double *);

int slaswp_plus(long, long, long, float,  float  *, long, float  *, long, blasint *, long);
int dlaswp_plus(long, long, long, double, double *, long, double *, long, blasint *, long);
int slaswp_minus(long, long, long, float,  float  *, long, float  *, long, blasint *, long);
int dlaswp_minus(long, long, long, double, double *, long, double *, long, blasint *, long);
int claswp_plus(long, long, long, float, float, float *, long, float *, long, blasint *, long);
int zlaswp_plus(long, long, long, double, double, double *, long, double *, long, blasint *, long);
int claswp_minus(long, long, long, float, float, float *, long, float *, long, blasint *, long);
int zlaswp_minus(long, long, long, double, double, double *, long, double *, long, blasint *, long);

int sgetrs_N(long, long, float  *, long, float  *, long, blasint *, float  *);
int sgetrs_T(long, long, float  *, long, float  *, long, blasint *, float  *);

int dgetrs_N(long, long, double *, long, double *, long, blasint *, double *);
int dgetrs_T(long, long, double *, long, double *, long, blasint *, double *);

int cgetrs_N(long, long, float  *, long, float  *, long, blasint *, float  *);
int cgetrs_T(long, long, float  *, long, float  *, long, blasint *, float  *);
int cgetrs_R(long, long, float  *, long, float  *, long, blasint *, float  *);
int cgetrs_C(long, long, float  *, long, float  *, long, blasint *, float  *);

int zgetrs_N(long, long, double *, long, double *, long, blasint *, double *);
int zgetrs_T(long, long, double *, long, double *, long, blasint *, double *);
int zgetrs_R(long, long, double *, long, double *, long, blasint *, double *);
int zgetrs_C(long, long, double *, long, double *, long, blasint *, double *);

int spotf2_U(long, float  *a, long, float *);
int spotf2_L(long, float  *a, long, float *);
int dpotf2_U(long, double *a, long, double *);
int dpotf2_L(long, double *a, long, double *);
int cpotf2_U(long, float  *a, long, float *);
int cpotf2_L(long, float  *a, long, float *);
int zpotf2_U(long, double *a, long, double *);
int zpotf2_L(long, double *a, long, double *);

int spotrf_U(long, float  *a, long, float *);
int spotrf_L(long, float  *a, long, float *);
int dpotrf_U(long, double *a, long, double *);
int dpotrf_L(long, double *a, long, double *);
int cpotrf_U(long, float  *a, long, float *);
int cpotrf_L(long, float  *a, long, float *);
int zpotrf_U(long, double *a, long, double *);
int zpotrf_L(long, double *a, long, double *);

#define SGETF2	sgetf2_k
#define DGETF2	dgetf2_k
#define CGETF2	cgetf2_k
#define ZGETF2	zgetf2_k

#ifndef SMP
#define SGETRF	sgetrf_single
#define DGETRF	dgetrf_single
#define CGETRF	cgetrf_single
#define ZGETRF	zgetrf_single
#else
#define SGETRF	sgetrf_parallel
#define DGETRF	dgetrf_parallel
#define CGETRF	cgetrf_parallel
#define ZGETRF	zgetrf_parallel
#endif

#define SGETRS_N	sgetrs_N
#define SGETRS_T	sgetrs_T
#define DGETRS_N	dgetrs_N
#define DGETRS_T	dgetrs_T
#define CGETRS_N	cgetrs_N
#define CGETRS_T	cgetrs_T
#define ZGETRS_N	zgetrs_N
#define ZGETRS_T	zgetrs_T

#define SLASWP_PLUS	slaswp_plus
#define SLASWP_MINUS	slaswp_minus
#define DLASWP_PLUS	dlaswp_plus
#define DLASWP_MINUS	dlaswp_minus
#define CLASWP_PLUS	claswp_plus
#define CLASWP_MINUS	claswp_minus
#define ZLASWP_PLUS	zlaswp_plus
#define ZLASWP_MINUS	zlaswp_minus

#ifndef COMPLEX
#ifndef DOUBLE
#define GETF2	SGETF2
#define GETRF	SGETRF
#define LASWP_PLUS	SLASWP_PLUS
#define LASWP_MINUS	SLASWP_MINUS
#define GETRS_N		SGETRS_N
#define GETRS_T		SGETRS_T
#define GETRF_SINGLE	sgetrf_single
#else
#define GETF2	DGETF2
#define GETRF	DGETRF
#define LASWP_PLUS	DLASWP_PLUS
#define LASWP_MINUS	DLASWP_MINUS
#define GETRS_N		DGETRS_N
#define GETRS_T		DGETRS_T
#define GETRF_SINGLE	dgetrf_single
#endif
#else
#ifndef DOUBLE
#define GETF2	CGETF2
#define GETRF	CGETRF
#define LASWP_PLUS	CLASWP_PLUS
#define LASWP_MINUS	CLASWP_MINUS
#define GETRS_N		CGETRS_N
#define GETRS_T		CGETRS_T
#define GETRF_SINGLE	cgetrf_single
#else
#define GETF2	ZGETF2
#define GETRF	ZGETRF
#define LASWP_PLUS	ZLASWP_PLUS
#define LASWP_MINUS	ZLASWP_MINUS
#define GETRS_N		ZGETRS_N
#define GETRS_T		ZGETRS_T
#define GETRF_SINGLE	zgetrf_single
#endif
#endif

#ifdef DOUBLE
#define  POTF2_U  dpotf2_U
#define  POTF2_L  dpotf2_L
#define ZPOTF2_U  zpotf2_U
#define ZPOTF2_L  zpotf2_L
#else
#define  POTF2_U  spotf2_U
#define  POTF2_L  spotf2_L
#define ZPOTF2_U  cpotf2_U
#define ZPOTF2_L  cpotf2_L
#endif

#endif
#endif
