/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LEVEL2
#define COMMON_LEVEL2

#ifndef ASSEMBLER

/* Level 2 Blas routines */

int sger_k (long, long, long, float,  float  *, long, float  *, long, float  *, long, float *);
int dger_k (long, long, long, double, double *, long, double *, long, double *, long, double *);

int cgeru_k(long, long, long,
	      float, float,  float  *, long, float  *, long, float  *, long, float *);
int cgerc_k(long, long, long,
	      float, float,  float  *, long, float  *, long, float  *, long, float *);
int zgeru_k(long, long, long,
	      double, double,  double *, long, double *, long, double *, long, double *);
int zgerc_k(long, long, long,
	      double, double,  double *, long, double *, long, double *, long, double *);

#if 1

void sgemv_n(long, long, long, float, 
		 float *, long, float *, long, float *, long, void *buffer);
void sgemv_t(long, long, long, float,
		 float *, long, float *, long, float *, long, void *buffer);

void dgemv_n(long, long, long, double,
		 double *, long, double *, long, double *, long, void *buffer);
void dgemv_t(long, long, long, double, 
		 double *, long, double *, long, double *, long, void *buffer);

void cgemv_n(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_t(long, long, long, float, float, 
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_r(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_c(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_o(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_u(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_s(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);
void cgemv_d(long, long, long, float, float,
		 float *, long, float *, long, float *, long, void *buffer);

void zgemv_n(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_t(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_r(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_c(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_o(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_u(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_s(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);
void zgemv_d(long, long, long, double, double,
		 double *, long, double *, long, double *, long, void *buffer);

#else

void sgemv_n(long, long, long, float, 
	     float *, long, float *, long, float *, long, void *buffer, long);
void sgemv_t(long, long, long, float,
	     float *, long, float *, long, float *, long, void *buffer, long);

void dgemv_n(long, long, long, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void dgemv_t(long, long, long, double, 
	     double *, long, double *, long, double *, long, void *buffer, long);

void cgemv_n(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_t(long, long, long, float, float, 
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_r(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_c(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_o(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_u(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_s(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);
void cgemv_d(long, long, long, float, float,
	     float *, long, float *, long, float *, long, void *buffer, long);

void zgemv_n(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_t(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_r(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_c(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_o(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_u(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_s(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);
void zgemv_d(long, long, long, double, double,
	     double *, long, double *, long, double *, long, void *buffer, long);

#endif


int strsv_NUU(long, float *, long, float *, long, void *);
int strsv_NUN(long, float *, long, float *, long, void *);
int strsv_NLU(long, float *, long, float *, long, void *);
int strsv_NLN(long, float *, long, float *, long, void *);
int strsv_TUU(long, float *, long, float *, long, void *);
int strsv_TUN(long, float *, long, float *, long, void *);
int strsv_TLU(long, float *, long, float *, long, void *);
int strsv_TLN(long, float *, long, float *, long, void *);

int dtrsv_NUU(long, double *, long, double *, long, void *);
int dtrsv_NUN(long, double *, long, double *, long, void *);
int dtrsv_NLU(long, double *, long, double *, long, void *);
int dtrsv_NLN(long, double *, long, double *, long, void *);
int dtrsv_TUU(long, double *, long, double *, long, void *);
int dtrsv_TUN(long, double *, long, double *, long, void *);
int dtrsv_TLU(long, double *, long, double *, long, void *);
int dtrsv_TLN(long, double *, long, double *, long, void *);

int ctrsv_NUU(long, float *, long, float *, long, void *);
int ctrsv_NUN(long, float *, long, float *, long, void *);
int ctrsv_NLU(long, float *, long, float *, long, void *);
int ctrsv_NLN(long, float *, long, float *, long, void *);
int ctrsv_TUU(long, float *, long, float *, long, void *);
int ctrsv_TUN(long, float *, long, float *, long, void *);
int ctrsv_TLU(long, float *, long, float *, long, void *);
int ctrsv_TLN(long, float *, long, float *, long, void *);
int ctrsv_RUU(long, float *, long, float *, long, void *);
int ctrsv_RUN(long, float *, long, float *, long, void *);
int ctrsv_RLU(long, float *, long, float *, long, void *);
int ctrsv_RLN(long, float *, long, float *, long, void *);
int ctrsv_CUU(long, float *, long, float *, long, void *);
int ctrsv_CUN(long, float *, long, float *, long, void *);
int ctrsv_CLU(long, float *, long, float *, long, void *);
int ctrsv_CLN(long, float *, long, float *, long, void *);

int ztrsv_NUU(long, double *, long, double *, long, void *);
int ztrsv_NUN(long, double *, long, double *, long, void *);
int ztrsv_NLU(long, double *, long, double *, long, void *);
int ztrsv_NLN(long, double *, long, double *, long, void *);
int ztrsv_TUU(long, double *, long, double *, long, void *);
int ztrsv_TUN(long, double *, long, double *, long, void *);
int ztrsv_TLU(long, double *, long, double *, long, void *);
int ztrsv_TLN(long, double *, long, double *, long, void *);
int ztrsv_RUU(long, double *, long, double *, long, void *);
int ztrsv_RUN(long, double *, long, double *, long, void *);
int ztrsv_RLU(long, double *, long, double *, long, void *);
int ztrsv_RLN(long, double *, long, double *, long, void *);
int ztrsv_CUU(long, double *, long, double *, long, void *);
int ztrsv_CUN(long, double *, long, double *, long, void *);
int ztrsv_CLU(long, double *, long, double *, long, void *);
int ztrsv_CLN(long, double *, long, double *, long, void *);

int strmv_NUU(long, float *, long, float *, long, void *);
int strmv_NUN(long, float *, long, float *, long, void *);
int strmv_NLU(long, float *, long, float *, long, void *);
int strmv_NLN(long, float *, long, float *, long, void *);
int strmv_TUU(long, float *, long, float *, long, void *);
int strmv_TUN(long, float *, long, float *, long, void *);
int strmv_TLU(long, float *, long, float *, long, void *);
int strmv_TLN(long, float *, long, float *, long, void *);

int dtrmv_NUU(long, double *, long, double *, long, void *);
int dtrmv_NUN(long, double *, long, double *, long, void *);
int dtrmv_NLU(long, double *, long, double *, long, void *);
int dtrmv_NLN(long, double *, long, double *, long, void *);
int dtrmv_TUU(long, double *, long, double *, long, void *);
int dtrmv_TUN(long, double *, long, double *, long, void *);
int dtrmv_TLU(long, double *, long, double *, long, void *);
int dtrmv_TLN(long, double *, long, double *, long, void *);

int ctrmv_NUU(long, float *, long, float *, long, void *);
int ctrmv_NUN(long, float *, long, float *, long, void *);
int ctrmv_NLU(long, float *, long, float *, long, void *);
int ctrmv_NLN(long, float *, long, float *, long, void *);
int ctrmv_TUU(long, float *, long, float *, long, void *);
int ctrmv_TUN(long, float *, long, float *, long, void *);
int ctrmv_TLU(long, float *, long, float *, long, void *);
int ctrmv_TLN(long, float *, long, float *, long, void *);
int ctrmv_RUU(long, float *, long, float *, long, void *);
int ctrmv_RUN(long, float *, long, float *, long, void *);
int ctrmv_RLU(long, float *, long, float *, long, void *);
int ctrmv_RLN(long, float *, long, float *, long, void *);
int ctrmv_CUU(long, float *, long, float *, long, void *);
int ctrmv_CUN(long, float *, long, float *, long, void *);
int ctrmv_CLU(long, float *, long, float *, long, void *);
int ctrmv_CLN(long, float *, long, float *, long, void *);

int ztrmv_NUU(long, double *, long, double *, long, void *);
int ztrmv_NUN(long, double *, long, double *, long, void *);
int ztrmv_NLU(long, double *, long, double *, long, void *);
int ztrmv_NLN(long, double *, long, double *, long, void *);
int ztrmv_TUU(long, double *, long, double *, long, void *);
int ztrmv_TUN(long, double *, long, double *, long, void *);
int ztrmv_TLU(long, double *, long, double *, long, void *);
int ztrmv_TLN(long, double *, long, double *, long, void *);
int ztrmv_RUU(long, double *, long, double *, long, void *);
int ztrmv_RUN(long, double *, long, double *, long, void *);
int ztrmv_RLU(long, double *, long, double *, long, void *);
int ztrmv_RLN(long, double *, long, double *, long, void *);
int ztrmv_CUU(long, double *, long, double *, long, void *);
int ztrmv_CUN(long, double *, long, double *, long, void *);
int ztrmv_CLU(long, double *, long, double *, long, void *);
int ztrmv_CLN(long, double *, long, double *, long, void *);

#ifndef PARAMTEST
int ssymv_L(long, float,  float  *, long, float  *, long, float  *, long, void *);
int ssymv_U(long, float,  float  *, long, float  *, long, float  *, long, void *);

int dsymv_L(long, double, double *, long, double *, long, double *, long, void *);
int dsymv_U(long, double, double *, long, double *, long, double *, long, void *);
#else
int ssymv_L(long, float,  float  *, long, float  *, long, float  *, long, void *, long, long);
int ssymv_U(long, float,  float  *, long, float  *, long, float  *, long, void *, long, long);

int dsymv_L(long, double, double *, long, double *, long, double *, long, void *, long, long);
int dsymv_U(long, double, double *, long, double *, long, double *, long, void *, long, long);
#endif

int ssyr_L(long, float,  float  *, long, float  *, long, float *);
int ssyr_U(long, float,  float  *, long, float  *, long, float *);

int dsyr_L(long, double, double *, long, double *, long, double *);
int dsyr_U(long, double, double *, long, double *, long, double *);

int ssyr2_L(long, float,  float  *, long, float *, long, float  *, long, float *);
int ssyr2_U(long, float,  float  *, long, float *, long, float  *, long, float *);

int dsyr2_L(long, double, double *, long, double *, long, double *, long, double *);
int dsyr2_U(long, double, double *, long, double *, long, double *, long, double *);

int cher_L(long, float, float  *, long, float  *, long, float *);
int cher_U(long, float, float  *, long, float  *, long, float *);

int zher_L(long, double, double *, long, double *, long, double *);
int zher_U(long, double, double *, long, double *, long, double *);

int cher2_L(long, float,  float, float  *, long, float *, long, float  *, long, float *);
int cher2_U(long, float,  float, float  *, long, float *, long, float  *, long, float *);

int zher2_L(long, double, double, double *, long, double *, long, double *, long, double *);
int zher2_U(long, double, double, double *, long, double *, long, double *, long, double *);


int chemv_L(long, float,  float, float  *, long, float  *, long, float  *, long, void *);
int chemv_U(long, float,  float, float  *, long, float  *, long, float  *, long, void *);

int zhemv_L(long, double, double, double *, long, double *, long, double *, long, void *);
int zhemv_U(long, double, double, double *, long, double *, long, double *, long, void *);

int snorm_n(long, long, float  *a, long);
int snorm_t(long, long, float  *a, long);
int dnorm_n(long, long, double *a, long);
int dnorm_t(long, long, double *a, long);
int cnorm_n(long, long, float  *a, long);
int cnorm_t(long, long, float  *a, long);
int znorm_n(long, long, double *a, long);
int znorm_t(long, long, double *a, long);

#ifdef DOUBLE

#define GEMV      dgemv_
#define ZGEMV_    zgemv_

#define GEMV_NCOPY      dgemv_ncopy
#define GEMV_TCOPY      dgemv_tcopy
#define GEMV_KERNEL     dgemv_kernel

#define GEMV_N    dgemv_n
#define GEMV_T    dgemv_t
#define ZGEMV_N   zgemv_n
#define ZGEMV_T   zgemv_t
#define ZGEMV_R   zgemv_r
#define ZGEMV_C   zgemv_c
#define ZGEMV_O   zgemv_o
#define ZGEMV_U   zgemv_u
#define ZGEMV_S   zgemv_s
#define ZGEMV_D   zgemv_d

#define GER_K      dger_k
#define ZGERU_K   zgeru_k
#define ZGERC_K   zgerc_k

#define TRSV_NUU dtrsv_NUU
#define TRSV_NUN dtrsv_NUN
#define TRSV_NLU dtrsv_NLU
#define TRSV_NLN dtrsv_NLN
#define TRSV_TUU dtrsv_TUU
#define TRSV_TUN dtrsv_TUN
#define TRSV_TLU dtrsv_TLU
#define TRSV_TLN dtrsv_TLN

#define ZTRSV_NUU ztrsv_NUU
#define ZTRSV_NUN ztrsv_NUN
#define ZTRSV_NLU ztrsv_NLU
#define ZTRSV_NLN ztrsv_NLN
#define ZTRSV_TUU ztrsv_TUU
#define ZTRSV_TUN ztrsv_TUN
#define ZTRSV_TLU ztrsv_TLU
#define ZTRSV_TLN ztrsv_TLN
#define ZTRSV_RUU ztrsv_RUU
#define ZTRSV_RUN ztrsv_RUN
#define ZTRSV_RLU ztrsv_RLU
#define ZTRSV_RLN ztrsv_RLN
#define ZTRSV_CUU ztrsv_CUU
#define ZTRSV_CUN ztrsv_CUN
#define ZTRSV_CLU ztrsv_CLU
#define ZTRSV_CLN ztrsv_CLN

#else

#define GEMV      sgemv_
#define ZGEMV_    cgemv_

#define GEMV_N    sgemv_n
#define GEMV_T    sgemv_t
#define ZGEMV_N   cgemv_n
#define ZGEMV_T   cgemv_t
#define ZGEMV_R   cgemv_r
#define ZGEMV_C   cgemv_c
#define ZGEMV_O   cgemv_o
#define ZGEMV_U   cgemv_u
#define ZGEMV_S   cgemv_s
#define ZGEMV_D   cgemv_d

#define GER_K     sger_k
#define ZGERU_K   cgeru_k
#define ZGERC_K   cgerc_k

#define TRSV_NUU strsv_NUU
#define TRSV_NUN strsv_NUN
#define TRSV_NLU strsv_NLU
#define TRSV_NLN strsv_NLN
#define TRSV_TUU strsv_TUU
#define TRSV_TUN strsv_TUN
#define TRSV_TLU strsv_TLU
#define TRSV_TLN strsv_TLN

#define ZTRSV_NUU ctrsv_NUU
#define ZTRSV_NUN ctrsv_NUN
#define ZTRSV_NLU ctrsv_NLU
#define ZTRSV_NLN ctrsv_NLN
#define ZTRSV_TUU ctrsv_TUU
#define ZTRSV_TUN ctrsv_TUN
#define ZTRSV_TLU ctrsv_TLU
#define ZTRSV_TLN ctrsv_TLN
#define ZTRSV_RUU ctrsv_RUU
#define ZTRSV_RUN ctrsv_RUN
#define ZTRSV_RLU ctrsv_RLU
#define ZTRSV_RLN ctrsv_RLN
#define ZTRSV_CUU ctrsv_CUU
#define ZTRSV_CUN ctrsv_CUN
#define ZTRSV_CLU ctrsv_CLU
#define ZTRSV_CLN ctrsv_CLN

#endif

#endif
#endif
