/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LEVEL3
#define COMMON_LEVEL3

#ifndef ASSEMBLER

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
extern long sgemm_p;
extern long sgemm_q;
extern long sgemm_r;
extern long dgemm_p;
extern long dgemm_q;
extern long dgemm_r;
extern long cgemm_p;
extern long cgemm_q;
extern long cgemm_r;
extern long zgemm_p;
extern long zgemm_q;
extern long zgemm_r;
#endif

int sgemm_beta(long, long, long, float, 
	       float  *, long, float   *, long, float  *, long);
int dgemm_beta(long, long, long, double, 
	       double *, long, double  *, long, double *, long);
int cgemm_beta(long, long, long, float,  float,
	       float  *, long, float   *, long, float  *, long);
int zgemm_beta(long, long, long, double, double,
	       double *, long, double  *, long, double *, long);

int ssyrk_beta_U(long, long, long, float, 
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int ssyrk_beta_L(long, long, long, float, 
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int dsyrk_beta_U(long, long, long, double, 
		 double *, long, double  *, long, double *, long, double *, long, long);
int dsyrk_beta_L(long, long, long, double, 
		 double *, long, double  *, long, double *, long, double *, long, long);
int csyrk_beta_U(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int csyrk_beta_L(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int zsyrk_beta_U(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int zsyrk_beta_L(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int cherk_beta_U(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int cherk_beta_L(long, long, long, float,  float,
		 float  *, long, float   *, long, float  *, long, float *, long, long);
int zherk_beta_U(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);
int zherk_beta_L(long, long, long, double, double,
		 double *, long, double  *, long, double *, long, double *, long, long);

void sgemm_incopy(long m, long n, float *a, long lda, float *b);
void sgemm_itcopy(long m, long n, float *a, long lda, float *b);
void sgemm_oncopy(long m, long n, float *a, long lda, float *b);
void sgemm_otcopy(long m, long n, float *a, long lda, float *b);
void dgemm_incopy(long m, long n, double *a, long lda, double *b);
void dgemm_itcopy(long m, long n, double *a, long lda, double *b);
void dgemm_oncopy(long m, long n, double *a, long lda, double *b);
void dgemm_otcopy(long m, long n, double *a, long lda, double *b);
void cgemm_incopy(long m, long n, float *a, long lda, float *b);
void cgemm_itcopy(long m, long n, float *a, long lda, float *b);
void cgemm_oncopy(long m, long n, float *a, long lda, float *b);
void cgemm_otcopy(long m, long n, float *a, long lda, float *b);
void zgemm_incopy(long m, long n, double *a, long lda, double *b);
void zgemm_itcopy(long m, long n, double *a, long lda, double *b);
void zgemm_oncopy(long m, long n, double *a, long lda, double *b);
void zgemm_otcopy(long m, long n, double *a, long lda, double *b);

void strsm_kernel_LN(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_LT(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_RN(long, long, long, float, float *, float *, float *, long, long);
void strsm_kernel_RT(long, long, long, float, float *, float *, float *, long, long);

void dtrsm_kernel_LN(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_LT(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_RN(long, long, long, double, double *, double *, double *, long, long);
void dtrsm_kernel_RT(long, long, long, double, double *, double *, double *, long, long);

void ctrsm_kernel_LN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_LC(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrsm_kernel_RC(long, long, long, float, float, float *, float *, float *, long, long);

void ztrsm_kernel_LN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_LC(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrsm_kernel_RC(long, long, long, double, double, double *, double *, double *, long, long);

void strmm_kernel_RN(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_RT(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_LN(long, long, long, float, float *, float *, float *, long, long);
void strmm_kernel_LT(long, long, long, float, float *, float *, float *, long, long);

void dtrmm_kernel_RN(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_RT(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_LN(long, long, long, double, double *, double *, double *, long, long);
void dtrmm_kernel_LT(long, long, long, double, double *, double *, double *, long, long);

void ctrmm_kernel_RN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_RC(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LN(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LT(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LR(long, long, long, float, float, float *, float *, float *, long, long);
void ctrmm_kernel_LC(long, long, long, float, float, float *, float *, float *, long, long);

void ztrmm_kernel_RN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_RC(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LN(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LT(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LR(long, long, long, double, double, double *, double *, double *, long, long);
void ztrmm_kernel_LC(long, long, long, double, double, double *, double *, double *, long, long);

void strmm_iunucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iunncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iutucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iutncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ounucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ounncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_outucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_outncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ilnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_ilnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_iltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_olnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_olnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_oltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void strmm_oltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);

void dtrmm_iunucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iunncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iutucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iutncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ounucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ounncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_outucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_outncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ilnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_ilnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_iltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_olnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_olnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_oltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dtrmm_oltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void ctrmm_iunucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iunncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iutucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iutncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ounucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ounncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_outucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_outncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ilnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_ilnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_iltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_olnucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_olnncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_oltucopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ctrmm_oltncopy(long m, long n, float *a, long lda, long posX, long posY, float *b);

void ztrmm_iunucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iunncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iutucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iutncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ounucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ounncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_outucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_outncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ilnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_ilnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_iltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_olnucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_olnncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_oltucopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void ztrmm_oltncopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void strsm_iunucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iunncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iutucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iutncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ounucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ounncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_outucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_outncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ilnucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_ilnncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iltucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_iltncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_olnucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_olnncopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_oltucopy(long m, long n, float *a, long lda, long offset, float *b);
void strsm_oltncopy(long m, long n, float *a, long lda, long offset, float *b);

void dtrsm_iunucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iunncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iutucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iutncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ounucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ounncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_outucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_outncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ilnucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_ilnncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iltucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_iltncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_olnucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_olnncopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_oltucopy(long m, long n, double *a, long lda, long offset, double *b);
void dtrsm_oltncopy(long m, long n, double *a, long lda, long offset, double *b);

void ctrsm_iunucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iunncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iutucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iutncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ounucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ounncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_outucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_outncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ilnucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_ilnncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iltucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_iltncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_olnucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_olnncopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_oltucopy(long m, long n, float *a, long lda, long offset, float *b);
void ctrsm_oltncopy(long m, long n, float *a, long lda, long offset, float *b);

void ztrsm_iunucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iunncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iutucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iutncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ounucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ounncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_outucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_outncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ilnucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_ilnncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iltucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_iltncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_olnucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_olnncopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_oltucopy(long m, long n, double *a, long lda, long offset, double *b);
void ztrsm_oltncopy(long m, long n, double *a, long lda, long offset, double *b);

void ssymm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void ssymm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void dsymm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void dsymm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void csymm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void csymm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void zsymm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zsymm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void chemm_iutcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_outcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_iltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void chemm_oltcopy(long m, long n, float *a, long lda, long posX, long posY, float *b);
void zhemm_iutcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_outcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_iltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);
void zhemm_oltcopy(long m, long n, double *a, long lda, long posX, long posY, double *b);

void ssyrk_kernel_U(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void ssyrk_kernel_L(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void dsyrk_kernel_U(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void dsyrk_kernel_L(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void csyrk_kernel_U(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void csyrk_kernel_L(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void zsyrk_kernel_U(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zsyrk_kernel_L(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void ssyr2k_kernel_U(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void ssyr2k_kernel_L(long m, long n, long k, float alpha, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void dsyr2k_kernel_U(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void dsyr2k_kernel_L(long m, long n, long k, double alpha, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void csyr2k_kernel_U(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void csyr2k_kernel_L(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void zsyr2k_kernel_U(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zsyr2k_kernel_L(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);

void cherk_kernel_UN(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_UC(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_LN(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void cherk_kernel_LC(long m, long n, long k, float alpha_r, float *a, float *b, float *c, long ldc, long offset, float *subbuffer);
void zherk_kernel_UN(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_UC(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_LN(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);
void zherk_kernel_LC(long m, long n, long k, double alpha_r, double *a, double *b, double *c, long ldc, long offset, double *subbuffer);

void cher2k_kernel_UN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_UC(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_LN(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void cher2k_kernel_LC(long m, long n, long k, float alpha_r, float alpha_i, float *a, float *b, float *c, long ldc, long offset, float *subbuffer, int flag);
void zher2k_kernel_UN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_UC(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_LN(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);
void zher2k_kernel_LC(long m, long n, long k, double alpha_r, double alpha_i, double *a, double *b, double *c, long ldc, long offset, double *subbuffer, int flag);

#if 1
unsigned int sgemm_kernel(long, long, long, float,  float  *, float  *, float  *, long);
unsigned int dgemm_kernel(long, long, long, double, double *, double *, double *, long);
unsigned int sgemm_kernel_b0(long, long, long, float,  float  *, float  *, float  *, long);
unsigned int dgemm_kernel_b0(long, long, long, double, double *, double *, double *, long);
#else
unsigned int sgemm_kernel(long, long, long, float,  float  *, float  *, float  *, long, long, long, long);
unsigned int dgemm_kernel(long, long, long, double, double *, double *, double *, long, long, long, long);
unsigned int sgemm_kernel_b0(long, long, long, float,  float  *, float  *, float  *, long, long, long, long);
unsigned int dgemm_kernel_b0(long, long, long, double, double *, double *, double *, long, long, long, long);
#endif

unsigned int cgemm_kernel_n(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_l(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_r(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_b(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_n_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_l_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_r_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);
unsigned int cgemm_kernel_b_b0(long, long, long, float,  float,  float  *, float  *, float  *, long);

unsigned int zgemm_kernel_n(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_l(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_r(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_b(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_n_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_l_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_r_b0(long, long, long, double, double, double *, double *, double *, long);
unsigned int zgemm_kernel_b_b0(long, long, long, double, double, double *, double *, double *, long);

GEMMRETTYPE sgemm_nn(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE sgemm_nt(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE sgemm_tn(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE sgemm_tt(long, long, long, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE dgemm_nn(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE dgemm_nt(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE dgemm_tn(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE dgemm_tt(long, long, long, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE cgemm_nn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_nc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_tc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rt(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_rc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cn(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_ct(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cr(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE cgemm_cc(long, long, long, float, float, float *, long,
		     float *, long, float *, long, float *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

GEMMRETTYPE zgemm_nn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_nc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_tc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rt(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_rc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cn(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_ct(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cr(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);
GEMMRETTYPE zgemm_cc(long, long, long, double, double, double *, long,
		     double *, long, double *, long, double *
#ifdef PARAMTEST
		     , long, long, long
#endif
#ifdef PREFETCHTEST
		     , long, long, long
#endif
);

int strsm_LNUU(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);
int strsm_LNUN(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);
int strsm_LNLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LNLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_LTLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RNLN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTUU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTUN(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTLU(long, long, long, float,
		   float *, long, float *, long, float *, long, float *);
int strsm_RTLN(long, long, long, float, 
		   float *, long, float *, long, float *, long, float *);

int dtrsm_LNUU(long, long, long, double, 
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LNLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_LTLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RNLN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTUU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTUN(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTLU(long, long, long, double,
		   double *, long, double *, long, double *, long, double *);
int dtrsm_RTLN(long, long, long, double, 
		   double *, long, double *, long, double *, long, double *);

int ctrsm_LNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_LCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrsm_RCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);

int ztrsm_LNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_LCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrsm_RCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);

int strmm_LNUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LNLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_LTLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RNLN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTUU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTUN(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTLU(long, long, long, float, float *, long, float *, long, float *, long, float *);
int strmm_RTLN(long, long, long, float, float *, long, float *, long, float *, long, float *);

int dtrmm_LNUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LNLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_LTLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RNLN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTUU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTUN(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTLU(long, long, long, double, double *, long, double *, long, double *, long, double *);
int dtrmm_RTLN(long, long, long, double, double *, long, double *, long, double *, long, double *);

int ctrmm_LNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_LCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RNLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RTLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RRLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCUU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCUN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCLU(long, long, long, float, float, float *, long, float *, long, float *, long, float *);
int ctrmm_RCLN(long, long, long, float, float, float *, long, float *, long, float *, long, float *);

int ztrmm_LNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_LCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RNLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RTLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RRLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCUU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCUN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCLU(long, long, long, double, double, double *, long, double *, long, double *, long, double *);
int ztrmm_RCLN(long, long, long, double, double, double *, long, double *, long, double *, long, double *);

int ssymm_LU(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_LL(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_RU(long, long, long, float , float  *, long, float  *, long, float *, long, float *);
int ssymm_RL(long, long, long, float , float  *, long, float  *, long, float *, long, float *);

int dsymm_LU(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_LL(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_RU(long, long, long, double, double *, long, double *, long, double *,long, double *);
int dsymm_RL(long, long, long, double, double *, long, double *, long, double *,long, double *);

int csymm_LU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_LL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_RU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int csymm_RL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);

int zsymm_LU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_LL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_RU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zsymm_RL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);

int ssyrk_UN(long, long, long, float , 
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_UT(long, long, long, float ,
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_LN(long, long, long, float , 
		 float  *, long, float *, long, float *, long, float *, long, long);
int ssyrk_LT(long, long, long, float ,
		 float  *, long, float *, long, float *, long, float *, long, long);

int dsyrk_UN(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_UT(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_LN(long, long, long, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int dsyrk_LT(long, long, long, double, 
		 double *, long, double*, long, double *, long, double *, long, long);

int csyrk_UN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_UT(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_LN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int csyrk_LT(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);

int zsyrk_UN(long, long, long, double, double, 
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_UT(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_LN(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);
int zsyrk_LT(long, long, long, double, double,
		 double *, long, double*, long, double *, long, double *, long, long);

int ssyr2k_UN(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_UT(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_LN(long, long, long, float , 
		  float  *, long, float  *, long, float *, long, float *, long, long);
int ssyr2k_LT(long, long, long, float ,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int dsyr2k_UN(long, long, long, double, 
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_UT(long, long, long, double, 
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_LN(long, long, long, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int dsyr2k_LT(long, long, long, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int csyr2k_UN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_UT(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_LN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int csyr2k_LT(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int zsyr2k_UN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_UT(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_LN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zsyr2k_LT(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int chemm_LU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_LL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_RU(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);
int chemm_RL(long, long, long, float, float, float  *, long, float  *, long, float *, long, float *);

int zhemm_LU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_LL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_RU(long, long, long, double, double, double *, long, double *, long, double *,long, double *);
int zhemm_RL(long, long, long, double, double, double *, long, double *, long, double *,long, double *);

int cherk_UN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_UC(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_LN(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);
int cherk_LC(long, long, long, float, float,
		 float  *, long, float *, long, float *, long, float *, long, long);

int zherk_UN(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_UC(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_LN(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);
int zherk_LC(long, long, long, double, double,
		 double *, long, double *, long, double *, long, double *, long, long);

int cher2k_UN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_UC(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_LN(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);
int cher2k_LC(long, long, long, float, float,
		  float  *, long, float  *, long, float *, long, float *, long, long);

int zher2k_UN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_UC(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_LN(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);
int zher2k_LC(long, long, long, double, double,
		  double *, long, double *, long, double*, long, double *, long, long);

int sgemt_n(long, long, float,  float   *, long, float    *, int);
int sgemt_t(long, long, float,  float   *, long, float    *, int);
int dgemt_n(long, long, double, double  *, long, double   *, int);
int dgemt_t(long, long, double, double  *, long, double   *, int);

int cgemt_n(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_t(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_r(long, long, float,  float,  float   *, long, float    *, int);
int cgemt_c(long, long, float,  float,  float   *, long, float    *, int);
int zgemt_n(long, long, double, double, double  *, long, double   *, int);
int zgemt_t(long, long, double, double, double  *, long, double   *, int);
int zgemt_r(long, long, double, double, double  *, long, double   *, int);
int zgemt_c(long, long, double, double, double  *, long, double   *, int);

int sgema_n(long, long, float,  float   *, long, float    *, long);
int sgema_t(long, long, float,  float   *, long, float    *, long);
int dgema_n(long, long, double, double  *, long, double   *, long);
int dgema_t(long, long, double, double  *, long, double   *, long);

int cgema_n(long, long, float,  float,  float   *, long, float    *, long);
int cgema_t(long, long, float,  float,  float   *, long, float    *, long);
int cgema_r(long, long, float,  float,  float   *, long, float    *, long);
int cgema_c(long, long, float,  float,  float   *, long, float    *, long);
int zgema_n(long, long, double, double, double  *, long, double   *, long);
int zgema_t(long, long, double, double, double  *, long, double   *, long);
int zgema_r(long, long, double, double, double  *, long, double   *, long);
int zgema_c(long, long, double, double, double  *, long, double   *, long);

#endif

#ifndef COMPLEX
#ifndef DOUBLE
/* REAL && SINGLE */
#define   GEMM_BETA   sgemm_beta
#define   GEMM_ONCOPY sgemm_oncopy
#define   GEMM_OTCOPY sgemm_otcopy
#if SGEMM_UNROLL_M != SGEMM_UNROLL_N
#define   GEMM_INCOPY sgemm_incopy
#define   GEMM_ITCOPY sgemm_itcopy
#else
#define   GEMM_INCOPY sgemm_oncopy
#define   GEMM_ITCOPY sgemm_otcopy
#endif
#define   GEMM_KERNEL_N sgemm_kernel
#define   GEMM_KERNEL_N_B0 sgemm_kernel_b0
#define   GEMM_NN     sgemm_nn
#define   GEMM_NT     sgemm_nt
#define   GEMM_TN     sgemm_tn
#define   GEMM_TT     sgemm_tt
#else
/* REAL && DOUBLE */
#define   GEMM_BETA   dgemm_beta
#define   GEMM_ONCOPY dgemm_oncopy
#define   GEMM_OTCOPY dgemm_otcopy
#if DGEMM_UNROLL_M != DGEMM_UNROLL_N
#define   GEMM_INCOPY dgemm_incopy
#define   GEMM_ITCOPY dgemm_itcopy
#else
#define   GEMM_INCOPY dgemm_oncopy
#define   GEMM_ITCOPY dgemm_otcopy
#endif
#define   GEMM_KERNEL_N dgemm_kernel
#define   GEMM_KERNEL_N_B0 dgemm_kernel_b0
#define   GEMM_NN     dgemm_nn
#define   GEMM_NT     dgemm_nt
#define   GEMM_TN     dgemm_tn
#define   GEMM_TT     dgemm_tt
#endif
#else
#ifndef DOUBLE
/* COMPLEX && SINGLE */
#define   GEMM_BETA   cgemm_beta
#define   GEMM_ONCOPY cgemm_oncopy
#define   GEMM_OTCOPY cgemm_otcopy
#if CGEMM_UNROLL_M != CGEMM_UNROLL_N
#define   GEMM_INCOPY cgemm_incopy
#define   GEMM_ITCOPY cgemm_itcopy
#else
#define   GEMM_INCOPY cgemm_oncopy
#define   GEMM_ITCOPY cgemm_otcopy
#endif
#define   GEMM_KERNEL_N cgemm_kernel_n
#define   GEMM_KERNEL_L cgemm_kernel_l
#define   GEMM_KERNEL_R cgemm_kernel_r
#define   GEMM_KERNEL_B cgemm_kernel_b
#define   GEMM_KERNEL_N_B0 cgemm_kernel_n_b0
#define   GEMM_KERNEL_L_B0 cgemm_kernel_l_b0
#define   GEMM_KERNEL_R_B0 cgemm_kernel_r_b0
#define   GEMM_KERNEL_B_B0 cgemm_kernel_b_b0
#define   GEMM_NN     cgemm_nn
#define   GEMM_CN     cgemm_cn
#define   GEMM_TN     cgemm_tn
#define   GEMM_NC     cgemm_nc
#define   GEMM_NT     cgemm_nt
#define   GEMM_CC     cgemm_cc
#define   GEMM_CT     cgemm_ct
#define   GEMM_TC     cgemm_tc
#define   GEMM_TT     cgemm_tt
#define   GEMM_NR     cgemm_nr
#define   GEMM_TR     cgemm_tr
#define   GEMM_CR     cgemm_cr
#define   GEMM_RN     cgemm_rn
#define   GEMM_RT     cgemm_rt
#define   GEMM_RC     cgemm_rc
#define   GEMM_RR     cgemm_rr
#else
/* COMPLEX && DOUBLE */
#define   GEMM_BETA   zgemm_beta
#define   GEMM_ONCOPY zgemm_oncopy
#define   GEMM_OTCOPY zgemm_otcopy
#if ZGEMM_UNROLL_M != ZGEMM_UNROLL_N
#define   GEMM_INCOPY zgemm_incopy
#define   GEMM_ITCOPY zgemm_itcopy
#else
#define   GEMM_INCOPY zgemm_oncopy
#define   GEMM_ITCOPY zgemm_otcopy
#endif
#define   GEMM_KERNEL_N zgemm_kernel_n
#define   GEMM_KERNEL_L zgemm_kernel_l
#define   GEMM_KERNEL_R zgemm_kernel_r
#define   GEMM_KERNEL_B zgemm_kernel_b
#define   GEMM_KERNEL_N_B0 zgemm_kernel_n_b0
#define   GEMM_KERNEL_L_B0 zgemm_kernel_l_b0
#define   GEMM_KERNEL_R_B0 zgemm_kernel_r_b0
#define   GEMM_KERNEL_B_B0 zgemm_kernel_b_b0
#define   GEMM_NN     zgemm_nn
#define   GEMM_CN     zgemm_cn
#define   GEMM_TN     zgemm_tn
#define   GEMM_NC     zgemm_nc
#define   GEMM_NT     zgemm_nt
#define   GEMM_CC     zgemm_cc
#define   GEMM_CT     zgemm_ct
#define   GEMM_TC     zgemm_tc
#define   GEMM_TT     zgemm_tt
#define   GEMM_NR     zgemm_nr
#define   GEMM_TR     zgemm_tr
#define   GEMM_CR     zgemm_cr
#define   GEMM_RN     zgemm_rn
#define   GEMM_RT     zgemm_rt
#define   GEMM_RC     zgemm_rc
#define   GEMM_RR     zgemm_rr
#endif
#endif

#ifndef COMPLEX
#ifndef DOUBLE
#define TRMM_KERNEL_LN  strmm_kernel_LN
#define TRMM_KERNEL_LT  strmm_kernel_LT
#define TRMM_KERNEL_RN  strmm_kernel_RN
#define TRMM_KERNEL_RT  strmm_kernel_RT

#define TRSM_KERNEL_LN  strsm_kernel_LN
#define TRSM_KERNEL_LT  strsm_kernel_LT
#define TRSM_KERNEL_RN  strsm_kernel_RN
#define TRSM_KERNEL_RT  strsm_kernel_RT
#else
#define TRMM_KERNEL_LN  dtrmm_kernel_LN
#define TRMM_KERNEL_LT  dtrmm_kernel_LT
#define TRMM_KERNEL_RN  dtrmm_kernel_RN
#define TRMM_KERNEL_RT  dtrmm_kernel_RT

#define TRSM_KERNEL_LN  dtrsm_kernel_LN
#define TRSM_KERNEL_LT  dtrsm_kernel_LT
#define TRSM_KERNEL_RN  dtrsm_kernel_RN
#define TRSM_KERNEL_RT  dtrsm_kernel_RT
#endif
#else
#ifndef DOUBLE
#define TRMM_KERNEL_LN  ctrmm_kernel_LN
#define TRMM_KERNEL_LT  ctrmm_kernel_LT
#define TRMM_KERNEL_LR  ctrmm_kernel_LR
#define TRMM_KERNEL_LC  ctrmm_kernel_LC
#define TRMM_KERNEL_RN  ctrmm_kernel_RN
#define TRMM_KERNEL_RT  ctrmm_kernel_RT
#define TRMM_KERNEL_RR  ctrmm_kernel_RR
#define TRMM_KERNEL_RC  ctrmm_kernel_RC

#define TRSM_KERNEL_LN  ctrsm_kernel_LN
#define TRSM_KERNEL_LT  ctrsm_kernel_LT
#define TRSM_KERNEL_LR  ctrsm_kernel_LR
#define TRSM_KERNEL_LC  ctrsm_kernel_LC
#define TRSM_KERNEL_RN  ctrsm_kernel_RN
#define TRSM_KERNEL_RT  ctrsm_kernel_RT
#define TRSM_KERNEL_RR  ctrsm_kernel_RR
#define TRSM_KERNEL_RC  ctrsm_kernel_RC
#else
#define TRMM_KERNEL_LN  ztrmm_kernel_LN
#define TRMM_KERNEL_LT  ztrmm_kernel_LT
#define TRMM_KERNEL_LR  ztrmm_kernel_LR
#define TRMM_KERNEL_LC  ztrmm_kernel_LC
#define TRMM_KERNEL_RN  ztrmm_kernel_RN
#define TRMM_KERNEL_RT  ztrmm_kernel_RT
#define TRMM_KERNEL_RR  ztrmm_kernel_RR
#define TRMM_KERNEL_RC  ztrmm_kernel_RC

#define TRSM_KERNEL_LN  ztrsm_kernel_LN
#define TRSM_KERNEL_LT  ztrsm_kernel_LT
#define TRSM_KERNEL_LR  ztrsm_kernel_LR
#define TRSM_KERNEL_LC  ztrsm_kernel_LC
#define TRSM_KERNEL_RN  ztrsm_kernel_RN
#define TRSM_KERNEL_RT  ztrsm_kernel_RT
#define TRSM_KERNEL_RR  ztrsm_kernel_RR
#define TRSM_KERNEL_RC  ztrsm_kernel_RC
#endif
#endif


#ifdef DOUBLE
#define TRSM_LNUU dtrsm_LNUU
#define TRSM_LNUN dtrsm_LNUN
#define TRSM_LNLU dtrsm_LNLU
#define TRSM_LNLN dtrsm_LNLN
#define TRSM_LTUU dtrsm_LTUU
#define TRSM_LTUN dtrsm_LTUN
#define TRSM_LTLU dtrsm_LTLU
#define TRSM_LTLN dtrsm_LTLN

#define TRSM_RNUU dtrsm_RNUU
#define TRSM_RNUN dtrsm_RNUN
#define TRSM_RNLU dtrsm_RNLU
#define TRSM_RNLN dtrsm_RNLN
#define TRSM_RTUU dtrsm_RTUU
#define TRSM_RTUN dtrsm_RTUN
#define TRSM_RTLU dtrsm_RTLU
#define TRSM_RTLN dtrsm_RTLN

#define ZTRSM_LNUU ztrsm_LNUU
#define ZTRSM_LNUN ztrsm_LNUN
#define ZTRSM_LNLU ztrsm_LNLU
#define ZTRSM_LNLN ztrsm_LNLN
#define ZTRSM_LTUU ztrsm_LTUU
#define ZTRSM_LTUN ztrsm_LTUN
#define ZTRSM_LTLU ztrsm_LTLU
#define ZTRSM_LTLN ztrsm_LTLN
#define ZTRSM_LRUU ztrsm_LRUU
#define ZTRSM_LRUN ztrsm_LRUN
#define ZTRSM_LRLU ztrsm_LRLU
#define ZTRSM_LRLN ztrsm_LRLN
#define ZTRSM_LCUU ztrsm_LCUU
#define ZTRSM_LCUN ztrsm_LCUN
#define ZTRSM_LCLU ztrsm_LCLU
#define ZTRSM_LCLN ztrsm_LCLN

#define ZTRSM_RNUU ztrsm_RNUU
#define ZTRSM_RNUN ztrsm_RNUN
#define ZTRSM_RNLU ztrsm_RNLU
#define ZTRSM_RNLN ztrsm_RNLN
#define ZTRSM_RTUU ztrsm_RTUU
#define ZTRSM_RTUN ztrsm_RTUN
#define ZTRSM_RTLU ztrsm_RTLU
#define ZTRSM_RTLN ztrsm_RTLN
#define ZTRSM_RRUU ztrsm_RRUU
#define ZTRSM_RRUN ztrsm_RRUN
#define ZTRSM_RRLU ztrsm_RRLU
#define ZTRSM_RRLN ztrsm_RRLN
#define ZTRSM_RCUU ztrsm_RCUU
#define ZTRSM_RCUN ztrsm_RCUN
#define ZTRSM_RCLU ztrsm_RCLU
#define ZTRSM_RCLN ztrsm_RCLN

#define  SYRK_UN   dsyrk_UN
#define  SYRK_UT   dsyrk_UT
#define  SYRK_LN   dsyrk_LN
#define  SYRK_LT   dsyrk_LT

#define ZSYRK_UN   zsyrk_UN
#define ZSYRK_UT   zsyrk_UT
#define ZSYRK_LN   zsyrk_LN
#define ZSYRK_LT   zsyrk_LT

#define  SYMCOPY_L dsymcopy_L
#define  SYMCOPY_U dsymcopy_U
#define ZSYMCOPY_L zsymcopy_L
#define ZSYMCOPY_U zsymcopy_U
#define ZHEMCOPY_L zhemcopy_L
#define ZHEMCOPY_U zhemcopy_U

#else
#define TRSM_LNUU strsm_LNUU
#define TRSM_LNUN strsm_LNUN
#define TRSM_LNLU strsm_LNLU
#define TRSM_LNLN strsm_LNLN
#define TRSM_LTUU strsm_LTUU
#define TRSM_LTUN strsm_LTUN
#define TRSM_LTLU strsm_LTLU
#define TRSM_LTLN strsm_LTLN

#define TRSM_RNUU strsm_RNUU
#define TRSM_RNUN strsm_RNUN
#define TRSM_RNLU strsm_RNLU
#define TRSM_RNLN strsm_RNLN
#define TRSM_RTUU strsm_RTUU
#define TRSM_RTUN strsm_RTUN
#define TRSM_RTLU strsm_RTLU
#define TRSM_RTLN strsm_RTLN

#define ZTRSM_LNUU ctrsm_LNUU
#define ZTRSM_LNUN ctrsm_LNUN
#define ZTRSM_LNLU ctrsm_LNLU
#define ZTRSM_LNLN ctrsm_LNLN
#define ZTRSM_LTUU ctrsm_LTUU
#define ZTRSM_LTUN ctrsm_LTUN
#define ZTRSM_LTLU ctrsm_LTLU
#define ZTRSM_LTLN ctrsm_LTLN
#define ZTRSM_LRUU ctrsm_LRUU
#define ZTRSM_LRUN ctrsm_LRUN
#define ZTRSM_LRLU ctrsm_LRLU
#define ZTRSM_LRLN ctrsm_LRLN
#define ZTRSM_LCUU ctrsm_LCUU
#define ZTRSM_LCUN ctrsm_LCUN
#define ZTRSM_LCLU ctrsm_LCLU
#define ZTRSM_LCLN ctrsm_LCLN

#define ZTRSM_RNUU ctrsm_RNUU
#define ZTRSM_RNUN ctrsm_RNUN
#define ZTRSM_RNLU ctrsm_RNLU
#define ZTRSM_RNLN ctrsm_RNLN
#define ZTRSM_RTUU ctrsm_RTUU
#define ZTRSM_RTUN ctrsm_RTUN
#define ZTRSM_RTLU ctrsm_RTLU
#define ZTRSM_RTLN ctrsm_RTLN
#define ZTRSM_RRUU ctrsm_RRUU
#define ZTRSM_RRUN ctrsm_RRUN
#define ZTRSM_RRLU ctrsm_RRLU
#define ZTRSM_RRLN ctrsm_RRLN
#define ZTRSM_RCUU ctrsm_RCUU
#define ZTRSM_RCUN ctrsm_RCUN
#define ZTRSM_RCLU ctrsm_RCLU
#define ZTRSM_RCLN ctrsm_RCLN

#define  SYRK_UN   ssyrk_UN
#define  SYRK_UT   ssyrk_UT
#define  SYRK_LN   ssyrk_LN
#define  SYRK_LT   ssyrk_LT

#define ZSYRK_UN   csyrk_UN
#define ZSYRK_UT   csyrk_UT
#define ZSYRK_LN   csyrk_LN
#define ZSYRK_LT   csyrk_LT

#define  SYMCOPY_L ssymcopy_L
#define  SYMCOPY_U ssymcopy_U
#define ZSYMCOPY_L csymcopy_L
#define ZSYMCOPY_U csymcopy_U
#define ZHEMCOPY_L chemcopy_L
#define ZHEMCOPY_U chemcopy_U
#endif

#ifdef UNIT
#ifndef COMPLEX
#ifndef DOUBLE
#define TRMM_IUNCOPY strmm_iunucopy
#define TRMM_IUTCOPY strmm_iutucopy
#define TRMM_ILNCOPY strmm_ilnucopy
#define TRMM_ILTCOPY strmm_iltucopy
#define TRMM_OUNCOPY strmm_ounucopy
#define TRMM_OUTCOPY strmm_outucopy
#define TRMM_OLNCOPY strmm_olnucopy
#define TRMM_OLTCOPY strmm_oltucopy

#define TRSM_IUNCOPY strsm_iunucopy
#define TRSM_IUTCOPY strsm_iutucopy
#define TRSM_ILNCOPY strsm_ilnucopy
#define TRSM_ILTCOPY strsm_iltucopy
#define TRSM_OUNCOPY strsm_ounucopy
#define TRSM_OUTCOPY strsm_outucopy
#define TRSM_OLNCOPY strsm_olnucopy
#define TRSM_OLTCOPY strsm_oltucopy
#else
#define TRMM_IUNCOPY dtrmm_iunucopy
#define TRMM_IUTCOPY dtrmm_iutucopy
#define TRMM_ILNCOPY dtrmm_ilnucopy
#define TRMM_ILTCOPY dtrmm_iltucopy
#define TRMM_OUNCOPY dtrmm_ounucopy
#define TRMM_OUTCOPY dtrmm_outucopy
#define TRMM_OLNCOPY dtrmm_olnucopy
#define TRMM_OLTCOPY dtrmm_oltucopy

#define TRSM_IUNCOPY dtrsm_iunucopy
#define TRSM_IUTCOPY dtrsm_iutucopy
#define TRSM_ILNCOPY dtrsm_ilnucopy
#define TRSM_ILTCOPY dtrsm_iltucopy
#define TRSM_OUNCOPY dtrsm_ounucopy
#define TRSM_OUTCOPY dtrsm_outucopy
#define TRSM_OLNCOPY dtrsm_olnucopy
#define TRSM_OLTCOPY dtrsm_oltucopy
#endif
#else
#ifndef DOUBLE
#define TRMM_IUNCOPY ctrmm_iunucopy
#define TRMM_IUTCOPY ctrmm_iutucopy
#define TRMM_ILNCOPY ctrmm_ilnucopy
#define TRMM_ILTCOPY ctrmm_iltucopy
#define TRMM_OUNCOPY ctrmm_ounucopy
#define TRMM_OUTCOPY ctrmm_outucopy
#define TRMM_OLNCOPY ctrmm_olnucopy
#define TRMM_OLTCOPY ctrmm_oltucopy

#define TRSM_IUNCOPY ctrsm_iunucopy
#define TRSM_IUTCOPY ctrsm_iutucopy
#define TRSM_ILNCOPY ctrsm_ilnucopy
#define TRSM_ILTCOPY ctrsm_iltucopy
#define TRSM_OUNCOPY ctrsm_ounucopy
#define TRSM_OUTCOPY ctrsm_outucopy
#define TRSM_OLNCOPY ctrsm_olnucopy
#define TRSM_OLTCOPY ctrsm_oltucopy
#else
#define TRMM_IUNCOPY ztrmm_iunucopy
#define TRMM_IUTCOPY ztrmm_iutucopy
#define TRMM_ILNCOPY ztrmm_ilnucopy
#define TRMM_ILTCOPY ztrmm_iltucopy
#define TRMM_OUNCOPY ztrmm_ounucopy
#define TRMM_OUTCOPY ztrmm_outucopy
#define TRMM_OLNCOPY ztrmm_olnucopy
#define TRMM_OLTCOPY ztrmm_oltucopy

#define TRSM_IUNCOPY ztrsm_iunucopy
#define TRSM_IUTCOPY ztrsm_iutucopy
#define TRSM_ILNCOPY ztrsm_ilnucopy
#define TRSM_ILTCOPY ztrsm_iltucopy
#define TRSM_OUNCOPY ztrsm_ounucopy
#define TRSM_OUTCOPY ztrsm_outucopy
#define TRSM_OLNCOPY ztrsm_olnucopy
#define TRSM_OLTCOPY ztrsm_oltucopy
#endif
#endif
#else
#ifndef COMPLEX
#ifndef DOUBLE
#define TRMM_IUNCOPY strmm_iunncopy
#define TRMM_IUTCOPY strmm_iutncopy
#define TRMM_ILNCOPY strmm_ilnncopy
#define TRMM_ILTCOPY strmm_iltncopy
#define TRMM_OUNCOPY strmm_ounncopy
#define TRMM_OUTCOPY strmm_outncopy
#define TRMM_OLNCOPY strmm_olnncopy
#define TRMM_OLTCOPY strmm_oltncopy

#define TRSM_IUNCOPY strsm_iunncopy
#define TRSM_IUTCOPY strsm_iutncopy
#define TRSM_ILNCOPY strsm_ilnncopy
#define TRSM_ILTCOPY strsm_iltncopy
#define TRSM_OUNCOPY strsm_ounncopy
#define TRSM_OUTCOPY strsm_outncopy
#define TRSM_OLNCOPY strsm_olnncopy
#define TRSM_OLTCOPY strsm_oltncopy
#else
#define TRMM_IUNCOPY dtrmm_iunncopy
#define TRMM_IUTCOPY dtrmm_iutncopy
#define TRMM_ILNCOPY dtrmm_ilnncopy
#define TRMM_ILTCOPY dtrmm_iltncopy
#define TRMM_OUNCOPY dtrmm_ounncopy
#define TRMM_OUTCOPY dtrmm_outncopy
#define TRMM_OLNCOPY dtrmm_olnncopy
#define TRMM_OLTCOPY dtrmm_oltncopy

#define TRSM_IUNCOPY dtrsm_iunncopy
#define TRSM_IUTCOPY dtrsm_iutncopy
#define TRSM_ILNCOPY dtrsm_ilnncopy
#define TRSM_ILTCOPY dtrsm_iltncopy
#define TRSM_OUNCOPY dtrsm_ounncopy
#define TRSM_OUTCOPY dtrsm_outncopy
#define TRSM_OLNCOPY dtrsm_olnncopy
#define TRSM_OLTCOPY dtrsm_oltncopy
#endif
#else
#ifndef DOUBLE
#define TRMM_IUNCOPY ctrmm_iunncopy
#define TRMM_IUTCOPY ctrmm_iutncopy
#define TRMM_ILNCOPY ctrmm_ilnncopy
#define TRMM_ILTCOPY ctrmm_iltncopy
#define TRMM_OUNCOPY ctrmm_ounncopy
#define TRMM_OUTCOPY ctrmm_outncopy
#define TRMM_OLNCOPY ctrmm_olnncopy
#define TRMM_OLTCOPY ctrmm_oltncopy

#define TRSM_IUNCOPY ctrsm_iunncopy
#define TRSM_IUTCOPY ctrsm_iutncopy
#define TRSM_ILNCOPY ctrsm_ilnncopy
#define TRSM_ILTCOPY ctrsm_iltncopy
#define TRSM_OUNCOPY ctrsm_ounncopy
#define TRSM_OUTCOPY ctrsm_outncopy
#define TRSM_OLNCOPY ctrsm_olnncopy
#define TRSM_OLTCOPY ctrsm_oltncopy
#else
#define TRMM_IUNCOPY ztrmm_iunncopy
#define TRMM_IUTCOPY ztrmm_iutncopy
#define TRMM_ILNCOPY ztrmm_ilnncopy
#define TRMM_ILTCOPY ztrmm_iltncopy
#define TRMM_OUNCOPY ztrmm_ounncopy
#define TRMM_OUTCOPY ztrmm_outncopy
#define TRMM_OLNCOPY ztrmm_olnncopy
#define TRMM_OLTCOPY ztrmm_oltncopy

#define TRSM_IUNCOPY ztrsm_iunncopy
#define TRSM_IUTCOPY ztrsm_iutncopy
#define TRSM_ILNCOPY ztrsm_ilnncopy
#define TRSM_ILTCOPY ztrsm_iltncopy
#define TRSM_OUNCOPY ztrsm_ounncopy
#define TRSM_OUTCOPY ztrsm_outncopy
#define TRSM_OLNCOPY ztrsm_olnncopy
#define TRSM_OLTCOPY ztrsm_oltncopy
#endif
#endif
#endif

#if GEMM_UNROLL_M == GEMM_UNROLL_N
#undef  TRMM_IUNCOPY
#undef  TRMM_IUTCOPY
#undef  TRMM_ILNCOPY
#undef  TRMM_ILTCOPY
#define TRMM_IUNCOPY TRMM_OUNCOPY
#define TRMM_IUTCOPY TRMM_OUTCOPY
#define TRMM_ILNCOPY TRMM_OLNCOPY
#define TRMM_ILTCOPY TRMM_OLTCOPY

#undef  TRSM_IUNCOPY
#undef  TRSM_IUTCOPY
#undef  TRSM_ILNCOPY
#undef  TRSM_ILTCOPY
#define TRSM_IUNCOPY TRSM_OUNCOPY
#define TRSM_IUTCOPY TRSM_OUTCOPY
#define TRSM_ILNCOPY TRSM_OLNCOPY
#define TRSM_ILTCOPY TRSM_OLTCOPY
#endif

#endif
