/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	16
#define ARGS	 0
	
#define M	 4 + STACK + ARGS(%esp)
#define ALPHA	16 + STACK + ARGS(%esp)
#ifdef DOUBLE
#define X	24 + STACK + ARGS(%esp)
#define INCX	28 + STACK + ARGS(%esp)
#define Y	32 + STACK + ARGS(%esp)
#define INCY	36 + STACK + ARGS(%esp)
#else
#define X	20 + STACK + ARGS(%esp)
#define INCX	24 + STACK + ARGS(%esp)
#define Y	28 + STACK + ARGS(%esp)
#define INCY	32 + STACK + ARGS(%esp)
#endif

.text
	.align 32
	.globl NAME
#if defined(__CYGWIN__) || defined(__WIN32__)
	.def NAME;.scl	2;.type	32;.endef
#endif
#ifdef linux
       .type NAME, @function
#endif
NAME:
	pushl	%edi
	pushl	%esi
	pushl	%ebx
	pushl	%ebp

#ifdef PROFILE
.data
	.align 4
.LP1:
	.long 0
.text
	movl $.LP1,%edx
	call mcount
#endif

	movsd	ALPHA, %xmm7

	movl	M,    %ebx
	movl	X,    %esi
	movl	INCX, %ecx
	movl	Y,    %edi
	movl	INCY, %edx

	leal	(, %ecx, SIZE), %ecx
	leal	(, %edx, SIZE), %edx

	cmpl	$SIZE, %ecx
	jne	.L14
	cmpl	$SIZE, %edx
	jne	.L14

	movl	%ebx, %eax
	sarl	$3,   %eax
	jle	.L15
	.align	8

.L16:
	movsd	0 * SIZE(%esi), %xmm0
	movsd	1 * SIZE(%esi), %xmm1
	movsd	2 * SIZE(%esi), %xmm2
	movsd	3 * SIZE(%esi), %xmm3

	mulsd	%xmm7, %xmm0
	addsd	0 * SIZE(%edi), %xmm0
	mulsd	%xmm7, %xmm1
	addsd	1 * SIZE(%edi), %xmm1
	mulsd	%xmm7, %xmm2
	addsd	2 * SIZE(%edi), %xmm2
	mulsd	%xmm7, %xmm3
	addsd	3 * SIZE(%edi), %xmm3

	movsd	%xmm0, 0 * SIZE(%edi)
	movsd	%xmm1, 1 * SIZE(%edi)
	movsd	%xmm2, 2 * SIZE(%edi)
	movsd	%xmm3, 3 * SIZE(%edi)

	movsd	4 * SIZE(%esi), %xmm0
	movsd	5 * SIZE(%esi), %xmm1
	movsd	6 * SIZE(%esi), %xmm2
	movsd	7 * SIZE(%esi), %xmm3

	mulsd	%xmm7, %xmm0
	addsd	4 * SIZE(%edi), %xmm0
	mulsd	%xmm7, %xmm1
	addsd	5 * SIZE(%edi), %xmm1
	mulsd	%xmm7, %xmm2
	addsd	6 * SIZE(%edi), %xmm2
	mulsd	%xmm7, %xmm3
	addsd	7 * SIZE(%edi), %xmm3
	addl	$8 * SIZE, %esi

	movsd	%xmm0, 4 * SIZE(%edi)
	movsd	%xmm1, 5 * SIZE(%edi)
	movsd	%xmm2, 6 * SIZE(%edi)
	movsd	%xmm3, 7 * SIZE(%edi)

	addl	$8 * SIZE, %edi
	decl	%eax
	jg	.L16
	.align 8

.L15:
	movl	%ebx, %eax
	andl	$7,   %eax
	jle	.L23
	.align	8

.L22:
	movsd	0 * SIZE(%esi), %xmm0
	addl	$SIZE, %esi
	mulsd	%xmm7, %xmm0
	addsd	0 * SIZE(%edi), %xmm0

	movsd	%xmm0, 	0 * SIZE(%edi)
	addl	$SIZE, %edi
	decl	%eax
	jg	.L22

.L23:
	xorl	%eax,%eax
	popl	%ebp
	popl	%ebx
	popl	%esi
	popl	%edi
	ret
	.align	8

.L14:
	movl	%ebx, %eax
	movl	%edi, %ebp
	sarl	$3,   %eax
	jle	.L28
	.align	8

.L29:
	movsd	(%esi), %xmm0
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm0
	addsd	(%ebp), %xmm0
	addl	%edx, %ebp

	movsd	(%esi), %xmm1
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm1
	addsd	(%ebp), %xmm1
	addl	%edx, %ebp

	movsd	(%esi), %xmm2
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm2
	addsd	(%ebp), %xmm2
	addl	%edx, %ebp

	movsd	(%esi), %xmm3
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm3
	addsd	(%ebp), %xmm3
	addl	%edx, %ebp

	movsd	%xmm0, (%edi)
	addl	%edx, %edi
	movsd	%xmm1, (%edi)
	addl	%edx, %edi
	movsd	%xmm2, (%edi)
	addl	%edx, %edi
	movsd	%xmm3, (%edi)
	addl	%edx, %edi

	movsd	(%esi), %xmm0
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm0
	addsd	(%ebp), %xmm0
	addl	%edx, %ebp

	movsd	(%esi), %xmm1
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm1
	addsd	(%ebp), %xmm1
	addl	%edx, %ebp

	movsd	(%esi), %xmm2
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm2
	addsd	(%ebp), %xmm2
	addl	%edx, %ebp

	movsd	(%esi), %xmm3
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm3
	addsd	(%ebp), %xmm3
	addl	%edx, %ebp

	movsd	%xmm0, (%edi)
	addl	%edx, %edi
	movsd	%xmm1, (%edi)
	addl	%edx, %edi
	movsd	%xmm2, (%edi)
	addl	%edx, %edi
	movsd	%xmm3, (%edi)
	addl	%edx, %edi

	decl	%eax
	jg	.L29
	.align 8

.L28:
	movl	%ebx, %eax
	andl	$7,   %eax
	jle	.L40
	.align	8

.L35:
	movsd	(%esi), %xmm0
	addl	%ecx, %esi
	mulsd	%xmm7, %xmm0
	addsd	(%edi), %xmm0
	movsd	%xmm0, (%edi)
	addl	%edx, %edi
	decl	%eax
	jg	.L35
	.align 8

.L40:
	xorl	%eax,%eax
	popl	%ebp
	popl	%ebx
	popl	%esi
	popl	%edi
	ret
#if !defined(__CYGWIN__) && !defined(__WIN32__)
.Lfe1:
	.size	 NAME, .Lfe1 - NAME
#endif
