/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	12
#define ARGS	 0
	
#define M	 4 + STACK + ARGS(%esp)
#ifdef DOUBLE
#define ALPHA_R	16 + STACK + ARGS(%esp)
#define ALPHA_I	24 + STACK + ARGS(%esp)
#define X	32 + STACK + ARGS(%esp)
#define INCX	36 + STACK + ARGS(%esp)
#define Y	40 + STACK + ARGS(%esp)
#define INCY	44 + STACK + ARGS(%esp)
#else
#define ALPHA_R	16 + STACK + ARGS(%esp)
#define ALPHA_I	20 + STACK + ARGS(%esp)
#define X	24 + STACK + ARGS(%esp)
#define INCX	28 + STACK + ARGS(%esp)
#define Y	32 + STACK + ARGS(%esp)
#define INCY	36 + STACK + ARGS(%esp)
#endif

#ifndef HAVE_SSE2
#define pxor	xorps
#define movsd	movlps
#endif

#ifdef PENTIUM4
#define PREFETCHSIZE	128
#endif

#ifdef OPTERON
#define PREFETCHSIZE	176
#define movsd		movlps
#endif

	PROLOGUE

	pushl	%edi
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movss	ALPHA_R, %xmm6		# 0    a
	movss	ALPHA_I, %xmm5		# 0    b

	pxor	%xmm7, %xmm7		#  0   0
	subss	%xmm5, %xmm7		#  0  -b

	unpcklps %xmm6, %xmm7		#  a   -b
	unpcklps %xmm5, %xmm6		#  b   a

	movlhps	 %xmm7, %xmm7
	movlhps	 %xmm6, %xmm6

	movl	M,    %ebx
	movl	X,    %esi
	movl	INCX, %ecx
	movl	Y,    %edi
	movl	INCY, %edx

	leal	(, %ecx, SIZE), %ecx
	leal	(, %edx, SIZE), %edx

	addl	%ecx, %ecx
	addl	%edx, %edx

	cmpl	$2 * SIZE, %ecx
	jne	.L14
	cmpl	$2 * SIZE, %edx
	jne	.L14

	movl	%ebx, %eax
	sarl	$3,   %eax
	jle	.L15
	.align	8

.L16:
#ifdef OPTERON
	prefetcht0     (PREFETCHSIZE + 0) * SIZE(%esi)
	prefetchw      (PREFETCHSIZE + 0) * SIZE(%edi)
#endif

	movss	 0 * SIZE(%esi), %xmm0		#  0  x1
	movss	 1 * SIZE(%esi), %xmm1		#  0  x2
	movss	 2 * SIZE(%esi), %xmm3		#  0  x1
	movss	 3 * SIZE(%esi), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(%edi), %xmm2
	movhps	 2 * SIZE(%edi), %xmm2
	movsd	 4 * SIZE(%edi), %xmm5
	movhps	 6 * SIZE(%edi), %xmm5

	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(%edi)
	movhps	%xmm2,  2 * SIZE(%edi)

	movss	 4 * SIZE(%esi), %xmm0		#  0  x1
	movss	 5 * SIZE(%esi), %xmm1		#  0  x2
	movss	 6 * SIZE(%esi), %xmm3		#  0  x1
	movss	 7 * SIZE(%esi), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	addps	%xmm0, %xmm5
	movsd	%xmm5,  4 * SIZE(%edi)
	movhps	%xmm5,  6 * SIZE(%edi)


	movss	 8 * SIZE(%esi), %xmm0		#  0  x1
	movss	 9 * SIZE(%esi), %xmm1		#  0  x2
	movss	10 * SIZE(%esi), %xmm3		#  0  x1
	movss	11 * SIZE(%esi), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 8 * SIZE(%edi), %xmm2
	movhps	10 * SIZE(%edi), %xmm2
	movsd	12 * SIZE(%edi), %xmm5
	movhps	14 * SIZE(%edi), %xmm5

	addps	%xmm0, %xmm2
	movsd	%xmm2,  8 * SIZE(%edi)
	movhps	%xmm2, 10 * SIZE(%edi)

	movss	12 * SIZE(%esi), %xmm0		#  0  x1
	movss	13 * SIZE(%esi), %xmm1		#  0  x2
	movss	14 * SIZE(%esi), %xmm3		#  0  x1
	movss	15 * SIZE(%esi), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	addps	%xmm0, %xmm5
	movsd	%xmm5, 12 * SIZE(%edi)
	movhps	%xmm5, 14 * SIZE(%edi)
	
	addl	$16 * SIZE, %esi
	addl	$16 * SIZE, %edi

	decl	%eax
	jg	.L16
	.align 8

.L15:
	movl	%ebx, %eax
	andl	$7,   %eax
	jle	.L40
	.align	8

.L22:
	movss	 0 * SIZE(%esi), %xmm0		#  0  x1
	movss	 1 * SIZE(%esi), %xmm1		#  0  x2

	unpcklps %xmm0, %xmm0			# x1  x1
	unpcklps %xmm1, %xmm1			# x2  x2
	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(%edi), %xmm2
	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(%edi)

	addl	$2 * SIZE, %esi
	addl	$2 * SIZE, %edi
	decl	%eax
	jg	.L22
	jmp	.L40
	.align	8

.L14:
	movl	%ebx, %eax
	sarl	$2,   %eax
	jle	.L28
	.align	8

.L29:
	movss	 0 * SIZE(%esi), %xmm0		#  0  x1
	movss	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi
	movss	 0 * SIZE(%esi), %xmm3		#  0  x1
	movss	 1 * SIZE(%esi), %xmm4		#  0  x2
	addl	%ecx, %esi

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(%edi), %xmm2
	movhps	 0 * SIZE(%edi, %edx), %xmm2

	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(%edi)
	addl	%edx, %edi
	movhps	%xmm2,  0 * SIZE(%edi)
	addl	%edx, %edi

	movss	 0 * SIZE(%esi), %xmm0		#  0  x1
	movss	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi
	movss	 0 * SIZE(%esi), %xmm3		#  0  x1
	movss	 1 * SIZE(%esi), %xmm4		#  0  x2
	addl	%ecx, %esi

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2

	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(%edi), %xmm5
	movhps	 0 * SIZE(%edi, %edx), %xmm5
	addps	%xmm0, %xmm5
	movsd	%xmm5,  0 * SIZE(%edi)
	addl	%edx, %edi
	movhps	%xmm5,  0 * SIZE(%edi)
	addl	%edx, %edi

	decl	%eax
	jg	.L29
	.align 8

.L28:
	movl	%ebx, %eax
	andl	$3,   %eax
	jle	.L40
	.align	8

.L35:
	movss	 0 * SIZE(%esi), %xmm0		#  0  x1
	movss	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi

	movsd	 0 * SIZE(%edi), %xmm2

	unpcklps %xmm0, %xmm0			# x1  x1
	unpcklps %xmm1, %xmm1			# x2  x2
	mulps	%xmm6, %xmm0			#  a * x1  b * x1
	mulps	%xmm7, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0
	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(%edi)
	addl	%edx, %edi

	decl	%eax
	jg	.L35
	.align 8

.L40:
	xorl	%eax,%eax
	popl	%ebx
	popl	%esi
	popl	%edi
	ret

	EPILOGUE

