/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	12
#define ARGS	 0
	
#define M	 4 + STACK + ARGS(%esp)
#ifdef DOUBLE
#define ALPHA_R	16 + STACK + ARGS(%esp)
#define ALPHA_I	24 + STACK + ARGS(%esp)
#define X	32 + STACK + ARGS(%esp)
#define INCX	36 + STACK + ARGS(%esp)
#define Y	40 + STACK + ARGS(%esp)
#define INCY	44 + STACK + ARGS(%esp)
#else
#define ALPHA_R	16 + STACK + ARGS(%esp)
#define ALPHA_I	20 + STACK + ARGS(%esp)
#define X	24 + STACK + ARGS(%esp)
#define INCX	28 + STACK + ARGS(%esp)
#define Y	32 + STACK + ARGS(%esp)
#define INCY	36 + STACK + ARGS(%esp)
#endif

#ifdef PENTIUM4
#define PREFETCHSIZE	128
#endif

#ifdef OPTERON
#define PREFETCHSIZE	 88
#define movsd	movlpd
#endif

	PROLOGUE

	pushl	%edi
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movsd	ALPHA_R, %xmm6		# 0    a
	movsd	ALPHA_I, %xmm5		# 0    b

	pxor	%xmm7, %xmm7		#  0   0
	subsd	%xmm5, %xmm7		#  0  -b

	unpcklpd %xmm6, %xmm7		#  a   -b
	unpcklpd %xmm5, %xmm6		#  b   a

	movl	M,    %ebx
	movl	X,    %esi
	movl	INCX, %ecx
	movl	Y,    %edi
	movl	INCY, %edx

	leal	(, %ecx, SIZE), %ecx
	leal	(, %edx, SIZE), %edx

	addl	%ecx, %ecx
	addl	%edx, %edx

	cmpl	$2 * SIZE, %ecx
	jne	.L14
	cmpl	$2 * SIZE, %edx
	jne	.L14

	movl	%ebx, %eax
	sarl	$3,   %eax
	jle	.L15
	.align	8

.L16:
#ifdef OPTERON
	prefetcht0     (PREFETCHSIZE + 0) * SIZE(%esi)
	prefetchw      (PREFETCHSIZE + 0) * SIZE(%edi)
#endif

	movsd	 0 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm1		#  0  x2
	movsd	 2 * SIZE(%esi), %xmm3		#  0  x1
	movsd	 3 * SIZE(%esi), %xmm4		#  0  x2

	movsd	 0 * SIZE(%edi), %xmm2
	movhpd	 1 * SIZE(%edi), %xmm2
	movsd	 2 * SIZE(%edi), %xmm5
	movhpd	 3 * SIZE(%edi), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2

	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2

	addpd	%xmm0, %xmm2
	addpd	%xmm1, %xmm2
	addpd	%xmm3, %xmm5
	addpd	%xmm4, %xmm5

	movsd	%xmm2,  0 * SIZE(%edi)
	movhpd	%xmm2,  1 * SIZE(%edi)
	movsd	%xmm5,  2 * SIZE(%edi)
	movhpd	%xmm5,  3 * SIZE(%edi)

	movsd	 4 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 5 * SIZE(%esi), %xmm1		#  0  x2
	movsd	 6 * SIZE(%esi), %xmm3		#  0  x1
	movsd	 7 * SIZE(%esi), %xmm4		#  0  x2

	movsd	 4 * SIZE(%edi), %xmm2
	movhpd	 5 * SIZE(%edi), %xmm2
	movsd	 6 * SIZE(%edi), %xmm5
	movhpd	 7 * SIZE(%edi), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2

	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2

	addpd	%xmm0, %xmm2
	addpd	%xmm1, %xmm2
	addpd	%xmm3, %xmm5
	addpd	%xmm4, %xmm5

	movsd	%xmm2,  4 * SIZE(%edi)
	movhpd	%xmm2,  5 * SIZE(%edi)
	movsd	%xmm5,  6 * SIZE(%edi)
	movhpd	%xmm5,  7 * SIZE(%edi)

#ifdef OPTERON
	prefetcht0     (PREFETCHSIZE + 8) * SIZE(%esi)
	prefetchw      (PREFETCHSIZE + 8) * SIZE(%edi)
#endif

	movsd	 8 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 9 * SIZE(%esi), %xmm1		#  0  x2
	movsd	10 * SIZE(%esi), %xmm3		#  0  x1
	movsd	11 * SIZE(%esi), %xmm4		#  0  x2

	movsd	 8 * SIZE(%edi), %xmm2
	movhpd	 9 * SIZE(%edi), %xmm2
	movsd	10 * SIZE(%edi), %xmm5
	movhpd	11 * SIZE(%edi), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2

	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2

	addpd	%xmm0, %xmm2
	addpd	%xmm1, %xmm2
	addpd	%xmm3, %xmm5
	addpd	%xmm4, %xmm5

	movsd	%xmm2,  8 * SIZE(%edi)
	movhpd	%xmm2,  9 * SIZE(%edi)
	movsd	%xmm5, 10 * SIZE(%edi)
	movhpd	%xmm5, 11 * SIZE(%edi)

	movsd	12 * SIZE(%esi), %xmm0		#  0  x1
	movsd	13 * SIZE(%esi), %xmm1		#  0  x2
	movsd	14 * SIZE(%esi), %xmm3		#  0  x1
	movsd	15 * SIZE(%esi), %xmm4		#  0  x2

	movsd	12 * SIZE(%edi), %xmm2
	movhpd	13 * SIZE(%edi), %xmm2
	movsd	14 * SIZE(%edi), %xmm5
	movhpd	15 * SIZE(%edi), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2

	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2

	addpd	%xmm0, %xmm2
	addpd	%xmm1, %xmm2
	addpd	%xmm3, %xmm5
	addpd	%xmm4, %xmm5

	movsd	%xmm2, 12 * SIZE(%edi)
	movhpd	%xmm2, 13 * SIZE(%edi)
	movsd	%xmm5, 14 * SIZE(%edi)
	movhpd	%xmm5, 15 * SIZE(%edi)

	addl	$16 * SIZE, %esi
	addl	$16 * SIZE, %edi
	decl	%eax
	jg	.L16
	.align 8

.L15:
	movl	%ebx, %eax
	andl	$7,   %eax
	jle	.L40
	.align	8

.L22:
	movsd	 0 * SIZE(%esi), %xmm0		#  0  c
	movsd	 1 * SIZE(%esi), %xmm1		#  0  d
	movsd	 0 * SIZE(%edi), %xmm2
	movhpd	 1 * SIZE(%edi), %xmm2

	unpcklpd %xmm0, %xmm0			# c  c
	unpcklpd %xmm1, %xmm1			# d  d
	mulpd	%xmm6, %xmm0			#  a * c  b * c
	mulpd	%xmm7, %xmm1			# -b * d  a * d
	addpd	%xmm1, %xmm2
	addpd	%xmm0, %xmm2

	movsd	%xmm2,  0 * SIZE(%edi)
	movhpd	%xmm2,  1 * SIZE(%edi)

	addl	$2 * SIZE, %esi
	addl	$2 * SIZE, %edi
	decl	%eax
	jg	.L22
	jmp	.L40
	.align	8

.L14:
	movl	%ebx, %eax
	sarl	$2,   %eax
	jle	.L28
	.align	8

.L29:
	movsd	 0 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi
	movsd	 0 * SIZE(%esi), %xmm3		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm4		#  0  x2
	addl	%ecx, %esi

	movsd	 0 * SIZE(%edi), %xmm2
	movhpd	 1 * SIZE(%edi), %xmm2
	movsd	 0 * SIZE(%edi, %edx), %xmm5
	movhpd	 1 * SIZE(%edi, %edx), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	addpd	%xmm1, %xmm0
	addpd	%xmm0, %xmm2

	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2
	addpd	%xmm4, %xmm3
	addpd	%xmm3, %xmm5

	movsd	%xmm2,  0 * SIZE(%edi)
	movhpd	%xmm2,  1 * SIZE(%edi)
	addl	%edx, %edi
	movsd	%xmm5,  0 * SIZE(%edi)
	movhpd	%xmm5,  1 * SIZE(%edi)
	addl	%edx, %edi

	movsd	 0 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi
	movsd	 0 * SIZE(%esi), %xmm3		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm4		#  0  x2
	addl	%ecx, %esi

	movsd	 0 * SIZE(%edi), %xmm2
	movhpd	 1 * SIZE(%edi), %xmm2
	movsd	 0 * SIZE(%edi, %edx), %xmm5
	movhpd	 1 * SIZE(%edi, %edx), %xmm5

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	addpd	%xmm1, %xmm0
	addpd	%xmm0, %xmm2

	unpcklpd %xmm3, %xmm3			# x1  x1
	unpcklpd %xmm4, %xmm4			# x2  x2
	mulpd	%xmm6, %xmm3			#  a * x1  b * x1
	mulpd	%xmm7, %xmm4			# -b * x2  a * x2
	addpd	%xmm4, %xmm3
	addpd	%xmm3, %xmm5

	movsd	%xmm2,  0 * SIZE(%edi)
	movhpd	%xmm2,  1 * SIZE(%edi)
	addl	%edx, %edi
	movsd	%xmm5,  0 * SIZE(%edi)
	movhpd	%xmm5,  1 * SIZE(%edi)
	addl	%edx, %edi

	decl	%eax
	jg	.L29
	.align 8

.L28:
	movl	%ebx, %eax
	andl	$3,   %eax
	jle	.L40
	.align	8

.L35:
	movsd	 0 * SIZE(%esi), %xmm0		#  0  x1
	movsd	 1 * SIZE(%esi), %xmm1		#  0  x2
	addl	%ecx, %esi

	movsd	 0 * SIZE(%edi), %xmm2
	movhpd	 1 * SIZE(%edi), %xmm2

	unpcklpd %xmm0, %xmm0			# x1  x1
	unpcklpd %xmm1, %xmm1			# x2  x2
	mulpd	%xmm6, %xmm0			#  a * x1  b * x1
	mulpd	%xmm7, %xmm1			# -b * x2  a * x2
	addpd	%xmm1, %xmm0
	addpd	%xmm0, %xmm2

	movsd	%xmm2,  0 * SIZE(%edi)
	movhpd	%xmm2,  1 * SIZE(%edi)
	addl	%edx, %edi

	decl	%eax
	jg	.L35
	.align 8

.L40:
	xorl	%eax,%eax
	popl	%ebx
	popl	%esi
	popl	%edi
	ret

	EPILOGUE
