/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define M	ARG1	/* rdi */
#define X	ARG4	/* rsi */
#define INCX	ARG5	/* rdx */
#define Y	ARG6	/* rcx */
#define INCY	ARG2	/* r8  */

#define	YY	%r10
#define ALPHA	%xmm15

#ifdef PENTIUM4
#define movlpd	movsd
#endif

#ifdef OPTERON
#define PREFETCH	prefetch
#define PREFETCHW	prefetchw
#define PREFETCHSIZE	86
#endif

	PROLOGUE
	PROFCODE

	movq	8(%rsp), INCY

	movapd	%xmm0,  ALPHA
	unpcklpd ALPHA, ALPHA

	leaq	(, INCX, SIZE), INCX
	leaq	(, INCY, SIZE), INCY

	cmpq	$SIZE, INCX
	jne	.L40
	cmpq	$SIZE, INCY
	jne	.L40

	testq	$SIZE, Y
	je	.L10

	movsd	0 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	0 * SIZE(Y), %xmm0
	movlpd	%xmm0, 0 * SIZE(Y)
	addq	$1 * SIZE, X
	addq	$1 * SIZE, Y
	decq	M
	jle	.L19
	.align	32

.L10:
	testq	$SIZE, X
	jne	.L20

	movq	M,  %rax
	sarq	$4, %rax
	jle	.L13

	movapd	0 * SIZE(X), %xmm0
	movapd	2 * SIZE(X), %xmm1
	movapd	4 * SIZE(X), %xmm2
	movapd	6 * SIZE(X), %xmm3

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5
	movapd	4 * SIZE(Y), %xmm6
	movapd	6 * SIZE(Y), %xmm7

	movapd	 8 * SIZE(X), %xmm8
	mulpd	ALPHA, %xmm0
	movapd	10 * SIZE(X), %xmm9
	mulpd	ALPHA, %xmm1
	movapd	12 * SIZE(X), %xmm10
	mulpd	ALPHA, %xmm2
	movapd	14 * SIZE(X), %xmm11
	mulpd	ALPHA, %xmm3

	decq	%rax
	jle .L12
	.align	8

.L11:
#ifdef OPTERON
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)
	PREFETCHW	(PREFETCHSIZE + 0) * SIZE(Y)
#endif
	addpd	%xmm4, %xmm0
	movapd	 8 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm8
	addpd	%xmm5, %xmm1
	movapd	10 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm9
	addpd	%xmm6, %xmm2
	movapd	12 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm10
	addpd	%xmm7, %xmm3
	movapd	14 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	16 * SIZE(X), %xmm0
	movapd	%xmm1, 2 * SIZE(Y)
	movapd	18 * SIZE(X), %xmm1
	movapd	%xmm2, 4 * SIZE(Y)
	movapd	20 * SIZE(X), %xmm2
	movapd	%xmm3, 6 * SIZE(Y)
	movapd	22 * SIZE(X), %xmm3

#ifdef OPTERON
	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)
	PREFETCHW	(PREFETCHSIZE + 8) * SIZE(Y)
#endif

	addpd	%xmm4, %xmm8
	movapd	16 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm0
	addpd	%xmm5, %xmm9
	movapd	18 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm1
	addpd	%xmm6, %xmm10
	movapd	20 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm2
	addpd	%xmm7, %xmm11
	movapd	22 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm3

	movapd	%xmm8,   8 * SIZE(Y)
	movapd	24 * SIZE(X), %xmm8
	movapd	%xmm9,  10 * SIZE(Y)
	movapd	26 * SIZE(X), %xmm9
	movapd	%xmm10, 12 * SIZE(Y)
	movapd	28 * SIZE(X), %xmm10
	movapd	%xmm11, 14 * SIZE(Y)
	movapd	30 * SIZE(X), %xmm11

	addq	$16 * SIZE, Y
	addq	$16 * SIZE, X
	decq	%rax
	jg	.L11
	.align 8

.L12:
	addpd	%xmm4, %xmm0
	movapd	 8 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm8
	addpd	%xmm5, %xmm1
	movapd	10 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm9
	addpd	%xmm6, %xmm2
	movapd	12 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm10
	addpd	%xmm7, %xmm3
	movapd	14 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	16 * SIZE(X), %xmm0
	movapd	%xmm1, 2 * SIZE(Y)
	movapd	18 * SIZE(X), %xmm1
	movapd	%xmm2, 4 * SIZE(Y)
	movapd	20 * SIZE(X), %xmm2
	movapd	%xmm3, 6 * SIZE(Y)
	movapd	22 * SIZE(X), %xmm3

	addpd	%xmm4, %xmm8
	mulpd	ALPHA, %xmm0
	addpd	%xmm5, %xmm9
	mulpd	ALPHA, %xmm1
	addpd	%xmm6, %xmm10
	mulpd	ALPHA, %xmm2
	addpd	%xmm7, %xmm11
	mulpd	ALPHA, %xmm3

	movapd	%xmm8,   8 * SIZE(Y)
	movapd	%xmm9,  10 * SIZE(Y)
	movapd	%xmm10, 12 * SIZE(Y)
	movapd	%xmm11, 14 * SIZE(Y)

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	.align 8

.L13:
	movq	M,  %rax
	andq	$8, %rax
	jle	.L14
	.align	8

	movapd	0 * SIZE(X), %xmm0
	movapd	2 * SIZE(X), %xmm1
	movapd	4 * SIZE(X), %xmm2
	movapd	6 * SIZE(X), %xmm3

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5
	movapd	4 * SIZE(Y), %xmm6
	movapd	6 * SIZE(Y), %xmm7

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	addpd	%xmm4, %xmm0
	addpd	%xmm5, %xmm1
	addpd	%xmm6, %xmm2
	addpd	%xmm7, %xmm3

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	%xmm1, 2 * SIZE(Y)
	movapd	%xmm2, 4 * SIZE(Y)
	movapd	%xmm3, 6 * SIZE(Y)

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	.align 8

.L14:
	movq	M,  %rax
	andq	$4, %rax
	jle	.L15
	.align	8

	movapd	0 * SIZE(X), %xmm0
	movapd	2 * SIZE(X), %xmm1

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1

	addpd	%xmm4, %xmm0
	addpd	%xmm5, %xmm1

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	%xmm1, 2 * SIZE(Y)

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	.align 8

.L15:
	movq	M,  %rax
	andq	$2, %rax
	jle	.L16
	.align	8

	movapd	0 * SIZE(X), %xmm0
	movapd	0 * SIZE(Y), %xmm4

	mulpd	ALPHA, %xmm0
	addpd	%xmm4, %xmm0

	movapd	%xmm0, 0 * SIZE(Y)

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	.align 8

.L16:
	movq	M,  %rax
	andq	$1, %rax
	jle	.L19
	.align	8

	movsd	0 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	0 * SIZE(Y), %xmm0

	movlpd	%xmm0, 	0 * SIZE(Y)
	addq	$SIZE, Y
	.align 8

.L19:
	xorq	%rax,%rax
	ret
	.align	8

.L20:
	movq	M,  %rax
	sarq	$4, %rax
	jle	.L23

	movlpd	0 * SIZE(X), %xmm0
	movhpd	1 * SIZE(X), %xmm0
	movlpd	2 * SIZE(X), %xmm1
	movhpd	3 * SIZE(X), %xmm1
	movlpd	4 * SIZE(X), %xmm2
	movhpd	5 * SIZE(X), %xmm2
	movlpd	6 * SIZE(X), %xmm3
	movhpd	7 * SIZE(X), %xmm3

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5
	movapd	4 * SIZE(Y), %xmm6
	movapd	6 * SIZE(Y), %xmm7

	movlpd	 8 * SIZE(X), %xmm8
	movhpd	 9 * SIZE(X), %xmm8
	mulpd	ALPHA, %xmm0
	movlpd	10 * SIZE(X), %xmm9
	movhpd	11 * SIZE(X), %xmm9
	mulpd	ALPHA, %xmm1
	movlpd	12 * SIZE(X), %xmm10
	movhpd	13 * SIZE(X), %xmm10
	mulpd	ALPHA, %xmm2
	movlpd	14 * SIZE(X), %xmm11
	movhpd	15 * SIZE(X), %xmm11
	mulpd	ALPHA, %xmm3

	decq	%rax
	jle .L22
	.align	8

.L21:
#ifdef OPTERON
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)
	PREFETCHW	(PREFETCHSIZE + 0) * SIZE(Y)
#endif
	addpd	%xmm4, %xmm0
	movapd	 8 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm8
	addpd	%xmm5, %xmm1
	movapd	10 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm9
	addpd	%xmm6, %xmm2
	movapd	12 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm10
	addpd	%xmm7, %xmm3
	movapd	14 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, 0 * SIZE(Y)
	movlpd	16 * SIZE(X), %xmm0
	movhpd	17 * SIZE(X), %xmm0
	movapd	%xmm1, 2 * SIZE(Y)
	movlpd	18 * SIZE(X), %xmm1
	movhpd	19 * SIZE(X), %xmm1
	movapd	%xmm2, 4 * SIZE(Y)
	movlpd	20 * SIZE(X), %xmm2
	movhpd	21 * SIZE(X), %xmm2
	movapd	%xmm3, 6 * SIZE(Y)
	movlpd	22 * SIZE(X), %xmm3
	movhpd	23 * SIZE(X), %xmm3

#ifdef OPTERON
	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)
	PREFETCHW	(PREFETCHSIZE + 8) * SIZE(Y)
#endif

	addpd	%xmm4, %xmm8
	movapd	16 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm0
	addpd	%xmm5, %xmm9
	movapd	18 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm1
	addpd	%xmm6, %xmm10
	movapd	20 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm2
	addpd	%xmm7, %xmm11
	movapd	22 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm3

	movapd	%xmm8,   8 * SIZE(Y)
	movlpd	24 * SIZE(X), %xmm8
	movhpd	25 * SIZE(X), %xmm8
	movapd	%xmm9,  10 * SIZE(Y)
	movlpd	26 * SIZE(X), %xmm9
	movhpd	27 * SIZE(X), %xmm9
	movapd	%xmm10, 12 * SIZE(Y)
	movlpd	28 * SIZE(X), %xmm10
	movhpd	29 * SIZE(X), %xmm10
	movapd	%xmm11, 14 * SIZE(Y)
	movlpd	30 * SIZE(X), %xmm11
	movhpd	31 * SIZE(X), %xmm11

	addq	$16 * SIZE, Y
	addq	$16 * SIZE, X
	decq	%rax
	jg	.L21
	.align 8

.L22:
	addpd	%xmm4, %xmm0
	movapd	 8 * SIZE(Y), %xmm4
	mulpd	ALPHA, %xmm8
	addpd	%xmm5, %xmm1
	movapd	10 * SIZE(Y), %xmm5
	mulpd	ALPHA, %xmm9
	addpd	%xmm6, %xmm2
	movapd	12 * SIZE(Y), %xmm6
	mulpd	ALPHA, %xmm10
	addpd	%xmm7, %xmm3
	movapd	14 * SIZE(Y), %xmm7
	mulpd	ALPHA, %xmm11

	movapd	%xmm0, 0 * SIZE(Y)
	movlpd	16 * SIZE(X), %xmm0
	movhpd	17 * SIZE(X), %xmm0
	movapd	%xmm1, 2 * SIZE(Y)
	movlpd	18 * SIZE(X), %xmm1
	movhpd	19 * SIZE(X), %xmm1
	movapd	%xmm2, 4 * SIZE(Y)
	movlpd	20 * SIZE(X), %xmm2
	movhpd	21 * SIZE(X), %xmm2
	movapd	%xmm3, 6 * SIZE(Y)
	movlpd	22 * SIZE(X), %xmm3
	movhpd	23 * SIZE(X), %xmm3

	addpd	%xmm4, %xmm8
	mulpd	ALPHA, %xmm0
	addpd	%xmm5, %xmm9
	mulpd	ALPHA, %xmm1
	addpd	%xmm6, %xmm10
	mulpd	ALPHA, %xmm2
	addpd	%xmm7, %xmm11
	mulpd	ALPHA, %xmm3

	movapd	%xmm8,   8 * SIZE(Y)
	movapd	%xmm9,  10 * SIZE(Y)
	movapd	%xmm10, 12 * SIZE(Y)
	movapd	%xmm11, 14 * SIZE(Y)

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	.align 8

.L23:
	movq	M,  %rax
	andq	$8, %rax
	jle	.L24
	.align	8

	movlpd	0 * SIZE(X), %xmm0
	movhpd	1 * SIZE(X), %xmm0
	movlpd	2 * SIZE(X), %xmm1
	movhpd	3 * SIZE(X), %xmm1
	movlpd	4 * SIZE(X), %xmm2
	movhpd	5 * SIZE(X), %xmm2
	movlpd	6 * SIZE(X), %xmm3
	movhpd	7 * SIZE(X), %xmm3

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5
	movapd	4 * SIZE(Y), %xmm6
	movapd	6 * SIZE(Y), %xmm7

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1
	mulpd	ALPHA, %xmm2
	mulpd	ALPHA, %xmm3

	addpd	%xmm4, %xmm0
	addpd	%xmm5, %xmm1
	addpd	%xmm6, %xmm2
	addpd	%xmm7, %xmm3

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	%xmm1, 2 * SIZE(Y)
	movapd	%xmm2, 4 * SIZE(Y)
	movapd	%xmm3, 6 * SIZE(Y)

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	.align 8

.L24:
	movq	M,  %rax
	andq	$4, %rax
	jle	.L25
	.align	8

	movlpd	0 * SIZE(X), %xmm0
	movhpd	1 * SIZE(X), %xmm0
	movlpd	2 * SIZE(X), %xmm1
	movhpd	3 * SIZE(X), %xmm1

	movapd	0 * SIZE(Y), %xmm4
	movapd	2 * SIZE(Y), %xmm5

	mulpd	ALPHA, %xmm0
	mulpd	ALPHA, %xmm1

	addpd	%xmm4, %xmm0
	addpd	%xmm5, %xmm1

	movapd	%xmm0, 0 * SIZE(Y)
	movapd	%xmm1, 2 * SIZE(Y)

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	.align 8

.L25:
	movq	M,  %rax
	andq	$2, %rax
	jle	.L26
	.align	8

	movlpd	0 * SIZE(X), %xmm0
	movhpd	1 * SIZE(X), %xmm0
	movapd	0 * SIZE(Y), %xmm4

	mulpd	ALPHA, %xmm0
	addpd	%xmm4, %xmm0

	movapd	%xmm0, 0 * SIZE(Y)

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	.align 8

.L26:
	movq	M,  %rax
	andq	$1, %rax
	jle	.L29
	.align	8

	movlpd	0 * SIZE(X), %xmm0
	mulsd	ALPHA, %xmm0
	addsd	0 * SIZE(Y), %xmm0

	movlpd	%xmm0, 	0 * SIZE(Y)
	addq	$SIZE, Y
	.align 8

.L29:
	xorq	%rax,%rax
	ret
	.align	8


.L40:
	movq	Y, YY
	movq	M,  %rax
	sarq	$3, %rax
	jle	.L45
	.align	8

.L41:
	movlpd	0 * SIZE(X), %xmm0
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm0
	addq	INCX, X
	mulpd	ALPHA, %xmm0

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm0

	movlpd	0 * SIZE(X), %xmm1
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm1
	addq	INCX, X
	mulpd	ALPHA, %xmm1

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm1

	movlpd	0 * SIZE(X), %xmm2
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulpd	ALPHA, %xmm2

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm2

	movlpd	0 * SIZE(X), %xmm3
	addq	INCX, X
	movhpd	0 * SIZE(X), %xmm3
	addq	INCX, X
	mulpd	ALPHA, %xmm3

	movlpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	movhpd	0 * SIZE(YY), %xmm6
	addq	INCY, YY
	addpd	%xmm6, %xmm3

	movlpd	%xmm0, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm0, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm1, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm1, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm2, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm2, 0 * SIZE(Y)
	addq	INCY, Y
	movlpd	%xmm3, 0 * SIZE(Y)
	addq	INCY, Y
	movhpd	%xmm3, 0 * SIZE(Y)
	addq	INCY, Y

	decq	%rax
	jg	.L41
	.align 8

.L45:
	movq	M,  %rax
	andq	$7, %rax
	jle	.L47
	.align	8

.L46:
	movlpd	(X), %xmm0
	addq	INCX, X
	mulsd	%xmm15, %xmm0
	addsd	(Y), %xmm0
	movlpd	%xmm0, (Y)
	addq	INCY, Y
	decq	%rax
	jg	.L46
	.align 8

.L47:
	xorq	%rax, %rax
	ret

	EPILOGUE
