/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define M	ARG1	/* rdi */
#define X	ARG4	/* rsi */
#define INCX	ARG5	/* rdx */
#define Y	ARG6	/* rcx */
#define INCY	ARG2	/* r8  */

#define	YY	%r10
#define ALPHA	%xmm15

#ifdef OPTERON
#define movsd	movlps
#endif

#undef HAVE_SSE3

#define PREFETCHSIZE 144

	PROLOGUE
	PROFCODE

	movq	8(%rsp), INCY

	movss	%xmm0, %xmm14		# 0    a
	movss	%xmm1, %xmm5		# 0    b

	pxor	%xmm15, %xmm15		#  0   0
	subss	%xmm5, %xmm15		#  0  -b

	unpcklps %xmm14, %xmm15		#  a   -b
	unpcklps %xmm5, %xmm14		#  b   a

	movlhps	 %xmm15, %xmm15
	movlhps	 %xmm14, %xmm14

	leaq	(, INCX, SIZE), INCX
	leaq	(, INCY, SIZE), INCY

	addq	INCX, INCX
	addq	INCY, INCY

	cmpq	$2 * SIZE, INCX
	jne	.L14
	cmpq	$2 * SIZE, INCY
	jne	.L14

	movq	M,  %rax
	sarq	$3, %rax
	jle	.L15
	.align	8

.L16:
#ifdef OPTERON
	prefetch	PREFETCHSIZE * SIZE(X)
	prefetchw	PREFETCHSIZE * SIZE(Y)
#endif

#ifndef HAVE_SSE3
	movss	 0 * SIZE(X), %xmm0		#  0  x1
	movss	 1 * SIZE(X), %xmm1		#  0  x2
	movss	 2 * SIZE(X), %xmm3		#  0  x1
	movss	 3 * SIZE(X), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1		# x2  x2
#else
	movsd	 0 * SIZE(X), %xmm0		#  0  x1
	movhps	 2 * SIZE(X), %xmm0		#  0  x2

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(Y), %xmm2
	movhps	 2 * SIZE(Y), %xmm2
	movsd	 4 * SIZE(Y), %xmm5
	movhps	 6 * SIZE(Y), %xmm5

	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(Y)
	movhps	%xmm2,  2 * SIZE(Y)

#ifndef HAVE_SSE3
	movss	 4 * SIZE(X), %xmm0		#  0  x1
	movss	 5 * SIZE(X), %xmm1		#  0  x2
	movss	 6 * SIZE(X), %xmm3		#  0  x1
	movss	 7 * SIZE(X), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2
#else
	movsd	 4 * SIZE(X), %xmm0		#  0  x1
	movhps	 6 * SIZE(X), %xmm0		#  0  x2

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	addps	%xmm0, %xmm5
	movsd	%xmm5,  4 * SIZE(Y)
	movhps	%xmm5,  6 * SIZE(Y)

#ifndef HAVE_SSE3
	movss	 8 * SIZE(X), %xmm0		#  0  x1
	movss	 9 * SIZE(X), %xmm1		#  0  x2
	movss	10 * SIZE(X), %xmm3		#  0  x1
	movss	11 * SIZE(X), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1		# x2  x2
#else
	movsd	 8 * SIZE(X), %xmm0		#  0  x1
	movhps	10 * SIZE(X), %xmm0		#  0  x2

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 8 * SIZE(Y), %xmm2
	movhps	10 * SIZE(Y), %xmm2
	movsd	12 * SIZE(Y), %xmm5
	movhps	14 * SIZE(Y), %xmm5

	addps	%xmm0, %xmm2
	movsd	%xmm2,  8 * SIZE(Y)
	movhps	%xmm2, 10 * SIZE(Y)

#ifndef HAVE_SSE3
	movss	12 * SIZE(X), %xmm0		#  0  x1
	movss	13 * SIZE(X), %xmm1		#  0  x2
	movss	14 * SIZE(X), %xmm3		#  0  x1
	movss	15 * SIZE(X), %xmm4		#  0  x2

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2
#else
	movsd	12 * SIZE(X), %xmm0		#  0  x1
	movhps	14 * SIZE(X), %xmm0		#  0  x2

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	addps	%xmm0, %xmm5
	movsd	%xmm5, 12 * SIZE(Y)
	movhps	%xmm5, 14 * SIZE(Y)

#ifdef PENTIUM4
	prefetcht0	PREFETCHSIZE * SIZE(Y)
#endif

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	decq	%rax
	jg	.L16
	.align 8

.L15:
	movq	M, %rax
	andq	$7,   %rax
	jle	.L40
	.align	8

.L22:
	movss	 0 * SIZE(X), %xmm0		#  0  x1
	movss	 1 * SIZE(X), %xmm1		#  0  x2

	unpcklps %xmm0, %xmm0			# x1  x1
	unpcklps %xmm1, %xmm1			# x2  x2
	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(Y), %xmm2
	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(Y)

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	decq	%rax
	jg	.L22
	jmp	.L40
	.align	8

.L14:
	movq	M, %rax
	sarq	$2,   %rax
	jle	.L28
	.align	8

.L29:
#ifndef HAVE_SSE3
	movss	 0 * SIZE(X), %xmm0		#  0  x1
	movss	 1 * SIZE(X), %xmm1		#  0  x2
	addq	INCX, X
	movss	 0 * SIZE(X), %xmm3		#  0  x1
	movss	 1 * SIZE(X), %xmm4		#  0  x2
	addq	INCX, X

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2
#else
	movsd	 0 * SIZE(X), %xmm0		#  0  x1
	addq	INCX, X
	movhps	 0 * SIZE(X), %xmm0		#  0  x2
	addq	INCX, X

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(Y), %xmm2
	movhps	 0 * SIZE(Y, INCY), %xmm2

	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(Y)
	addq	INCY, Y
	movhps	%xmm2,  0 * SIZE(Y)
	addq	INCY, Y

#ifndef HAVE_SSE3
	movss	 0 * SIZE(X), %xmm0		#  0  x1
	movss	 1 * SIZE(X), %xmm1		#  0  x2
	addq	INCX, X
	movss	 0 * SIZE(X), %xmm3		#  0  x1
	movss	 1 * SIZE(X), %xmm4		#  0  x2
	addq	INCX, X

	shufps	$0, %xmm3, %xmm0
	shufps	$0, %xmm4, %xmm1			# x2  x2
#else
	movsd	 0 * SIZE(X), %xmm0		#  0  x1
	addq	INCX, X
	movhps	 0 * SIZE(X), %xmm0		#  0  x2
	addq	INCX, X

	movshdup	 %xmm0, %xmm1		#  0  x1
	movsldup	 %xmm0, %xmm0		#  0  x1
#endif

	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0

	movsd	 0 * SIZE(Y), %xmm5
	movhps	 0 * SIZE(Y, INCY), %xmm5
	addps	%xmm0, %xmm5
	movsd	%xmm5,  0 * SIZE(Y)
	addq	INCY, Y
	movhps	%xmm5,  0 * SIZE(Y)
	addq	INCY, Y

	decq	%rax
	jg	.L29
	.align 8

.L28:
	movq	M, %rax
	andq	$3,   %rax
	jle	.L40
	.align	8

.L35:
	movss	 0 * SIZE(X), %xmm0		#  0  x1
	movss	 1 * SIZE(X), %xmm1		#  0  x2
	addq	INCX, X

	movsd	 0 * SIZE(Y), %xmm2

	unpcklps %xmm0, %xmm0			# x1  x1
	unpcklps %xmm1, %xmm1			# x2  x2
	mulps	%xmm14, %xmm0			#  a * x1  b * x1
	mulps	%xmm15, %xmm1			# -b * x2  a * x2
	addps	%xmm1, %xmm0
	addps	%xmm0, %xmm2
	movsd	%xmm2,  0 * SIZE(Y)
	addq	INCY, Y

	decq	%rax
	jg	.L35
	.align 8

.L40:
	xorq	%rax, %rax
	ret

	EPILOGUE

