/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	12
#define ARGS     0
	
#define STACK_N		 4 + STACK + ARGS(%esp)
#define STACK_X		 8 + STACK + ARGS(%esp)
#define STACK_INCX	12 + STACK + ARGS(%esp)
#define STACK_Y		16 + STACK + ARGS(%esp)
#define STACK_INCY	20 + STACK + ARGS(%esp)

#define N	%ecx
#define X	%esi
#define INCX	%ebx
#define Y	%edi
#define INCY	%edx


#ifdef PENTIUM4
#define PREFETCHSIZE 88
#endif

#ifdef OPTERON
#define PREFETCHSIZE 80
#define movsd	     movlps
#endif

#ifndef HAVE_SSE2
#define movsd	movlps
#endif
	
	PROLOGUE

	pushl	%edi
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	STACK_N,     N
	movl	STACK_X,     X
	movl	STACK_INCX,  INCX
	movl	STACK_Y,     Y
	movl	STACK_INCY,  INCY

#ifdef F_INTERFACE
	movl	(N),   N		# N
	movl	(INCX),INCX		# INCX
	movl	(INCY),INCY		# INCY
#endif

	leal	(, INCX, SIZE), INCX	
	leal	(, INCY, SIZE), INCY	

#ifdef HAVE_SSE2
	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
#else
	xorps	%xmm0, %xmm0
	xorps	%xmm1, %xmm1
#endif

	cmpl	$0, N
	jle	.L999

	cmpl	$SIZE, INCX
	jne	.L14
	cmpl	$SIZE, INCY
	jne	.L14

	cmpl	$7, N
	jle	.L15

	testl	$SIZE, Y
	je	.L00

	movss	0 * SIZE(X), %xmm0
	mulss	0 * SIZE(Y), %xmm0
	addl	$1 * SIZE, X
	addl	$1 * SIZE, Y
	decl	N
	.align 4

.L00:
	testl	$SIZE * 2, Y
	je	.L01

	movss	0 * SIZE(X), %xmm1
	mulss	0 * SIZE(Y), %xmm1
	movss	1 * SIZE(X), %xmm2
	mulss	1 * SIZE(Y), %xmm2
	addss	%xmm2, %xmm0

	addl	$2 * SIZE, X
	addl	$2 * SIZE, Y
	subl	$2, N
	.align 8

.L01:
	movl	N,  %eax
	sarl	$5, %eax
	jle	.L15
	.align	8

.L16:
#ifdef PENTIUM4
	prefetcht1	(PREFETCHSIZE +  0) * SIZE(X)
#endif

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(Y)
#endif

	movsd	 0 * SIZE(X), %xmm2
	movhps	 2 * SIZE(X), %xmm2

	mulps	 0 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	 4 * SIZE(X), %xmm2
	movhps	 6 * SIZE(X), %xmm2

	mulps	 4 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	movsd	 8 * SIZE(X), %xmm2
	movhps	10 * SIZE(X), %xmm2

	mulps	 8 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	12 * SIZE(X), %xmm2
	movhps	14 * SIZE(X), %xmm2

	mulps	12 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE + 16) * SIZE(X)
	prefetcht0	(PREFETCHSIZE + 16) * SIZE(Y)
#endif

	movsd	16 * SIZE(X), %xmm2
	movhps	18 * SIZE(X), %xmm2

	mulps	16 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	20 * SIZE(X), %xmm2
	movhps	22 * SIZE(X), %xmm2

	mulps	20 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	movsd	24 * SIZE(X), %xmm2
	movhps	26 * SIZE(X), %xmm2

	mulps	24 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	28 * SIZE(X), %xmm2
	movhps	30 * SIZE(X), %xmm2

	mulps	28 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	addl	$32 * SIZE, X
	addl	$32 * SIZE, Y

	decl	%eax
	jg	.L16
	.align 8

.L15:
	movl	N,   %eax
	andl	$31, %eax
	jle	.L999
	.align	8

.L22:
	movss	0 * SIZE(X), %xmm2
	mulss	0 * SIZE(Y), %xmm2
	addl	$SIZE, X
	addl	$SIZE, Y

	addss	%xmm2, %xmm0
	decl	%eax
	jg	.L22
	jmp	.L999
	.align	8

.L14:
#ifdef F_INTERFACE
	testl	INCX, INCX		# if (incx < 0)
	jge	.L28

	movl	N, %eax		# n
	decl	%eax			# n - 1
	imull	INCX, %eax		# (n - 1) * incx
	subl	%eax, X
	.align 8

.L28:
	testl	INCY, INCY		# if (incy < 0)
	jge	.L29

	movl	N, %eax
	decl	%eax			# (n - 1)
	imull	INCY, %eax		# (n - 1) * incy
	subl	%eax, Y
	.align 8
.L29:
#endif
	movl	N,  %eax
	sarl	$2, %eax
	jle	.L30
	.align	8

.L31:
	movss	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	addss	%xmm2, %xmm0

	movss	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	addss	%xmm2, %xmm1

	movss	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	addss	%xmm2, %xmm0

	movss	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	addss	%xmm2, %xmm1

	decl	%eax
	jg	.L31
	.align 8

.L30:
	movl	N, %eax
	andl	$3,   %eax
	jle	.L999
	.align	8

.L37:
	movss	0 * SIZE(X), %xmm2
	mulss	0 * SIZE(Y), %xmm2
	addl	INCX, X
	addl	INCY, Y
	addss	%xmm2, %xmm0
	decl	%eax
	jg	.L37
	.align 8

.L999:
	addps	%xmm1, %xmm0

	movhlps	%xmm0, %xmm1
	addps	%xmm1, %xmm0
	
	movaps	%xmm0, %xmm1
	shufps  $1, %xmm0, %xmm0
	addss	 %xmm1, %xmm0

	movss	%xmm0, STACK_N
	flds	STACK_N

	popl	%ebx
	popl	%esi
	popl	%edi
	ret

	EPILOGUE
