/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	12
#define ARGS     0
	
#define STACK_N		 4 + STACK + ARGS(%esp)
#define STACK_X		 8 + STACK + ARGS(%esp)
#define STACK_INCX	12 + STACK + ARGS(%esp)
#define STACK_Y		16 + STACK + ARGS(%esp)
#define STACK_INCY	20 + STACK + ARGS(%esp)

#define N	%ecx
#define X	%esi
#define INCX	%ebx
#define Y	%edi
#define INCY	%edx


#ifdef PENTIUM4
#define PREFETCHSIZE 88
#endif

#ifdef OPTERON
#define PREFETCHSIZE 80
#define movsd	     movlpd
#endif

	PROLOGUE

	pushl	%edi
	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	STACK_N,     N
	movl	STACK_X,     X
	movl	STACK_INCX,  INCX
	movl	STACK_Y,     Y
	movl	STACK_INCY,  INCY

#ifdef F_INTERFACE
	movl	(N),   N		# N
	movl	(INCX),INCX		# INCX
	movl	(INCY),INCY		# INCY
#endif

	leal	(, INCX, SIZE), INCX	
	leal	(, INCY, SIZE), INCY	

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1

	cmpl	$0, N
	jle	.L999

	cmpl	$SIZE, INCX
	jne	.L14
	cmpl	$SIZE, INCY
	jne	.L14

	testl	$SIZE, Y
	je	.L00

	movsd	0 * SIZE(X), %xmm0
	mulsd	0 * SIZE(Y), %xmm0
	addl	$1 * SIZE, X
	addl	$1 * SIZE, Y
	decl	N
	.align 4

.L00:
	movl	N,  %eax
	sarl	$4, %eax
	jle	.L15
	.align	8

.L16:
#ifdef PENTIUM4
	prefetcht1	(PREFETCHSIZE +  0) * SIZE(X)
#endif

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(Y)
#endif

	movsd	 0 * SIZE(X), %xmm2
	movhpd	 1 * SIZE(X), %xmm2

	mulpd	 0 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm0

	movsd	 2 * SIZE(X), %xmm2
	movhpd	 3 * SIZE(X), %xmm2

	mulpd	 2 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm1

	movsd	 4 * SIZE(X), %xmm2
	movhpd	 5 * SIZE(X), %xmm2

	mulpd	 4 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm0

	movsd	 6 * SIZE(X), %xmm2
	movhpd	 7 * SIZE(X), %xmm2

	mulpd	 6 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm1

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  8) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  8) * SIZE(Y)
#endif

	movsd	 8 * SIZE(X), %xmm2
	movhpd	 9 * SIZE(X), %xmm2

	mulpd	 8 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm0

	movsd	10 * SIZE(X), %xmm2
	movhpd	11 * SIZE(X), %xmm2

	mulpd	10 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm1

	movsd	12 * SIZE(X), %xmm2
	movhpd	13 * SIZE(X), %xmm2

	mulpd	12 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm0

	movsd	14 * SIZE(X), %xmm2
	movhpd	15 * SIZE(X), %xmm2

	mulpd	14 * SIZE(Y), %xmm2
	addpd	%xmm2, %xmm1

	addl	$16 * SIZE, X
	addl	$16 * SIZE, Y

	decl	%eax
	jg	.L16
	.align 8

.L15:
	movl	N,   %eax
	andl	$15, %eax
	jle	.L999
	.align	8

.L22:
	movsd	0 * SIZE(X), %xmm2
	mulsd	0 * SIZE(Y), %xmm2
	addl	$SIZE, X
	addl	$SIZE, Y

	addsd	%xmm2, %xmm0
	decl	%eax
	jg	.L22
	jmp	.L999
	.align	8

.L14:
#ifdef F_INTERFACE
	testl	INCX, INCX		# if (incx < 0)
	jge	.L28

	movl	N, %eax		# n
	decl	%eax			# n - 1
	imull	INCX, %eax		# (n - 1) * incx
	subl	%eax, X
	.align 8

.L28:
	testl	INCY, INCY		# if (incy < 0)
	jge	.L29

	movl	N, %eax
	decl	%eax			# (n - 1)
	imull	INCY, %eax		# (n - 1) * incy
	subl	%eax, Y
	.align 8
.L29:
#endif
	movl	N, %eax
	sarl	$2,   %eax
	jle	.L30
	.align	8

.L31:
	movsd	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulsd	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	movsd	0 * SIZE(X), %xmm3
	addl	INCX, X
	mulsd	0 * SIZE(Y), %xmm3
	addl	INCY, Y
	movsd	0 * SIZE(X), %xmm4
	addl	INCX, X
	mulsd	0 * SIZE(Y), %xmm4
	addl	INCY, Y
	movsd	0 * SIZE(X), %xmm5
	addl	INCX, X
	mulsd	0 * SIZE(Y), %xmm5
	addl	INCY, Y

	addsd	%xmm2, %xmm0
	addsd	%xmm3, %xmm1
	addsd	%xmm4, %xmm0
	addsd	%xmm5, %xmm1

	decl	%eax
	jg	.L31
	.align 8

.L30:
	movl	N, %eax
	andl	$3,   %eax
	jle	.L999
	.align	8

.L37:
	movsd	0 * SIZE(X), %xmm2
	addl	INCX, X
	mulsd	0 * SIZE(Y), %xmm2
	addl	INCY, Y
	addsd	%xmm2, %xmm0
	decl	%eax
	jg	.L37
	.align 8

.L999:
	addpd	%xmm1, %xmm0
	movapd	%xmm0, %xmm1
	unpckhpd	%xmm0, %xmm0
	addsd	%xmm1, %xmm0

	movsd	%xmm0, STACK_N
	fldl	STACK_N

	popl	%ebx
	popl	%esi
	popl	%edi
	ret

	EPILOGUE
