/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */

	PROLOGUE
	PROFCODE

#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	leaq	(, INCX, SIZE), INCX	
	fldz
	fldz
	leaq	(, INCY, SIZE), INCY	
	fldz
	fldz
	testq	N, N
	jle	.L27

	cmpq	$SIZE, INCX
	jne	.L14
	cmpq	$SIZE, INCY
	jne	.L14

	movq	N, %rax
	sarq	$3,   %rax
	jle	.L15
	.align	8

#define PREFETCHSIZE 96

.L16:
	FLD	0 * SIZE(X)
	FMUL	0 * SIZE(Y)
	faddp	%st,%st(1)
	FLD	1 * SIZE(X)
	FMUL	1 * SIZE(Y)
	faddp	%st,%st(2)
	FLD	2 * SIZE(X)
	FMUL	2 * SIZE(Y)
	faddp	%st,%st(3)
	FLD	3 * SIZE(X)
	FMUL	3 * SIZE(Y)
	faddp	%st,%st(4)

	FLD	4 * SIZE(X)
	FMUL	4 * SIZE(Y)
	faddp	%st,%st(1)
	FLD	5 * SIZE(X)
	FMUL	5 * SIZE(Y)
	faddp	%st,%st(2)
	FLD	6 * SIZE(X)
	FMUL	6 * SIZE(Y)
	faddp	%st,%st(3)
	FLD	7 * SIZE(X)
	FMUL	7 * SIZE(Y)
	faddp	%st,%st(4)

	prefetcht0 PREFETCHSIZE * SIZE(X)
	prefetcht0 PREFETCHSIZE * SIZE(Y)

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	decq	%rax
	jg	.L16
	.align 8

.L15:
	movq	N, %rax
	andq	$7,   %rax
	jle	.L27
	.align	8

.L22:
	FLD	(X)
	addq	$SIZE, X
	FMUL	(Y)
	addq	$SIZE, Y
	faddp	%st,%st(1)
	decq	%rax
	jg	.L22

	jmp	.L27
	.align	8

.L14:
#ifdef F_INTERFACE
	testq	INCX, INCX		# if (incx < 0)
	jge	.L28

	movq	N, %rax		# n
	decq	%rax			# n - 1
	imulq	INCX, %rax		# (n - 1) * incx
	subq	%rax, X
	.align 8

.L28:
	testq	INCY, INCY		# if (incy < 0)
	jge	.L29

	movq	N, %rax
	decq	%rax			# (n - 1)
	imulq	INCY, %rax		# (n - 1) * incy
	subq	%rax, Y
	.align 8
.L29:
#endif
	movq	N, %rax
	sarq	$2,   %rax
	jle	.L30
	.align	8

.L31:
	FLD	(X)
	addq	INCX, X
	FMUL	(Y)
	addq	INCY, Y
	faddp	%st,%st(1)

	FLD	(X)
	addq	INCX, X
	FMUL	(Y)
	addq	INCY, Y
	faddp	%st,%st(2)

	FLD	(X)
	addq	INCX, X
	FMUL	(Y)
	addq	INCY, Y
	faddp	%st,%st(3)

	FLD	(X)
	addq	INCX, X
	FMUL	(Y)
	addq	INCY, Y
	faddp	%st,%st(4)

	decq	%rax
	jg	.L31
	.align 8

.L30:
	movq	N, %rax
	andq	$3,   %rax
	jle	.L27
	.align	8

.L37:
	FLD	(X)
	addq	INCX, X
	FMUL	(Y)
	addq	INCY, Y
	faddp	%st, %st(1)
	decq	%rax
	jg	.L37
	.align 8

.L27:
	faddp	%st,%st(2)
	faddp	%st,%st(2)
	faddp	%st,%st(1)

	sub	$2 * SIZE, %rsp

#if defined(F_INTERFACE) && defined(NEED_F2CCONV)
	fstpl	(%rsp)
	movsd	(%rsp), %xmm0
#else
	FST	(%rsp)
	MOVSD	(%rsp), %xmm0
#endif

	add	$2 * SIZE, %rsp
	ret

	EPILOGUE

