/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */

#ifdef PENTIUM4
#define PREFETCHSIZE 88
#endif

#ifdef OPTERON
#define PREFETCHSIZE 80
#define movsd	     movlps
#endif

#ifndef HAVE_SSE2
#define movsd	movlps
#endif
	
	PROLOGUE
	PROFCODE

#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	leaq	(, INCX, SIZE), INCX	
	leaq	(, INCY, SIZE), INCY	

#ifdef HAVE_SSE2
	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
#else
	xorps	%xmm0, %xmm0
	xorps	%xmm1, %xmm1
#endif

	cmpq	$0, N
	jle	.L999

	cmpq	$SIZE, INCX
	jne	.L14
	cmpq	$SIZE, INCY
	jne	.L14

	cmpq	$7, N
	jle	.L15

	testq	$SIZE, Y
	je	.L00

	movss	0 * SIZE(X), %xmm0
	mulss	0 * SIZE(Y), %xmm0
	addq	$1 * SIZE, X
	addq	$1 * SIZE, Y
	decq	N
	.align 4

.L00:
	testq	$SIZE * 2, Y
	je	.L01

	movss	0 * SIZE(X), %xmm1
	mulss	0 * SIZE(Y), %xmm1
	movss	1 * SIZE(X), %xmm2
	mulss	1 * SIZE(Y), %xmm2
	addss	%xmm2, %xmm0

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	subq	$2, N
	.align 8

.L01:
	movq	N,  %rax
	sarq	$5, %rax
	jle	.L15
	.align	8

.L16:
#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(Y)
#endif

	movsd	 0 * SIZE(X), %xmm2
	movhps	 2 * SIZE(X), %xmm2

	mulps	 0 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	 4 * SIZE(X), %xmm2
	movhps	 6 * SIZE(X), %xmm2

	mulps	 4 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	movsd	 8 * SIZE(X), %xmm2
	movhps	10 * SIZE(X), %xmm2

	mulps	 8 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	12 * SIZE(X), %xmm2
	movhps	14 * SIZE(X), %xmm2

	mulps	12 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE + 16) * SIZE(X)
	prefetcht0	(PREFETCHSIZE + 16) * SIZE(Y)
#endif

	movsd	16 * SIZE(X), %xmm2
	movhps	18 * SIZE(X), %xmm2

	mulps	16 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	20 * SIZE(X), %xmm2
	movhps	22 * SIZE(X), %xmm2

	mulps	20 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	movsd	24 * SIZE(X), %xmm2
	movhps	26 * SIZE(X), %xmm2

	mulps	24 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm0

	movsd	28 * SIZE(X), %xmm2
	movhps	30 * SIZE(X), %xmm2

#ifdef PENTIUM4
	prefetcht1	(PREFETCHSIZE +  0) * SIZE(X)
#endif

	mulps	28 * SIZE(Y), %xmm2
	addps	%xmm2, %xmm1

	addq	$32 * SIZE, X
	addq	$32 * SIZE, Y

	decq	%rax
	jg	.L16
	.align 8

.L15:
	movq	N,   %rax
	andq	$31, %rax
	jle	.L999
	.align	8

.L22:
	movss	0 * SIZE(X), %xmm2
	mulss	0 * SIZE(Y), %xmm2
	addq	$SIZE, X
	addq	$SIZE, Y

	addss	%xmm2, %xmm0
	decq	%rax
	jg	.L22
	jmp	.L999
	.align	8

.L14:
#ifdef F_INTERFACE
	testq	INCX, INCX		# if (incx < 0)
	jge	.L28

	movq	N, %rax		# n
	decq	%rax			# n - 1
	imulq	INCX, %rax		# (n - 1) * incx
	subq	%rax, X
	.align 8

.L28:
	testq	INCY, INCY		# if (incy < 0)
	jge	.L29

	movq	N, %rax
	decq	%rax			# (n - 1)
	imulq	INCY, %rax		# (n - 1) * incy
	subq	%rax, Y
	.align 8
.L29:
#endif
	movq	N,  %rax
	sarq	$2, %rax
	jle	.L30
	.align	8

.L31:
	movss	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addq	INCY, Y
	addss	%xmm2, %xmm0

	movss	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addq	INCY, Y
	addss	%xmm2, %xmm1

	movss	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addq	INCY, Y
	addss	%xmm2, %xmm0

	movss	0 * SIZE(X), %xmm2
	addq	INCX, X
	mulss	0 * SIZE(Y), %xmm2
	addq	INCY, Y
	addss	%xmm2, %xmm1

	decq	%rax
	jg	.L31
	.align 8

.L30:
	movq	N, %rax
	andq	$3,   %rax
	jle	.L999
	.align	8

.L37:
	movss	0 * SIZE(X), %xmm2
	mulss	0 * SIZE(Y), %xmm2
	addq	INCX, X
	addq	INCY, Y
	addss	%xmm2, %xmm0
	decq	%rax
	jg	.L37
	.align 8

.L999:
	addps	%xmm1, %xmm0

#ifndef HAVE_SSE3
	movhlps	%xmm0, %xmm1
	addps	%xmm1, %xmm0
	
	movaps	%xmm0, %xmm1
	shufps  $1, %xmm0, %xmm0
	addss	 %xmm1, %xmm0
#else
	haddps	%xmm0, %xmm0
	haddps	%xmm0, %xmm0
#endif

#if !defined(DOUBLE) && defined(F_INTERFACE) && defined(NEED_F2CCONV)
	cvtss2sd	%xmm0, %xmm0
#endif
	ret

	EPILOGUE
