/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */

#ifdef OPTERON
#define PREFETCH	prefetch
#define PREFETCHSIZE	64
#endif

	PROLOGUE
	PROFCODE

#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	leaq	(, INCX, SIZE), INCX	
	leaq	(, INCY, SIZE), INCY	

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

	cmpq	$0, N
	jle	.L999

	cmpq	$SIZE, INCX
	jne	.L50
	cmpq	$SIZE, INCY
	jne	.L50

	testq	$SIZE, Y
	je	.L10

	movsd	0 * SIZE(X), %xmm0
	mulsd	0 * SIZE(Y), %xmm0
	addq	$1 * SIZE, X
	addq	$1 * SIZE, Y
	decq	N
	.align 4

.L10:
	movq	N,  %rax
	sarq	$4, %rax
	jle	.L24

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5
	movlpd	 4 * SIZE(X), %xmm6
	movhpd	 5 * SIZE(X), %xmm6
	movlpd	 6 * SIZE(X), %xmm7
	movhpd	 7 * SIZE(X), %xmm7

	movapd	 0 * SIZE(Y), %xmm8
	movapd	 2 * SIZE(Y), %xmm9
	movapd	 4 * SIZE(Y), %xmm10
	movapd	 6 * SIZE(Y), %xmm11

	movlpd	 8 * SIZE(X), %xmm12
	movhpd	 9 * SIZE(X), %xmm12
	movlpd	10 * SIZE(X), %xmm13
	movhpd	11 * SIZE(X), %xmm13
	movlpd	12 * SIZE(X), %xmm14
	movhpd	13 * SIZE(X), %xmm14
	movlpd	14 * SIZE(X), %xmm15
	movhpd	15 * SIZE(X), %xmm15

	mulpd	%xmm8,  %xmm4
	movapd	 8 * SIZE(Y), %xmm8
	mulpd	%xmm9,  %xmm5
	movapd	10 * SIZE(Y), %xmm9
	mulpd	%xmm10, %xmm6
	movapd	12 * SIZE(Y), %xmm10
	mulpd	%xmm11, %xmm7
	movapd	14 * SIZE(Y), %xmm11
	decq	%rax
	jle	.L22

	.align	8

.L21:
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(X)
	addpd	%xmm4,  %xmm0
	movlpd	16 * SIZE(X), %xmm4
	movhpd	17 * SIZE(X), %xmm4
	mulpd	%xmm8,  %xmm12
	PREFETCH	(PREFETCHSIZE + 0) * SIZE(Y)
	addpd	%xmm5,  %xmm1
	movlpd	18 * SIZE(X), %xmm5
	movhpd	19 * SIZE(X), %xmm5
	mulpd	%xmm9,  %xmm13

	addpd	%xmm6,  %xmm2
	movlpd	20 * SIZE(X), %xmm6
	movhpd	21 * SIZE(X), %xmm6
	mulpd	%xmm10, %xmm14
	addpd	%xmm7,  %xmm3
	movlpd	22 * SIZE(X), %xmm7
	movhpd	23 * SIZE(X), %xmm7
	mulpd	%xmm11, %xmm15

	movapd	16 * SIZE(Y), %xmm8
	movapd	18 * SIZE(Y), %xmm9
	movapd	20 * SIZE(Y), %xmm10
	movapd	22 * SIZE(Y), %xmm11

	PREFETCH	(PREFETCHSIZE + 8) * SIZE(X)
	addpd	%xmm12,  %xmm0
	movlpd	24 * SIZE(X), %xmm12
	movhpd	25 * SIZE(X), %xmm12
	mulpd	%xmm8,  %xmm4
	PREFETCH	(PREFETCHSIZE + 8) * SIZE(Y)
	addpd	%xmm13,  %xmm1
	movlpd	26 * SIZE(X), %xmm13
	movhpd	27 * SIZE(X), %xmm13
	mulpd	%xmm9,  %xmm5

	addpd	%xmm14,  %xmm2
	movlpd	28 * SIZE(X), %xmm14
	movhpd	29 * SIZE(X), %xmm14
	mulpd	%xmm10, %xmm6
	addpd	%xmm15,  %xmm3
	movlpd	30 * SIZE(X), %xmm15
	movhpd	31 * SIZE(X), %xmm15
	mulpd	%xmm11, %xmm7

	movapd	24 * SIZE(Y), %xmm8
	movapd	26 * SIZE(Y), %xmm9
	movapd	28 * SIZE(Y), %xmm10
	movapd	30 * SIZE(Y), %xmm11

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y

	decq	%rax
	jg	.L21
	.align 8

.L22:
	addpd	%xmm4,  %xmm0
	mulpd	%xmm8,  %xmm12
	addpd	%xmm5,  %xmm1
	mulpd	%xmm9,  %xmm13

	addpd	%xmm6,  %xmm2
	mulpd	%xmm10, %xmm14
	addpd	%xmm7,  %xmm3
	mulpd	%xmm11, %xmm15

	addpd	%xmm12,  %xmm0
	addpd	%xmm13,  %xmm1
	addpd	%xmm14,  %xmm2
	addpd	%xmm15,  %xmm3

	addq	$16 * SIZE, X
	addq	$16 * SIZE, Y
	.align 8

.L24:
	testq	$15, %rax
	jne	.L999

	movq	N,  %rax
	andq	$8, %rax
	jle	.L25

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5
	movlpd	 4 * SIZE(X), %xmm6
	movhpd	 5 * SIZE(X), %xmm6
	movlpd	 6 * SIZE(X), %xmm7
	movhpd	 7 * SIZE(X), %xmm7

	movapd	 0 * SIZE(Y), %xmm8
	movapd	 2 * SIZE(Y), %xmm9
	movapd	 4 * SIZE(Y), %xmm10
	movapd	 6 * SIZE(Y), %xmm11

	mulpd	%xmm8,  %xmm4
	mulpd	%xmm9,  %xmm5
	mulpd	%xmm10, %xmm6
	mulpd	%xmm11, %xmm7

	addpd	%xmm4,  %xmm0
	addpd	%xmm5,  %xmm1
	addpd	%xmm6,  %xmm2
	addpd	%xmm7,  %xmm3

	addq	$8 * SIZE, X
	addq	$8 * SIZE, Y
	.align 8

.L25:
	movq	N,  %rax
	andq	$4, %rax
	jle	.L26

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movlpd	 2 * SIZE(X), %xmm5
	movhpd	 3 * SIZE(X), %xmm5
	movapd	 0 * SIZE(Y), %xmm8
	movapd	 2 * SIZE(Y), %xmm9

	mulpd	%xmm8,  %xmm4
	mulpd	%xmm9,  %xmm5

	addpd	%xmm4,  %xmm0
	addpd	%xmm5,  %xmm1

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	.align 8

.L26:
	movq	N,  %rax
	andq	$2, %rax
	jle	.L27

	movlpd	 0 * SIZE(X), %xmm4
	movhpd	 1 * SIZE(X), %xmm4
	movapd	 0 * SIZE(Y), %xmm8

	mulpd	%xmm8,  %xmm4
	addpd	%xmm4,  %xmm0

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	.align 8

.L27:
	movq	N,  %rax
	andq	$1, %rax
	jle	.L999

	movsd	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm8

	mulsd	%xmm8,  %xmm4
	addsd	%xmm4,  %xmm0
	jmp	.L999
	.align 8


.L50:
#ifdef F_INTERFACE
	testq	INCX, INCX
	jge	.L51

	movq	N, %rax
	decq	%rax
	imulq	INCX, %rax
	subq	%rax, X
	.align 8

.L51:
	testq	INCY, INCY
	jge	.L52

	movq	N, %rax
	decq	%rax
	imulq	INCY, %rax
	subq	%rax, Y
	.align 8
.L52:
#endif

	movq	N,  %rax
	sarq	$2, %rax
	jle	.L55
	.align	8

.L53:
	movsd	0 * SIZE(X), %xmm4
	addq	INCX, X
	mulsd	0 * SIZE(Y), %xmm4
	addq	INCY, Y
	movsd	0 * SIZE(X), %xmm5
	addq	INCX, X
	mulsd	0 * SIZE(Y), %xmm5
	addq	INCY, Y
	movsd	0 * SIZE(X), %xmm6
	addq	INCX, X
	mulsd	0 * SIZE(Y), %xmm6
	addq	INCY, Y
	movsd	0 * SIZE(X), %xmm7
	addq	INCX, X
	mulsd	0 * SIZE(Y), %xmm7
	addq	INCY, Y

	addsd	%xmm4, %xmm0
	addsd	%xmm5, %xmm1
	addsd	%xmm6, %xmm2
	addsd	%xmm7, %xmm3

	decq	%rax
	jg	.L53
	.align 8

.L55:
	movq	N, %rax
	andq	$3,   %rax
	jle	.L999
	.align	8

.L56:
	movsd	0 * SIZE(X), %xmm4
	addq	INCX, X
	mulsd	0 * SIZE(Y), %xmm4
	addq	INCY, Y
	addsd	%xmm4, %xmm0
	decq	%rax
	jg	.L56
	.align 8

.L999:
	addpd	%xmm1, %xmm0
	addpd	%xmm3, %xmm2
	addpd	%xmm2, %xmm0

#ifndef HAVE_SSE3
	movapd	%xmm0, %xmm1
	unpckhpd	%xmm0, %xmm0
	addsd	%xmm1, %xmm0
#else
	haddpd	%xmm0, %xmm0
#endif

	ret

	EPILOGUE
