/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
#define RESULT	ARG1	/* rdi */
#define N	ARG2	/* rsi */
#define X	ARG3	/* rdx */
#define INCX	ARG4	/* rcx */
#define Y	ARG5	/* r8  */
#define INCY	ARG6	/* r9  */
#else
#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */
#endif

	PROLOGUE
	PROFCODE
	
#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	testq	N, N
	je	.L89

	addq	INCX, INCX
	fldz
	addq	INCY, INCY
	fldz

	leaq	(, INCX, SIZE), INCX	
	fldz
	leaq	(, INCY, SIZE), INCY	
	fldz

	cmpq	$2 * SIZE, INCX
	jne	.L14
	cmpq	$2 * SIZE, INCY
	jne	.L14

	movq	N, %rax
	sarq	$1,   %rax
	jle	.L15
	.align	8

.L16:
	FLD	 0 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 1 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1
	FLD	 2 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 2 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 3 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 3 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 2 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 3 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1

	addq	$4 * SIZE, X
	addq	$4 * SIZE, Y
	decq	%rax
	jg	.L16
	.align 8

.L15:
	movq	N, %rax
	andq	$1,   %rax
	jle	.L27
	.align	8

.L22:
	FLD	 0 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 1 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1
	jmp	.L27
	.align	8

.L14:
#ifdef F_INTERFACE
	testq	INCX, INCX		# if (incx < 0)
	jge	.L28

	movq	N, %rax		# n
	decq	%rax			# n - 1
	imulq	INCX, %rax		# (n - 1) * incx
	subq	%rax, X
	.align 8

.L28:
	testq	INCY, INCY		# if (incy < 0)
	jge	.L29

	movq	N, %rax
	decq	%rax			# (n - 1)
	imulq	INCY, %rax		# (n - 1) * incy
	subq	%rax, Y
	.align 8

.L29:
#endif

	movq	N, %rax
	sarq	$1,   %rax
	jle	.L30
	.align	8


.L31:
	FLD	 0 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 1 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1
	addq	INCX, X

	FLD	 0 * SIZE(X)		# bt1  = *(b_offset + 1)
	addq	INCY, Y

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 1 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1
	addq	INCX, X
	addq	INCY, Y

	decq	%rax
	jg	.L31
	.align 8

.L30:
	movq	N, %rax
	andq	$1,   %rax
	jle	.L27
	.align	8

.L37:
	FLD	 0 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(2)		# ct1 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(2)		# ct2 += bt1
	FLD	 1 * SIZE(X)		# bt1  = *(b_offset + 1)

	FLD	 0 * SIZE(Y)		# at1  = *(a_offset + 0)
	fmul	%st(1)			# at1 *= bt1
	faddp	%st, %st(4)		# ct3 += at1

	FMUL	 1 * SIZE(Y)		# bt1 *= *(a_offset + 1)
	faddp	%st, %st(4)		# ct4 += bt1
	.align 8

.L27:
#ifndef CONJ
	fsubp	%st, %st(3)
	faddp	%st, %st(1)
#else
	faddp	%st, %st(3)
	fsubp	%st, %st(1)
#endif

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
	FST	1 * SIZE(RESULT)
	FST	0 * SIZE(RESULT)
#else
	subq	$2 * SIZE, %rsp
	FST	1 * SIZE(%rsp)
	FST	0 * SIZE(%rsp)
#if !defined(DOUBLE) && defined(AMD_ABI)
	movsd	0 * SIZE(%rsp), %xmm0
#else
	MOVSD	0 * SIZE(%rsp), %xmm0
	MOVSD	1 * SIZE(%rsp), %xmm1
#endif
	addq	$2 * SIZE, %rsp
#endif
	ret
	.align 4

.L89:
	xorps	%xmm0, %xmm0
	xorps	%xmm1, %xmm1

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
	MOVSD	%xmm0, 0 * SIZE(RESULT)
	MOVSD	%xmm1, 1 * SIZE(RESULT)
#endif
	ret

	EPILOGUE
