/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
#define RESULT	ARG1	/* rdi */
#define N	ARG2	/* rsi */
#define X	ARG3	/* rdx */
#define INCX	ARG4	/* rcx */
#define Y	ARG5	/* r8  */
#define INCY	ARG6	/* r9  */
#else
#define N	ARG1	/* rdi */
#define X	ARG2	/* rsi */
#define INCX	ARG3	/* rdx */
#define Y	ARG4	/* rcx */
#define INCY	ARG5	/* r8  */
#endif

#ifdef PENTIUM4
#define PREFETCHSIZE 88
#endif

#ifdef OPTERON
#define PREFETCHSIZE 80
#define movsd	     movlpd
#endif

	PROLOGUE
	PROFCODE

#ifdef F_INTERFACE
#ifndef USE64BITINT
	movslq	(N), N			# N
	movslq	(INCX),INCX		# INCX
	movslq	(INCY),INCY		# INCY
#else
	movq	(N), N			# N
	movq	(INCX),INCX		# INCX
	movq	(INCY),INCY		# INCY
#endif
#endif

	addq	INCX, INCX
	addq	INCY, INCY

	leaq	(, INCX, SIZE), INCX	
	leaq	(, INCY, SIZE), INCY	

	pxor	%xmm0, %xmm0
	pxor	%xmm1, %xmm1
	pxor	%xmm2, %xmm2
	pxor	%xmm3, %xmm3

	cmpq	$0, N
	jle	.L999

	cmpq	$2 * SIZE, INCX
	jne	.L50
	cmpq	$2 * SIZE, INCY
	jne	.L50

	movq	N,  %rax
	sarq	$4, %rax
	jle	.L15
	.align	8

.L11:
#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  0) * SIZE(Y)
#endif

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	 2 * SIZE(X), %xmm4
	movsd	 2 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	 4 * SIZE(X), %xmm4
	movsd	 4 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	 6 * SIZE(X), %xmm4
	movsd	 6 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	 8 * SIZE(X), %xmm4
	movsd	 8 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	10 * SIZE(X), %xmm4
	movsd	10 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	12 * SIZE(X), %xmm4
	movsd	12 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	14 * SIZE(X), %xmm4
	movsd	14 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

#ifdef OPTERON
	prefetcht0	(PREFETCHSIZE +  8) * SIZE(X)
	prefetcht0	(PREFETCHSIZE +  8) * SIZE(Y)
#endif

	movddup	16 * SIZE(X), %xmm4
	movsd	16 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	18 * SIZE(X), %xmm4
	movsd	18 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	20 * SIZE(X), %xmm4
	movsd	20 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	22 * SIZE(X), %xmm4
	movsd	22 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	24 * SIZE(X), %xmm4
	movsd	24 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	movddup	26 * SIZE(X), %xmm4
	movsd	26 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	movddup	28 * SIZE(X), %xmm4
	movsd	28 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

#ifdef PENTIUM4
	prefetcht1	(PREFETCHSIZE +  0) * SIZE(X)
#endif

	movddup	30 * SIZE(X), %xmm4
	movsd	30 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1

	addq	$32 * SIZE, X
	addq	$32 * SIZE, Y

	decq	%rax
	jg	.L11
	.align 8

.L15:
	movq	N,  %rax
	andq	$15, %rax
	jle	.L999
	.align	8

.L16:
	movddup	 0 * SIZE(X), %xmm4	/* a b a b */
	movsd	 0 * SIZE(Y), %xmm5	/* c d x x */

	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	addq	$2 * SIZE, X
	addq	$2 * SIZE, Y
	decq	%rax
	jg	.L16
	jmp	.L999
	.align	8

.L50:
#ifdef F_INTERFACE
	testq	INCX, INCX		# if (incx < 0)
	jge	.L51

	movq	N, %rax		# n
	decq	%rax			# n - 1
	imulq	INCX, %rax		# (n - 1) * incx
	subq	%rax, X
	.align 8

.L51:
	testq	INCY, INCY		# if (incy < 0)
	jge	.L60

	movq	N, %rax
	decq	%rax			# (n - 1)
	imulq	INCY, %rax		# (n - 1) * incy
	subq	%rax, Y
	.align 8
.L60:
#endif
	movq	N,  %rax
	sarq	$3, %rax
	jle	.L65
	.align	8

.L62:
	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0
	addq	INCX, X
	addq	INCY, Y

	movddup	 0 * SIZE(X), %xmm4
	movsd	 0 * SIZE(Y), %xmm5
	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm1
	addq	INCX, X
	addq	INCY, Y

	decq	%rax
	jg	.L62
	.align 8

.L65:
	movq	N, %rax
	andq	$7,   %rax
	jle	.L999
	.align	8

.L66:
	movddup	 0 * SIZE(X), %xmm4	/* a b a b */
	movsd	 0 * SIZE(Y), %xmm5	/* c d x x */

	shufps	 $0x50, %xmm5, %xmm5
	mulps	 %xmm4, %xmm5
	addps	 %xmm5, %xmm0

	addq	INCX, X
	addq	INCY, Y
	decq	%rax
	jg	.L66
	.align 8

.L999:
	addps	%xmm1, %xmm0

	movhlps	%xmm0, %xmm1	/* a b c d */

	movaps	%xmm0,  %xmm2
	shufps	$1, %xmm2, %xmm2

	movaps	%xmm1, %xmm3
	shufps	$1, %xmm3, %xmm3

#ifndef CONJ
	subss	 %xmm3, %xmm0
	addss	 %xmm2, %xmm1
#else
	addss	 %xmm3, %xmm0
	subss	 %xmm2, %xmm1
#endif

#if defined(F_INTERFACE) && defined(F_INTERFACE_F2C)
	movss	%xmm0, 0 * SIZE(RESULT)
	movss	%xmm1, 1 * SIZE(RESULT)
#elif !defined(DOUBLE) && defined(AMD_ABI)
	unpcklps       %xmm1, %xmm0	
#endif
	ret

	EPILOGUE
