/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define N	%i0
#if defined(DOUBLE) && !defined(__64BIT__)
#define X	%i5
#define INCX	%i1
#else
#define X	%i4
#define INCX	%i5
#endif

#define I	%i2
#define XX	%i3

#ifdef DOUBLE
#define c1	%f0
#define c2	%f2
#define c3	%f4
#define c4	%f6
#define c5	%f8
#define c6	%f10
#define c7	%f12
#define c8	%f14

#define FZERO	%f60
#define ALPHA	%f62
#else
#define c1	%f0
#define c2	%f1
#define c3	%f2
#define c4	%f3
#define c5	%f4
#define c6	%f5
#define c7	%f6
#define c8	%f7

#define FZERO	%f29
#define ALPHA	%f30
#endif

	PROLOGUE
	SAVESP

#ifndef __64BIT__
#ifdef DOUBLE
	st	%i3, [%fp + STACK_START +  8]   /* ALPHA */
	st	%i4, [%fp + STACK_START + 12]
	st	%g0, [%fp + STACK_START + 16]
	st	%g0, [%fp + STACK_START + 20]

	ld	[%fp+  STACK_START + 28], INCX
#else
	st	%i3, [%fp + STACK_START +  8]   /* ALPHA */
	st	%g0, [%fp + STACK_START + 16]
#endif

	LDF	[%fp + STACK_START +  8], ALPHA
	LDF	[%fp + STACK_START + 16], FZERO
#else
#ifdef DOUBLE
	stx	%g0, [%fp + STACK_START + 32]
	FMOV	%f6, ALPHA
#else
	st	%g0, [%fp + STACK_START + 32]
	FMOV	%f7, ALPHA
#endif
	LDF	[%fp + STACK_START + 32], FZERO
#endif
	
	FCMP	ALPHA, FZERO
	fbne	.LL100
	sll	INCX, BASE_SHIFT, INCX

	cmp	INCX, SIZE
	bne	.LL50
	nop

	sra	N, 3, I
	cmp	I, 0
	ble,pn	%icc, .LL15
	nop

.LL11:
	STF	FZERO, [X +  0 * SIZE]
	add	I, -1, I
	STF	FZERO, [X +  1 * SIZE]
	cmp	I, 0
	STF	FZERO, [X +  2 * SIZE]
	STF	FZERO, [X +  3 * SIZE]
	STF	FZERO, [X +  4 * SIZE]
	STF	FZERO, [X +  5 * SIZE]
	add	X, 8 * SIZE, X
	STF	FZERO, [X -  2 * SIZE]
	bg,pt	%icc, .LL11
	STF	FZERO, [X -  1 * SIZE]

.LL15:
	and	N, 7, I
	cmp	I,  0
	ble,a,pn %icc, .LL19
	nop

.LL16:
	STF	FZERO, [X +  0 * SIZE]
	add	I, -1, I
	cmp	I, 0
	bg,pt	%icc, .LL16
	add	X, 1 * SIZE, X

.LL19:
	return	%i7 + 8
	clr	%o0

.LL50:
	sra	N, 3, I
	cmp	I, 0
	ble,pn	%icc, .LL55
	nop

.LL51:
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	add	I, -1, I
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	cmp	I, 0
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	STF	FZERO, [X +  0 * SIZE]
	add	X, INCX, X
	STF	FZERO, [X +  0 * SIZE]
	bg,pt	%icc, .LL51
	add	X, INCX, X

.LL55:
	and	N, 7, I
	cmp	I,  0
	ble,a,pn %icc, .LL59
	nop

.LL56:
	STF	FZERO, [X +  0 * SIZE]
	add	I, -1, I
	cmp	I, 0
	bg,pt	%icc, .LL56
	add	X, INCX, X

.LL59:
	return	%i7 + 8
	clr	%o0

.LL100:
	cmp	INCX, SIZE
	bne	.LL150
	sra	N, 3, I

	cmp	I, 0
	ble,pn	%icc, .LL115
	nop

#define PREFETCHSIZE 32

.LL111:
	prefetch [X  + PREFETCHSIZE * SIZE], 0

	LDF	[X +  0 * SIZE], c1
	LDF	[X +  1 * SIZE], c2
	LDF	[X +  2 * SIZE], c3
	LDF	[X +  3 * SIZE], c4
	LDF	[X +  4 * SIZE], c5
	FMUL	ALPHA, c1, c1
	LDF	[X +  5 * SIZE], c6
	FMUL	ALPHA, c2, c2
	LDF	[X +  6 * SIZE], c7
	FMUL	ALPHA, c3, c3
	LDF	[X +  7 * SIZE], c8
	FMUL	ALPHA, c4, c4

	STF	c1, [X +  0 * SIZE]
	FMUL	ALPHA, c5, c5
	add	I, -1, I
	STF	c2, [X +  1 * SIZE]
	FMUL	ALPHA, c6, c6
	cmp	I, 0
	STF	c3, [X +  2 * SIZE]
	FMUL	ALPHA, c7, c7
	STF	c4, [X +  3 * SIZE]
	FMUL	ALPHA, c8, c8
	STF	c5, [X +  4 * SIZE]
	STF	c6, [X +  5 * SIZE]
	add	X, 8 * SIZE, X
	STF	c7, [X -  2 * SIZE]
	bg,pt	%icc, .LL111
	STF	c8, [X -  1 * SIZE]

.LL115:
	and	N, 7, I
	cmp	I,  0
	ble,a,pn %icc, .LL119
	nop

.LL116:
	LDF	[X +  0 * SIZE], c1
	add	I, -1, I
	FMUL	ALPHA, c1, c1
	cmp	I, 0
	STF	c1, [X +  0 * SIZE]
	bg,pt	%icc, .LL116
	add	X, 1 * SIZE, X

.LL119:
	return	%i7 + 8
	clr	%o0

.LL150:
	sra	N, 3, I
	cmp	I, 0
	ble,pn	%icc, .LL155
	mov	X, XX

.LL151:
	LDF	[X +  0 * SIZE], c1
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c2
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c3
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c4
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c5
	FMUL	ALPHA, c1, c1
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c6
	FMUL	ALPHA, c2, c2
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c7
	FMUL	ALPHA, c3, c3
	add	X, INCX, X
	LDF	[X +  0 * SIZE], c8
	FMUL	ALPHA, c4, c4
	add	X, INCX, X

	STF	c1, [XX +  0 * SIZE]
	FMUL	ALPHA, c5, c5
	add	XX, INCX, XX
	STF	c2, [XX +  0 * SIZE]
	FMUL	ALPHA, c6, c6
	add	XX, INCX, XX
	STF	c3, [XX +  0 * SIZE]
	FMUL	ALPHA, c7, c7
	add	XX, INCX, XX
	STF	c4, [XX +  0 * SIZE]
	FMUL	ALPHA, c8, c8
	add	XX, INCX, XX
	STF	c5, [XX +  0 * SIZE]
	add	XX, INCX, XX
	add	I, -1, I
	STF	c6, [XX +  0 * SIZE]
	add	XX, INCX, XX
	cmp	I, 0
	STF	c7, [XX +  0 * SIZE]
	add	XX, INCX, XX
	STF	c8, [XX +  0 * SIZE]

	bg,pt	%icc, .LL151
	add	XX, INCX, XX

.LL155:
	and	N, 7, I
	cmp	I,  0
	ble,a,pn %icc, .LL159
	nop

.LL156:
	LDF	[X +  0 * SIZE], c1
	add	I, -1, I
	FMUL	ALPHA, c1, c1
	cmp	I, 0
	STF	c1, [X +  0 * SIZE]
	bg,pt	%icc, .LL156
	add	X, INCX, X

.LL159:
	return	%i7 + 8
	clr	%o0


	EPILOGUE
