/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define STACK	 8
	
#define N	 4 + STACK(%esp)
#ifdef DOUBLE
#define ALPHA_R	16 + STACK(%esp)
#define ALPHA_I	24 + STACK(%esp)
#define X	32 + STACK(%esp)
#define INCX	36 + STACK(%esp)
#else
#define ALPHA_R	16 + STACK(%esp)
#define ALPHA_I	20 + STACK(%esp)
#define X	24 + STACK(%esp)
#define INCX	28 + STACK(%esp)
#endif

	PROLOGUE

	pushl	%esi
	pushl	%ebx

	PROFCODE

	movl	N, %esi
	movl	X, %edx
	movl	INCX, %ebx

	fldz
	FLD	ALPHA_R
	FLD	ALPHA_I

	testl	%esi,%esi
	jle	.L57

#ifdef DOUBLE
	sall	$4, %ebx	# incx *= 8
#else
	sall	$3, %ebx	# incx *= 8
#endif

	fld	%st(1)
	fabs
	fld	%st(1)
	fabs
	faddp	%st, %st(1)	# fabs(alpha_r) + fabs(alpha_i)

	ftst
	fnstsw	%ax
	ffreep	%st
	andb	$68,%ah
	xorb	$64,%ah
	jne	.L30

	cmpl	$2 * SIZE, %ebx
	jne	.L31		# if incx == 2 goto L31

	movl	%esi, %ecx	# i = (n >> 2)
	sarl	$2,   %ecx
	jle	.L32
	.align	16

.L33:
	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	FSTU	2 * SIZE(%edx)
	FSTU	3 * SIZE(%edx)
	FSTU	4 * SIZE(%edx)
	FSTU	5 * SIZE(%edx)
	FSTU	6 * SIZE(%edx)
	FSTU	7 * SIZE(%edx)
	addl	$8 * SIZE, %edx
	decl	%ecx
	jg	.L33
	.align 8

.L32:
	movl	%esi, %ecx
	andl	$3,   %ecx
	jle	.L57		# if (n & 3) goto L57
	.align	4

.L39:
	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	$2 * SIZE, %edx
	decl	%ecx
	jg	.L39

	jmp	.L57
	.align	4

.L31:
	movl	%esi,%ecx
	sarl	$2,  %ecx
	jle	.L45
	.align	8

.L46:
	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	%ebx, %edx

	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	%ebx, %edx

	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	%ebx, %edx

	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	%ebx, %edx

	decl	%ecx
	jg	.L46
	.align 8

.L45:
	movl	%esi, %ecx
	andl	$3,   %ecx
	jle	.L57
	.align 8

.L52:
	FSTU	0 * SIZE(%edx)
	FSTU	1 * SIZE(%edx)
	addl	%ebx, %edx
	decl	%ecx
	jg	.L52
	jmp	.L57
	.align 8

.L30:
	movl	%esi,%ecx
	.align 4
.L58:
	FLD	0 * SIZE(%edx)
	fmul	%st(1),%st
	FLD	1 * SIZE(%edx)
	fmul	%st(3),%st
	faddp	%st,%st(1)

	FLD	0 * SIZE(%edx)
	fmul	%st(3),%st
	FLD	1 * SIZE(%edx)
	fmul	%st(3),%st
	fsubrp	%st,%st(1)

	FST	0 * SIZE(%edx)
	FST	1 * SIZE(%edx)
	addl	%ebx, %edx
	decl	%ecx
	jg	.L58
	.align 4

.L57:
	ffreep	%st
	ffreep	%st
	ffreep	%st
	xorl	%eax,%eax
	popl	%ebx
	popl	%esi
	ret

	EPILOGUE
