/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"

#define M	ARG1
#define X	ARG4
#define INCX	ARG5

#define I	%rsi
#define XX	%rdx

	PROLOGUE
	PROFCODE

	pxor	%xmm1, %xmm1
	COMISD	%xmm0, %xmm1
	lea	(, INCX, SIZE), INCX
	jne	.L300		# Alpha != ZERO

/* Alpha == ZERO */
	cmpq	$SIZE, INCX
	jne	.L104

	movq	M,  I	# rcx = n
	sarq	$3, I	# (n >> 3)
	jle	.L102
	.align 16

.L101:
#ifdef HAVE_3DNOW
	prefetchw	88 * SIZE(X)
#endif
	MOVSD	%xmm0, 0 * SIZE(X)
	MOVSD	%xmm0, 1 * SIZE(X)
	MOVSD	%xmm0, 2 * SIZE(X)
	MOVSD	%xmm0, 3 * SIZE(X)
	MOVSD	%xmm0, 4 * SIZE(X)
	MOVSD	%xmm0, 5 * SIZE(X)
	MOVSD	%xmm0, 6 * SIZE(X)
	MOVSD	%xmm0, 7 * SIZE(X)
	addq	$8 * SIZE, X
	decq	I
	jg	.L101
	.align 16

.L102:
	movq	M,  I
	andq	$7, I
	jle	.L999
	.align 16

.L103:
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	$SIZE, X
	decq	I
	jg	.L103
	jmp	.L999
	.align 16

.L104:
	movq	M,  I		# rcx = n
	sarq	$3, I		# (n >> 3)
	jle	.L106
	.align 16

.L105:
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X

	decq	I
	jg	.L105
	.align 16

.L106:
	movq	M, I
	andq	$7,   I
	jle	.L999
	.align 16

.L107:
	MOVSD	%xmm0, 0 * SIZE(X)
	addq	INCX, X
	decq	I
	jg	.L107
	jmp	.L999
	.align 16

/* Alpha != ZERO */

.L300:
	cmpq	$SIZE, INCX
	jne	.L304

	movq	M,  I		# rcx = n
	sarq	$3, I		# (n >> 3)
	jle	.L302
	.align 16

.L301:
#ifdef HAVE_3DNOW
	prefetchw	40 * SIZE(X)
#endif
	MOVSD	0 * SIZE(X), %xmm1
	MOVSD	1 * SIZE(X), %xmm2
	MOVSD	2 * SIZE(X), %xmm3
	MOVSD	3 * SIZE(X), %xmm4
	MOVSD	4 * SIZE(X), %xmm5
	MOVSD	5 * SIZE(X), %xmm6
	MOVSD	6 * SIZE(X), %xmm7
	MOVSD	7 * SIZE(X), %xmm8

	MULSD	%xmm0, %xmm1
	MULSD	%xmm0, %xmm2
	MULSD	%xmm0, %xmm3
	MULSD	%xmm0, %xmm4
	MULSD	%xmm0, %xmm5
	MULSD	%xmm0, %xmm6
	MULSD	%xmm0, %xmm7
	MULSD	%xmm0, %xmm8

	MOVSD	%xmm1, 0 * SIZE(X)
	MOVSD	%xmm2, 1 * SIZE(X)
	MOVSD	%xmm3, 2 * SIZE(X)
	MOVSD	%xmm4, 3 * SIZE(X)
	MOVSD	%xmm5, 4 * SIZE(X)
	MOVSD	%xmm6, 5 * SIZE(X)
	MOVSD	%xmm7, 6 * SIZE(X)
	MOVSD	%xmm8, 7 * SIZE(X)

	addq	$8 * SIZE, X
	decq	I
	jg	.L301
	.align 16

.L302:
	movq	M,  I
	andq	$7, I
	jle	.L999
	.align 16

.L303:
	MOVSD	0 * SIZE(X), %xmm1
	MULSD	%xmm0, %xmm1
	MOVSD	%xmm1, 0 * SIZE(X)
	addq	$SIZE, X
	decq	I
	jg	.L303
	jmp	.L999
	.align 16

.L304:
	movq	X, XX
	movq	M,  I		# rcx = n
	sarq	$3, I		# (n >> 3)
	jle	.L306
	.align 16

.L305:
	MOVSD	0 * SIZE(X), %xmm1
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm2
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm3
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm4
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm5
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm6
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm7
	addq	INCX, X
	MOVSD	0 * SIZE(X), %xmm8
	addq	INCX, X

	MULSD	%xmm0, %xmm1
	MULSD	%xmm0, %xmm2
	MULSD	%xmm0, %xmm3
	MULSD	%xmm0, %xmm4
	MULSD	%xmm0, %xmm5
	MULSD	%xmm0, %xmm6
	MULSD	%xmm0, %xmm7
	MULSD	%xmm0, %xmm8

	MOVSD	%xmm1, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm2, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm3, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm4, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm5, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm6, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm7, 0 * SIZE(XX)
	addq	INCX, XX
	MOVSD	%xmm8, 0 * SIZE(XX)
	addq	INCX, XX
	decq	I
	jg	.L305
	.align 16

.L306:
	movq	M, I
	andq	$7,   I
	jle	.L999
	.align 16

.L307:
	MOVSD	0 * SIZE(X), %xmm1
	MULSD	%xmm0, %xmm1
	MOVSD	%xmm1, 0 * SIZE(X)
	addq	INCX, X
	decq	I
	jg	.L307
	.align 16

.L999:
	xorq	%rax, %rax
	ret

	EPILOGUE
