/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

#ifndef SYMV_P
#define SYMV_P (DTB_ENTRIES / 2)
#endif

int CNAME(long m, FLOAT alpha, FLOAT *a,
	      long lda, FLOAT *x, long incx, FLOAT *y, long incy, void *buffer){

  long is;
  long min_i;

  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;  /* Buffer */

  for(is = 0; is < m; is += SYMV_P){

    min_i = MIN(SYMV_P, m - is);
      
#ifndef LOWER
    if (is >= SYMV_P){
      GEMV_T(is, min_i, 0, alpha, 
	     a + is * lda, lda,
	     x,            incx,
	     y + is * incy,incy, buffer);

      GEMV_N(is, min_i, 0, alpha, 
	     a + is * lda,  lda,
	     x + is * incx, incx,
	     y,             incy, buffer);
    }
#endif

#ifdef LOWER
    SYMCOPY_L(min_i, a + is + is * lda, lda, subbuffer);
#else
    SYMCOPY_U(min_i, a + is + is * lda, lda, subbuffer);
#endif
    
    GEMV_N(min_i, min_i, 0, alpha, 
	   subbuffer, min_i,
	   x + is * incx, incx, 
	   y + is * incy, incy, buffer);


#ifdef LOWER
    if (m - is - SYMV_P > 0){
      GEMV_T(m - is - SYMV_P, min_i, 0, alpha, 
	     a + (is + SYMV_P) + is * lda, lda,
	     x + (is + SYMV_P) * incx,  incx,
	     y + is * incy,         incy, buffer);
      
      GEMV_N(m - is - SYMV_P, min_i, 0, alpha, 
	     a + (is + SYMV_P) + is * lda, lda,
	     x + is * incx,         incx,
	     y + (is + SYMV_P) * incy,  incy, buffer);
    }
#endif

  } /* end of is */

  return 0;
}

