/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

int CNAME(long m, FLOAT alpha, FLOAT *x, long incx, 
		       FLOAT *y, long incy, FLOAT *a, long lda, FLOAT *buffer){

  long i, j, is, ilength;
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;

  for (is = 0; is < m; is += SYR2_P1){
    
    ilength = MIN(SYR2_P1, m - is);
    
#ifndef LOWER
    if (is > 0){
      
      GER_K(is, ilength, 0, alpha,
	  x            , incx,
	  y + is * incy, incy, 
	  a + is * lda , lda, subbuffer);
      GER_K(is, ilength, 0, alpha,
	  y            , incy,
	  x + is * incx, incx, 
	  a + is * lda , lda, subbuffer);
    }
#endif
    
    GEMM_BETA(ilength, ilength, 0, ZERO, NULL, 0, NULL, 0, buffer, ilength);
    
    GER_K(ilength, ilength, 1, alpha,
	x + is * incx, incx,
	y + is * incy, incy, 
	buffer, ilength, subbuffer);
    
    for (j = 0; j < ilength; j++){
#ifndef LOWER
      for (i = 0; i < j; i++){
	a[i + is + (j + is) * lda] += buffer[i + j * ilength]
	  + buffer[j + i * ilength];
      }
      a[j + is + (j + is) * lda] += 2 * buffer[j + j * ilength];
#else
      for (i = j + 1; i < ilength; i++){
	a[i + is + (j + is) * lda] += buffer[i + j * ilength]
	  + buffer[j + i * ilength];
      }
      a[j + is + (j + is) * lda] += 2 * buffer[j + j * ilength];
#endif
    }
    
#ifdef LOWER
    if (m - is - SYR2_P1> 0){
      GER_K(m - is - SYR2_P1, SYR2_P1, 1, alpha, 
	  x + (is + SYR2_P1) * incx, incx,
	  y + is * incy       , incy, 
	  a + is + SYR2_P1 + is *lda, lda,
	  subbuffer);
      GER_K(m - is - SYR2_P1, SYR2_P1, 1, alpha, 
	  y + (is + SYR2_P1) * incy, incy,
	  x + is * incx       , incx, 
	  a + is + SYR2_P1 + is *lda, lda,
	  subbuffer);
    }
#endif
  }

  return 0;
}
