/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

const static FLOAT dp1 = 1.;

int CNAME(long m, FLOAT *a, long lda, FLOAT *b, long incb, void *buffer){

  long i, is;
  long length;
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;
  FLOAT *B;

  B = b;

  if (incb != 1) {
    B = subbuffer;
    COPY_K(m, b, incb, subbuffer, 1);
  }

  for (is =0; is < m; is += TRSV_P1){
    
#ifndef TRANSA
    if (is >= TRSV_P1){
      GEMV_N(is, MIN(TRSV_P1, m - is), 0, dp1, 
	     a + is * lda,  lda,
	     B + is, 1,
	     B,      1, buffer);
    }
#endif

    for (i = is; i < MIN(is + TRSV_P1, m); i++) {
#ifndef UNIT
      B[i] *= a[i +  i * lda];
#endif

      length = MIN(is + TRSV_P1, m) - i - 1;
      
#ifndef TRANSA
      B[i] += DOT_K(length, a + i + (i + 1) * lda, lda, 
			B + i + 1, 1);
#else
      B[i] += DOT_K(length, a + i * lda + i + 1, 1, 
			B + i + 1, 1);
#endif
    }

#ifdef TRANSA
    if (m - is > TRSV_P1){
      GEMV_T(m - is - TRSV_P1, TRSV_P1, 0, dp1, 
	     a + is + TRSV_P1 + is * lda,  lda,
	     B + is + TRSV_P1, 1,
	     B + is,           1, buffer);
    }
#endif
  }

  if (incb != 1) {
    COPY_K(m, subbuffer, 1, b, incb);
  }

  return 0;
}

