/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

static FLOAT dp1 = 1.;

int CNAME(long m, FLOAT *a, long lda, FLOAT *b, long incb, void *buffer){

  long i, is, s;
  FLOAT temp[2];
#ifndef UNIT
  FLOAT atemp1, atemp2, btemp1, btemp2;
#endif
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;
  FLOAT *B;

  B = b;

  if (incb != 1) {
    B = subbuffer;
    ZCOPY_K(m, b, incb, subbuffer, 1);
  }

  for (is = m; is > 0; is -= TRSV_P1){

    s = MAX(is - TRSV_P1, 0);

#if (TRANSA == 1) || (TRANSA == 3)
    if (m - is > 0){
#if   TRANSA == 1
      ZGEMV_N(m - is, MIN(TRSV_P1, is), 0, dp1, ZERO,
	      a +  (is + s * lda) * 2, lda,
	      B +  s * 2, 1,
	      B + is * 2, 1, buffer);
#else
      ZGEMV_R(m - is, MIN(TRSV_P1, is), 0, dp1, ZERO,
	      a +  (is + s * lda) * 2, lda,
	      B +  s * 2, 1,
	      B + is * 2, 1, buffer);
#endif
    }
#endif

    for (i = is - 1; i >= s; i--) {

#ifndef UNIT
      atemp1 = a[(i + i * lda) * 2 + 0];
      atemp2 = a[(i + i * lda) * 2 + 1];

      btemp1 = B[i * 2 + 0];
      btemp2 = B[i * 2 + 1];

#if (TRANSA == 1) || (TRANSA == 2)
      B[i * 2 + 0] = atemp1 * btemp1 - atemp2 * btemp2;
      B[i * 2 + 1] = atemp1 * btemp2 + atemp2 * btemp1;
#else
      B[i * 2 + 0] = atemp1 * btemp1 + atemp2 * btemp2;
      B[i * 2 + 1] = atemp1 * btemp2 - atemp2 * btemp1;
#endif
#endif

      if (i - s > 0){
#if   TRANSA == 1
	temp[0] = ZDOTU_K(i - s, a + (i + s * lda) * 2, lda, B + s * 2, 1);
#elif TRANSA == 2
	temp[0] = ZDOTU_K(i - s, a + (s + i * lda) * 2, 1,   B + s * 2, 1);
#elif TRANSA == 3
	temp[0] = ZDOTC_K(i - s, a + (i + s * lda) * 2, lda, B + s * 2, 1);
#else
	temp[0] = ZDOTC_K(i - s, a + (s + i * lda) * 2, 1,   B + s * 2, 1);
#endif

	GET_IMAGE(temp[1]);

	B[i * 2 + 0] += temp[0];
	B[i * 2 + 1] += temp[1];
      }
    }
    
#if (TRANSA == 2) || (TRANSA == 4)
    if (s >0){
#if TRANSA == 2
      ZGEMV_T(s, TRSV_P1, 0, dp1, ZERO,
	      a +  s * lda * 2, lda,
	      B,         1,
	      B + s * 2, 1, buffer);
#else
      ZGEMV_C(s, TRSV_P1, 0, dp1, ZERO,
	      a +  s * lda * 2, lda,
	      B,         1,
	      B + s * 2, 1, buffer);
#endif
    }
#endif
  }

  if (incb != 1) {
    ZCOPY_K(m, subbuffer, 1, b, incb);
  }

  return 0;
}

