/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

const static FLOAT dm1 = -1.;

int CNAME(long m, FLOAT *a, long lda, FLOAT *b, long incb, void *buffer){
  
  long i, is;
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;
  FLOAT *B;

  B = b;

  if (incb != 1) {
    B = subbuffer;
    COPY_K(m, b, incb, subbuffer, 1);
  }

  for (is = 0; is < m; is += TRSV_P1){
    
#ifdef TRANSA
    if (is >= TRSV_P1){
      GEMV_T(is, MIN(m - is, TRSV_P1), 0, dm1, 
	     a + is * lda ,  lda,
	     B,       1, 
	     B +  is, 1, buffer);
    }
#endif

    for (i = is; i < MIN(is+TRSV_P1, m); i++) {

#ifndef TRANSA
	B[i] -= DOT_K(i - is, a + i  + is * lda, lda, B + is, 1);

#else
	B[i] -= DOT_K(i - is, a + is + i  * lda, 1,   B + is, 1);
#endif
      
#ifndef UNIT
      B[i] /= a[i +  i * lda];
#endif
    }
      
#ifndef TRANSA
    if (m - is > TRSV_P1){
      GEMV_N(m - is - TRSV_P1, TRSV_P1, 0, dm1, 
	     a + is + TRSV_P1 + is * lda,  lda,
	     B           + is,    1, 
	     B + (is + TRSV_P1),  1, buffer);
    }
#endif

  }

  if (incb != 1) {
    COPY_K(m, subbuffer, 1, b, incb);
  }

  return 0;
}
