/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

static FLOAT dm1 = -1.;

int CNAME(long m, FLOAT *a, long lda, FLOAT *b, long incb, void *buffer){

  long i, is;
  FLOAT result[2];
#ifndef UNIT
  FLOAT ar, ai, br, bi, ratio, den;
#endif
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;
  FLOAT *B;

  B = b;

  if (incb != 1) {
    B = subbuffer;
    ZCOPY_K(m, b, incb, subbuffer, 1);
  }

  for (is = m; is > 0; is -= TRSV_P1){

#if TRANSA == 2
    i = MAX(is - TRSV_P1, 0);
    if (m > is){
      ZGEMV_T(m - is, MIN(is, TRSV_P1), 0, dm1, ZERO,
	      a + (is + i  * lda) * COMPSIZE, lda,
	      B + is * COMPSIZE, 1,
	      B + i  * COMPSIZE, 1, buffer);
    }
#endif
#if TRANSA == 4
    i = MAX(is - TRSV_P1, 0);
    if (m > is){
      ZGEMV_C(m - is, MIN(is, TRSV_P1), 0, dm1, ZERO,
	      a + (is + i  * lda) * COMPSIZE, lda,
	      B + is * COMPSIZE, 1,
	      B + i  * COMPSIZE, 1, buffer);
    }
#endif

    for (i = is - 1; i >= MAX((is - TRSV_P1), 0); i--) {

#if   TRANSA == 1
	result[0] = ZDOTU_K(is - i - 1, a + (i + (i + 1)*lda) * COMPSIZE, lda,
			 B + (i + 1) * COMPSIZE, 1);
#elif TRANSA == 2
	result[0] = ZDOTU_K(is - i - 1, a + (i * lda + i + 1) * COMPSIZE, 1,
			 B + (i + 1) * COMPSIZE, 1);
#elif TRANSA == 3
	result[0] = ZDOTC_K(is - i - 1, a + (i + (i + 1)*lda) * COMPSIZE, lda,
			 B + (i + 1) * COMPSIZE, 1);
#else
	result[0] = ZDOTC_K(is - i - 1, a + (i * lda + i + 1) * COMPSIZE, 1,
			 B + (i + 1) * COMPSIZE, 1);
#endif

      GET_IMAGE(result[1]);

      B[i * 2 + 0] -= result[0];
      B[i * 2 + 1] -= result[1];
      
#ifndef UNIT
      ar = a[(i +  i * lda) * COMPSIZE + 0];
      ai = a[(i +  i * lda) * COMPSIZE + 1];
      
      if (fabs(ar) >= fabs(ai)){
	ratio = ai / ar;
	den = 1./(ar * ( 1 + ratio * ratio));
	
	ar =  den;
#if TRANSA < 3
	ai = -ratio * den;
#else
	ai =  ratio * den;
#endif
      } else {
	ratio = ar / ai;
	den = 1./(ai * ( 1 + ratio * ratio));
	ar =  ratio * den;
#if TRANSA < 3
	ai = -den;
#else
	ai =  den;
#endif
    }

      br = B[i * 2 + 0];
      bi = B[i * 2 + 1];
      
      B[i * 2 + 0] = ar*br - ai*bi;
      B[i * 2 + 1] = ar*bi + ai*br;
#endif
    }
    
#if   TRANSA == 1
    if (is > TRSV_P1){
      ZGEMV_N(is - TRSV_P1, TRSV_P1, 0, dm1, ZERO,
	      a +  (is - TRSV_P1) * lda * COMPSIZE, lda,
	      B +  (is - TRSV_P1)       * COMPSIZE, 1,
	      B,                                    1, buffer);
    }
#endif
#if TRANSA == 3
    if (is > TRSV_P1){
      ZGEMV_R(is - TRSV_P1, TRSV_P1, 0, dm1, ZERO,
	      a +  (is - TRSV_P1) * lda * COMPSIZE, lda,
	      B +  (is - TRSV_P1)       * COMPSIZE, 1,
	      B,                                    1, buffer);
    }
#endif
  }

  if (incb != 1) {
    ZCOPY_K(m, subbuffer, 1, b, incb);
  }

  return 0;
}

