/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"
		
#ifndef __64BIT__
#define LOAD	lwz
#else
#define LOAD	ld
#endif

#ifdef __64BIT__
#define STACKSIZE 360
#else
#define STACKSIZE 272
#endif

#define ALIGN_SIZE	0xffff
#define	BUFFER		128
#define ALPHA		  0
#define FZERO		 16
#define BUFFERSIZE	((GEMM_Q * 4 * 16) + ALIGN_SIZE + BUFFER)

#include "sparam.h"

#define	M	r3
#define	N	r4
#define	K	r5

#ifdef linux
#ifndef __64BIT__
#define A	r6
#define	B	r7
#define	C	r8
#define	LDC	r9
#else
#define A	r7
#define	B	r8
#define	C	r9
#define	LDC	r10
#endif
#endif

#if defined(_AIX) || defined(__APPLE__)
#if !defined(__64BIT__) && defined(DOUBLE)
#define A	r8
#define	B	r9
#define	C	r10
#define	LDC	r7
#else
#define A	r7
#define	B	r8
#define	C	r9
#define	LDC	r10
#endif
#endif

#define STACK	r11

#define	I	r21
#define J	r22
#define AO	r23
#define	BO	r24
#define	CO1	r25
#define CO2	r26
#define	CO3	r27
#define	CO4	r28

#define PREA	r29
#define PREC	r30
#define VREG	r31


#define OFFSET_0	  0
#define OFFSET_1	r14
#define OFFSET_2	r15
#define OFFSET_3	r16
#define OFFSET_4	r17
#define OFFSET_5	r18
#define OFFSET_6	r19
#define OFFSET_7	r20

#define PERMRSHIFT	 b6
#define PERMLSHIFT	 b7
#define alpha		 b8

#define	c01	v0
#define	c02	v1
#define	c03	v2
#define	c04	v3
#define	c05	v4
#define	c06	v5
#define	c07	v6
#define	c08	v7
#define	c09	v8
#define	c10	v9
#define	c11	v10
#define	c12	v11
#define	c13	v12
#define	c14	v13
#define	c15	v14
#define	c16	v15

#define	a1	v16
#define	a2	v17
#define	a3	v18
#define	a4	v19
#define	a5	v20
#define	a6	v21
#define	a7	v22
#define	a8	v23

#define	b1	v24
#define	b2	v25
#define	b3	v26
#define	b4	v27
#define	b5	v28
#define	b6	v29
#define	b7	v30
#define	b8	v31

#ifndef NEEDPARAM

	PROLOGUE
	PROFCODE

	addi	SP, SP, -STACKSIZE
	mr	STACK, SP

	li	r0,  0 * 16
	stvx	v20, SP, r0
	li	r0,  1 * 16
	stvx	v21, SP, r0
	li	r0,  2 * 16
	stvx	v22, SP, r0
	li	r0,  3 * 16
	stvx	v23, SP, r0
	li	r0,  4 * 16
	stvx	v24, SP, r0
	li	r0,  5 * 16
	stvx	v25, SP, r0
	li	r0,  6 * 16
	stvx	v26, SP, r0
	li	r0,  7 * 16
	stvx	v27, SP, r0
	li	r0,  8 * 16
	stvx	v28, SP, r0
	li	r0,  9 * 16
	stvx	v29, SP, r0
	li	r0, 10 * 16
	stvx	v30, SP, r0
	li	r0, 11 * 16
	stvx	v31, SP, r0

#ifdef __64BIT__
	std	r31,  192(SP)
	std	r30,  200(SP)
	std	r29,  208(SP)
	std	r28,  216(SP)
	std	r27,  224(SP)
	std	r26,  232(SP)
	std	r25,  240(SP)
	std	r24,  248(SP)
	std	r23,  256(SP)
	std	r22,  264(SP)
	std	r21,  272(SP)
	std	r20,  280(SP)
	std	r19,  288(SP)
	std	r18,  296(SP)
	std	r17,  304(SP)
	std	r16,  312(SP)
	std	r15,  320(SP)
	std	r14,  328(SP)
#else
	stw	r31,  192(SP)
	stw	r30,  196(SP)
	stw	r29,  200(SP)
	stw	r28,  204(SP)
	stw	r27,  208(SP)
	stw	r26,  212(SP)
	stw	r25,  216(SP)
	stw	r24,  220(SP)
	stw	r23,  224(SP)
	stw	r22,  228(SP)
	stw	r21,  232(SP)
	stw	r20,  236(SP)
	stw	r19,  240(SP)
	stw	r18,  244(SP)
	stw	r17,  248(SP)
	stw	r16,  252(SP)
	stw	r15,  256(SP)
	stw	r14,  260(SP)
#endif


#if defined(_AIX) || defined(__APPLE__)
#if !defined(__64BIT__) && defined(DOUBLE)
	lwz	LDC,    56 + STACKSIZE(SP)
#endif
#endif

#ifndef PREFETCHTEST
#ifdef PPC970
	li	PREC,   16 * SIZE
#endif
#else

#ifdef linux
#ifndef __64BIT__
	lwz	PREA,   16 + STACKSIZE(SP)
	lwz	PREC,   20 + STACKSIZE(SP)
#else
	ld	PREA,  136 + STACKSIZE(SP)
	ld	PREC,  144 + STACKSIZE(SP)
#endif
#endif

#if defined(_AIX) || defined(__APPLE__)
#ifdef __64BIT__
	ld	PREA,  136 + STACKSIZE(SP)
	ld	PREC,  144 + STACKSIZE(SP)
#else
#ifdef DOUBLE
	lwz	PREA,   72 + STACKSIZE(SP)
	lwz	PREC,   76 + STACKSIZE(SP)
#else
	lwz	PREA,   68 + STACKSIZE(SP)
	lwz	PREC,   72 + STACKSIZE(SP)
#endif
#endif
#endif

#endif

#ifndef PREFETCHTEST
#ifdef PPC970
#ifdef ALLOC_HUGETLB
	li	PREA,   (32 *  5 * SIZE + 0)
#else
	li	PREA,   (32 *  7 * SIZE + 0)
#endif
#endif
#endif

	li	r0, -1
	mfspr	VREG, VRsave

	mtspr	VRsave, r0

	addis	SP, SP, -Address_H(BUFFERSIZE)
	addi	SP, SP, -Address_L(BUFFERSIZE)

	li	r0, -8192

	and	SP, SP, r0

	stfs	f1,  ALPHA +  0(SP)
	stfs	f1,  ALPHA +  4(SP)
	stfs	f1,  ALPHA +  8(SP)
	stfs	f1,  ALPHA + 12(SP)
	li	r0, 0
	stw	r0, FZERO(SP)

	slwi	LDC, LDC, BASE_SHIFT

	li	OFFSET_1,  4 * SIZE
	li	OFFSET_2,  8 * SIZE
	li	OFFSET_3, 12 * SIZE
	li	OFFSET_4, 16 * SIZE
	li	OFFSET_5, 20 * SIZE
	li	OFFSET_6, 24 * SIZE
	li	OFFSET_7, 28 * SIZE

	cmpwi	cr0, M, 0
	ble	LL(999)
	cmpwi	cr0, N, 0
	ble	LL(999)
	cmpwi	cr0, K, 0
	ble	LL(999)

	srawi.	J, N,  2
	ble	LL(60)
	.align 4

LL(01):
	la	BO, BUFFER(SP)
	srawi.	r0, K,  2
	mtspr	CTR, r0
	ble	LL(05)
	.align 4

LL(02):
	lvx	a1, OFFSET_0, B
	lvx	a2, OFFSET_1, B
	lvx	a3, OFFSET_2, B
	lvx	a4, OFFSET_3, B

	vspltw	c01, a1,  0
	vspltw	c02, a1,  1
	vspltw	c03, a1,  2
	vspltw	c04, a1,  3

	vspltw	c05, a2,  0
	vspltw	c06, a2,  1
	vspltw	c07, a2,  2
	vspltw	c08, a2,  3

	stvx	c01, OFFSET_0, BO
	stvx	c02, OFFSET_1, BO
	stvx	c03, OFFSET_2, BO
	stvx	c04, OFFSET_3, BO

	stvx	c05, OFFSET_4, BO
	stvx	c06, OFFSET_5, BO
	stvx	c07, OFFSET_6, BO
	stvx	c08, OFFSET_7, BO

	addi	BO, BO, 32 * SIZE

	vspltw	c09, a3,  0
	vspltw	c10, a3,  1
	vspltw	c11, a3,  2
	vspltw	c12, a3,  3

	vspltw	c13, a4,  0
	vspltw	c14, a4,  1
	vspltw	c15, a4,  2
	vspltw	c16, a4,  3

	stvx	c09, OFFSET_0, BO
	stvx	c10, OFFSET_1, BO
	stvx	c11, OFFSET_2, BO
	stvx	c12, OFFSET_3, BO

	stvx	c13, OFFSET_4, BO
	stvx	c14, OFFSET_5, BO
	stvx	c15, OFFSET_6, BO
	stvx	c16, OFFSET_7, BO

	addi	B,  B,  16 * SIZE
	addi	BO, BO, 32 * SIZE
	bdnz	LL(02)
	.align 4
	

LL(05):
	andi.	r0, K,  3
	mtspr	CTR, r0
	ble	LL(10)
	.align 4

LL(06):
	lvx	a1, 0, B

	vspltw	b1, a1,  0
	vspltw	b2, a1,  1
	vspltw	b3, a1,  2
	vspltw	b4, a1,  3

	stvx	b1, OFFSET_0, BO
	stvx	b2, OFFSET_1, BO
	stvx	b3, OFFSET_2, BO
	stvx	b4, OFFSET_3, BO

	addi	B,  B,   4 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(06)
	.align 4

LL(10):
	mr	CO1, C
	add	CO2, C,  LDC
	add	CO3, CO2, LDC
	add	CO4, CO3, LDC
	add	C,   CO4, LDC

	mr	AO, A
	srawi.	I, M,  4
	ble	LL(20)
	.align 4

LL(11):
	la	BO, BUFFER(SP)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	lvxl	a5, OFFSET_4, AO
	lvxl	a6, OFFSET_5, AO
	lvxl	a7, OFFSET_6, AO
	lvxl	a8, OFFSET_7, AO

	lvx	b5, OFFSET_4, BO
	lvx	b6, OFFSET_5, BO
	lvx	b7, OFFSET_6, BO
	lvx	b8, OFFSET_7, BO
	
	addi	AO, AO, 32 * SIZE
	addi	BO, BO, 32 * SIZE

	vxor	c01, c01, c01
	vxor	c02, c02, c02
	vxor	c03, c03, c03
	vxor	c04, c04, c04
	vxor	c05, c05, c05
	vxor	c06, c06, c06
	vxor	c07, c07, c07
	vxor	c08, c08, c08
	vxor	c09, c09, c09
	vxor	c10, c10, c10
	vxor	c11, c11, c11
	vxor	c12, c12, c12
	vxor	c13, c13, c13
	vxor	c14, c14, c14
	vxor	c15, c15, c15
	vxor	c16, c16, c16

	PREFETCH_C1
	PREFETCH_C2
	PREFETCH_C3
	PREFETCH_C4

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(15)
	.align 4

#define	fnop00  fmr	f0, f0
#define	fnop01  fmr	f1, f1
#define	fnop02  fmr	f2, f2
#define	fnop03  fmr	f3, f3
#define	fnop04  fmr	f4, f4
#define	fnop05  fmr	f5, f5
#define	fnop06  fmr	f6, f6
#define	fnop07  fmr	f7, f7
#define	fnop08  fmr	f8, f8
#define	fnop09  fmr	f9, f9

LL(12):
	vmaddfp	c01, a1, b1, c01
	PREFETCH_A
	vmaddfp	c05, a1, b2, c05
	PREFETCH_B
	vmaddfp	c09, a1, b3, c09
	nop
	vmaddfp	c13, a1, b4, c13
	lvxl	a1, OFFSET_0, AO

	vmaddfp	c02, a2, b1, c02
	nop
	vmaddfp	c06, a2, b2, c06
	nop
	vmaddfp	c10, a2, b3, c10
	nop
	vmaddfp	c14, a2, b4, c14
	lvxl	a2, OFFSET_1, AO

	vmaddfp	c03, a3, b1, c03
	nop
	vmaddfp	c07, a3, b2, c07
	nop
	vmaddfp	c11, a3, b3, c11
	nop
	vmaddfp	c15, a3, b4, c15
	lvxl	a3, OFFSET_2, AO

	vmaddfp	c04, a4, b1, c04
	lvx	b1, OFFSET_0, BO
	vmaddfp	c08, a4, b2, c08
	lvx	b2, OFFSET_1, BO
	vmaddfp	c12, a4, b3, c12
	lvx	b3, OFFSET_2, BO
	vmaddfp	c16, a4, b4, c16
	lvxl	a4, OFFSET_3, AO

	vmaddfp	c01, a5, b5, c01
	lvx	b4, OFFSET_3, BO
	vmaddfp	c05, a5, b6, c05
	nop
	vmaddfp	c09, a5, b7, c09
	nop
	vmaddfp	c13, a5, b8, c13
	lvxl	a5, OFFSET_4, AO

	vmaddfp	c02, a6, b5, c02
	nop
	vmaddfp	c06, a6, b6, c06
	nop
	vmaddfp	c10, a6, b7, c10
	nop
	vmaddfp	c14, a6, b8, c14
	lvxl	a6, OFFSET_5, AO

	vmaddfp	c03, a7, b5, c03
	nop
	vmaddfp	c07, a7, b6, c07
	nop
	vmaddfp	c11, a7, b7, c11
	nop
	vmaddfp	c15, a7, b8, c15
	lvxl	a7, OFFSET_6, AO

	vmaddfp	c04, a8, b5, c04
	lvx	b5, OFFSET_4, BO
	vmaddfp	c08, a8, b6, c08
	lvx	b6, OFFSET_5, BO
	vmaddfp	c12, a8, b7, c12
	lvx	b7, OFFSET_6, BO
	vmaddfp	c16, a8, b8, c16
	lvxl	a8, OFFSET_7, AO

	addi	AO, AO, 32 * SIZE
	lvx	b8, OFFSET_7, BO
	addi	BO, BO, 32 * SIZE
	bdnz	LL(12)
	.align 4

LL(15):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	lfs	f2, 0(CO1)
	lfs	f3, 0(CO2)
	lfs	f4, 0(CO3)
	lfs	f5, 0(CO4)

	addi	AO, AO, -32 * SIZE
	addi	BO, BO, -32 * SIZE
	ble+	LL(18)
	.align 4

LL(16):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c09, a1, b3, c09
	vmaddfp	c13, a1, b4, c13

	vmaddfp	c02, a2, b1, c02
	vmaddfp	c06, a2, b2, c06
	vmaddfp	c10, a2, b3, c10
	vmaddfp	c14, a2, b4, c14

	vmaddfp	c03, a3, b1, c03
	vmaddfp	c07, a3, b2, c07
	vmaddfp	c11, a3, b3, c11
	vmaddfp	c15, a3, b4, c15

	vmaddfp	c04, a4, b1, c04
	vmaddfp	c08, a4, b2, c08
	vmaddfp	c12, a4, b3, c12
	vmaddfp	c16, a4, b4, c16

	addi	BO, BO, 16 * SIZE
	addi	AO, AO, 16 * SIZE
	.align 4

LL(18):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	lvx	a5, OFFSET_3, CO1
	lvx	a6, OFFSET_4, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2
	vmaddfp	c03, alpha, c03, b3
	vmaddfp	c04, alpha, c04, b4

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, c03, PERMRSHIFT
	vperm	a5, c03, c04, PERMRSHIFT
	vperm	a6, c04, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1
	stvx	a5, OFFSET_3, CO1
	stvx	a6, OFFSET_4, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	lvx	a4, OFFSET_2, CO2
	lvx	a5, OFFSET_3, CO2
	lvx	a6, OFFSET_4, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1
	vmaddfp	c06, alpha, c06, b2
	vmaddfp	c07, alpha, c07, b3
	vmaddfp	c08, alpha, c08, b4

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, c06, PERMRSHIFT
	vperm	a4, c06, c07, PERMRSHIFT
	vperm	a5, c07, c08, PERMRSHIFT
	vperm	a6, c08, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2
	stvx	a4, OFFSET_2, CO2
	stvx	a5, OFFSET_3, CO2
	stvx	a6, OFFSET_4, CO2

	lvx	a2, OFFSET_0, CO3
	lvx	a3, OFFSET_1, CO3
	lvx	a4, OFFSET_2, CO3
	lvx	a5, OFFSET_3, CO3
	lvx	a6, OFFSET_4, CO3
	
	lvsl	PERMLSHIFT, 0, CO3
	lvsr	PERMRSHIFT, 0, CO3

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c09, alpha, c09, b1
	vmaddfp	c10, alpha, c10, b2
	vmaddfp	c11, alpha, c11, b3
	vmaddfp	c12, alpha, c12, b4

	vperm	a2, a1,  c09, PERMRSHIFT
	vperm	a3, c09, c10, PERMRSHIFT
	vperm	a4, c10, c11, PERMRSHIFT
	vperm	a5, c11, c12, PERMRSHIFT
	vperm	a6, c12, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO3
	stvx	a3, OFFSET_1, CO3
	stvx	a4, OFFSET_2, CO3
	stvx	a5, OFFSET_3, CO3
	stvx	a6, OFFSET_4, CO3

	lvx	a2, OFFSET_0, CO4
	lvx	a3, OFFSET_1, CO4
	lvx	a4, OFFSET_2, CO4
	lvx	a5, OFFSET_3, CO4
	lvx	a6, OFFSET_4, CO4
	
	lvsl	PERMLSHIFT, 0, CO4
	lvsr	PERMRSHIFT, 0, CO4

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c13, alpha, c13, b1
	vmaddfp	c14, alpha, c14, b2
	vmaddfp	c15, alpha, c15, b3
	vmaddfp	c16, alpha, c16, b4

	vperm	a2, a1,  c13, PERMRSHIFT
	vperm	a3, c13, c14, PERMRSHIFT
	vperm	a4, c14, c15, PERMRSHIFT
	vperm	a5, c15, c16, PERMRSHIFT
	vperm	a6, c16, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO4
	stvx	a3, OFFSET_1, CO4
	stvx	a4, OFFSET_2, CO4
	stvx	a5, OFFSET_3, CO4
	stvx	a6, OFFSET_4, CO4

	addi	CO1, CO1, 16 * SIZE
	addi	CO2, CO2, 16 * SIZE
	addi	CO3, CO3, 16 * SIZE
	addi	CO4, CO4, 16 * SIZE
	
	addic.	I, I, -1
	bgt+	LL(11)
	.align 4

LL(20):
	andi.	I, M,  8
	la	BO, BUFFER(SP)
	ble	LL(30)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	lvx	b5, OFFSET_4, BO
	lvx	b6, OFFSET_5, BO
	lvx	b7, OFFSET_6, BO
	lvx	b8, OFFSET_7, BO
	
	addi	AO, AO, 16 * SIZE
	addi	BO, BO, 32 * SIZE

	vxor	c01, c01, c01
	vxor	c05, c05, c05
	vxor	c09, c09, c09
	vxor	c13, c13, c13
	vxor	c02, c02, c02
	vxor	c06, c06, c06
	vxor	c10, c10, c10
	vxor	c14, c14, c14

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(25)
	.align 4

LL(22):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c09, a1, b3, c09
	vmaddfp	c13, a1, b4, c13
	lvxl	a1, OFFSET_0, AO

	vmaddfp	c02, a2, b1, c02
	lvx	b1, OFFSET_0, BO
	vmaddfp	c06, a2, b2, c06
	lvx	b2, OFFSET_1, BO
	vmaddfp	c10, a2, b3, c10
	lvx	b3, OFFSET_2, BO
	vmaddfp	c14, a2, b4, c14
	lvxl	a2, OFFSET_1, AO

	vmaddfp	c01, a3, b5, c01
	lvx	b4, OFFSET_3, BO
	vmaddfp	c05, a3, b6, c05
	vmaddfp	c09, a3, b7, c09
	vmaddfp	c13, a3, b8, c13
	lvxl	a3, OFFSET_2, AO

	vmaddfp	c02, a4, b5, c02
	lvx	b5, OFFSET_4, BO
	vmaddfp	c06, a4, b6, c06
	lvx	b6, OFFSET_5, BO
	vmaddfp	c10, a4, b7, c10
	lvx	b7, OFFSET_6, BO
	vmaddfp	c14, a4, b8, c14
	lvx	b8, OFFSET_7, BO

	addi	BO, BO, 32 * SIZE
	lvxl	a4, OFFSET_3, AO
	addi	AO, AO, 16 * SIZE
	bdnz	LL(22)
	.align 4

LL(25):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO, -16 * SIZE
	addi	BO, BO, -32 * SIZE
	ble+	LL(28)
	.align 4

LL(26):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c09, a1, b3, c09
	vmaddfp	c13, a1, b4, c13

	vmaddfp	c02, a2, b1, c02
	vmaddfp	c06, a2, b2, c06
	vmaddfp	c10, a2, b3, c10
	vmaddfp	c14, a2, b4, c14

	addi	AO, AO,  8 * SIZE
	addi	BO, BO, 16 * SIZE
	.align 4

LL(28):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	lvx	a4, OFFSET_2, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1
	vmaddfp	c06, alpha, c06, b2

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, c06, PERMRSHIFT
	vperm	a4, c06, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2
	stvx	a4, OFFSET_2, CO2

	lvx	a2, OFFSET_0, CO3
	lvx	a3, OFFSET_1, CO3
	lvx	a4, OFFSET_2, CO3
	
	lvsl	PERMLSHIFT, 0, CO3
	lvsr	PERMRSHIFT, 0, CO3

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c09, alpha, c09, b1
	vmaddfp	c10, alpha, c10, b2

	vperm	a2, a1,  c09, PERMRSHIFT
	vperm	a3, c09, c10, PERMRSHIFT
	vperm	a4, c10, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO3
	stvx	a3, OFFSET_1, CO3
	stvx	a4, OFFSET_2, CO3

	lvx	a2, OFFSET_0, CO4
	lvx	a3, OFFSET_1, CO4
	lvx	a4, OFFSET_2, CO4
	
	lvsl	PERMLSHIFT, 0, CO4
	lvsr	PERMRSHIFT, 0, CO4

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c13, alpha, c13, b1
	vmaddfp	c14, alpha, c14, b2

	vperm	a2, a1,  c13, PERMRSHIFT
	vperm	a3, c13, c14, PERMRSHIFT
	vperm	a4, c14, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO4
	stvx	a3, OFFSET_1, CO4
	stvx	a4, OFFSET_2, CO4

	addi	CO1, CO1, 8 * SIZE
	addi	CO2, CO2, 8 * SIZE
	addi	CO3, CO3, 8 * SIZE
	addi	CO4, CO4, 8 * SIZE
	.align 4

LL(30):
	andi.	I, M,  4
	la	BO, BUFFER(SP)
	ble	LL(40)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	lvx	b5, OFFSET_4, BO
	lvx	b6, OFFSET_5, BO
	lvx	b7, OFFSET_6, BO
	lvx	b8, OFFSET_7, BO
	
	addi	AO, AO,  8 * SIZE
	addi	BO, BO, 32 * SIZE

	vxor	c01, c01, c01
	vxor	c05, c05, c05
	vxor	c09, c09, c09
	vxor	c13, c13, c13

	vxor	c02, c02, c02
	vxor	c06, c06, c06
	vxor	c10, c10, c10
	vxor	c14, c14, c14

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(35)
	.align 4

LL(32):
	vmaddfp	c01, a1, b1, c01
	lvx	b1, OFFSET_0, BO
	vmaddfp	c05, a1, b2, c05
	lvx	b2, OFFSET_1, BO
	vmaddfp	c09, a1, b3, c09
	lvx	b3, OFFSET_2, BO
	vmaddfp	c13, a1, b4, c13
	lvx	b4, OFFSET_3, BO
	lvxl	a1, OFFSET_0, AO

	vmaddfp	c02, a2, b5, c02
	lvx	b5, OFFSET_4, BO
	vmaddfp	c06, a2, b6, c06
	lvx	b6, OFFSET_5, BO
	vmaddfp	c10, a2, b7, c10
	lvx	b7, OFFSET_6, BO
	vmaddfp	c14, a2, b8, c14
	lvx	b8, OFFSET_7, BO
	lvxl	a2, OFFSET_1, AO

	addi	AO, AO,  8 * SIZE
	addi	BO, BO, 32 * SIZE
	bdnz	LL(32)
	.align 4

LL(35):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO,  -8 * SIZE
	addi	BO, BO, -32 * SIZE
	ble+	LL(38)
	.align 4

LL(36):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c09, a1, b3, c09
	vmaddfp	c13, a1, b4, c13

	addi	AO, AO,  4 * SIZE
	addi	BO, BO, 16 * SIZE
	.align 4

LL(38):
	vaddfp	c01, c01, c02
	vaddfp	c05, c05, c06
	vaddfp	c09, c09, c10
	vaddfp	c13, c13, c14

	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2

	lvx	a2, OFFSET_0, CO3
	lvx	a3, OFFSET_1, CO3
	
	lvsl	PERMLSHIFT, 0, CO3
	lvsr	PERMRSHIFT, 0, CO3

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c09, alpha, c09, b1

	vperm	a2, a1,  c09, PERMRSHIFT
	vperm	a3, c09, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO3
	stvx	a3, OFFSET_1, CO3

	lvx	a2, OFFSET_0, CO4
	lvx	a3, OFFSET_1, CO4
	
	lvsl	PERMLSHIFT, 0, CO4
	lvsr	PERMRSHIFT, 0, CO4

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c13, alpha, c13, b1

	vperm	a2, a1,  c13, PERMRSHIFT
	vperm	a3, c13, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO4
	stvx	a3, OFFSET_1, CO4

	addi	CO1, CO1, 4 * SIZE
	addi	CO2, CO2, 4 * SIZE
	addi	CO3, CO3, 4 * SIZE
	addi	CO4, CO4, 4 * SIZE
	.align 4

LL(40):
	andi.	I, M,  2
	la	BO, BUFFER(SP)
	ble	LL(50)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)

	LFD	f10,  0 * SIZE(BO)
	LFD	f11,  4 * SIZE(BO)
	LFD	f12,  8 * SIZE(BO)
	LFD	f13, 12 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0
	fmr	f2,  f0
	fmr	f3,  f0

	fmr	f4,  f0
	fmr	f5,  f0
	fmr	f6,  f0
	fmr	f7,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(45)
	.align 4

LL(42):
	FMADD	f0,  f8, f10, f0
	FMADD	f2,  f8, f11, f2
	FMADD	f4,  f8, f12, f4
	FMADD	f6,  f8, f13, f6

	FMADD	f1,  f9, f10, f1
	FMADD	f3,  f9, f11, f3
	FMADD	f5,  f9, f12, f5
	FMADD	f7,  f9, f13, f7

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	FMADD	f0,  f8, f10, f0
	FMADD	f2,  f8, f11, f2
	FMADD	f4,  f8, f12, f4
	FMADD	f6,  f8, f13, f6

	FMADD	f1,  f9, f10, f1
	FMADD	f3,  f9, f11, f3
	FMADD	f5,  f9, f12, f5
	FMADD	f7,  f9, f13, f7

	LFD	f8,   4 * SIZE(AO)
	LFD	f9,   5 * SIZE(AO)

	LFD	f10, 32 * SIZE(BO)
	LFD	f11, 36 * SIZE(BO)
	LFD	f12, 40 * SIZE(BO)
	LFD	f13, 44 * SIZE(BO)

	addi	AO, AO,  4 * SIZE
	addi	BO, BO, 32 * SIZE
	bdnz	LL(42)
	.align 4

LL(45):
	andi.	r0,  K,  1
	ble	LL(48)
	.align 4

LL(46):
	FMADD	f0,  f8, f10, f0
	FMADD	f2,  f8, f11, f2
	FMADD	f4,  f8, f12, f4
	FMADD	f6,  f8, f13, f6

	FMADD	f1,  f9, f10, f1
	FMADD	f3,  f9, f11, f3
	FMADD	f5,  f9, f12, f5
	FMADD	f7,  f9, f13, f7

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	addi	AO, AO,  2 * SIZE
	addi	BO, BO, 16 * SIZE
	.align 4

LL(48):
	lfs	f13,  ALPHA(SP)

	LFD	f8,  0 * SIZE(CO1)
	LFD	f9,  1 * SIZE(CO1)
	LFD	f10, 0 * SIZE(CO2)
	LFD	f11, 1 * SIZE(CO2)

	FMADD	f0,  f0, f13, f8
	FMADD	f1,  f1, f13, f9
	FMADD	f2,  f2, f13, f10
	FMADD	f3,  f3, f13, f11

	LFD	f8,  0 * SIZE(CO3)
	LFD	f9,  1 * SIZE(CO3)
	LFD	f10, 0 * SIZE(CO4)
	LFD	f11, 1 * SIZE(CO4)

	FMADD	f4,  f4, f13, f8
	FMADD	f5,  f5, f13, f9
	FMADD	f6,  f6, f13, f10
	FMADD	f7,  f7, f13, f11

	STFD	f0,  0 * SIZE(CO1)
	STFD	f1,  1 * SIZE(CO1)
	STFD	f2,  0 * SIZE(CO2)
	STFD	f3,  1 * SIZE(CO2)

	STFD	f4,  0 * SIZE(CO3)
	STFD	f5,  1 * SIZE(CO3)
	STFD	f6,  0 * SIZE(CO4)
	STFD	f7,  1 * SIZE(CO4)

	addi	CO1, CO1, 2 * SIZE
	addi	CO2, CO2, 2 * SIZE
	addi	CO3, CO3, 2 * SIZE
	addi	CO4, CO4, 2 * SIZE
	.align 4

LL(50):
	andi.	I, M,  1
	la	BO, BUFFER(SP)
	ble	LL(59)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)

	LFD	f10,  0 * SIZE(BO)
	LFD	f11,  4 * SIZE(BO)
	LFD	f12,  8 * SIZE(BO)
	LFD	f13, 12 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0
	fmr	f2,  f0
	fmr	f3,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(55)
	.align 4

LL(52):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f8, f11, f1
	FMADD	f2,  f8, f12, f2
	FMADD	f3,  f8, f13, f3

	LFD	f8,   2 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	FMADD	f0,  f9, f10, f0
	FMADD	f1,  f9, f11, f1
	FMADD	f2,  f9, f12, f2
	FMADD	f3,  f9, f13, f3

	LFD	f9,   3 * SIZE(AO)

	LFD	f10, 32 * SIZE(BO)
	LFD	f11, 36 * SIZE(BO)
	LFD	f12, 40 * SIZE(BO)
	LFD	f13, 44 * SIZE(BO)

	addi	AO, AO,  2 * SIZE
	addi	BO, BO, 32 * SIZE
	bdnz	LL(52)
	.align 4

LL(55):
	andi.	r0,  K,  1
	ble	LL(58)
	.align 4

LL(56):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f8, f11, f1
	FMADD	f2,  f8, f12, f2
	FMADD	f3,  f8, f13, f3

	LFD	f8,   2 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	addi	AO, AO,  1 * SIZE
	addi	BO, BO, 16 * SIZE
	.align 4

LL(58):
	lfs	f13,  ALPHA(SP)

	LFD	f8,  0 * SIZE(CO1)
	LFD	f9,  0 * SIZE(CO2)
	LFD	f10, 0 * SIZE(CO3)
	LFD	f11, 0 * SIZE(CO4)

	FMADD	f0,  f0, f13, f8
	FMADD	f1,  f1, f13, f9
	FMADD	f2,  f2, f13, f10
	FMADD	f3,  f3, f13, f11

	STFD	f0,  0 * SIZE(CO1)
	STFD	f1,  0 * SIZE(CO2)
	STFD	f2,  0 * SIZE(CO3)
	STFD	f3,  0 * SIZE(CO4)
	.align 4

LL(59):
	addic.	J, J, -1
	bgt	LL(01)
	.align 4

LL(60):
	andi.	J, N,  2
	ble	LL(120)

	la	BO, BUFFER(SP)
	srawi.	r0, K,  1
	mtspr	CTR, r0
	ble	LL(65)
	.align 4

LL(62):
	lfs	f0,  0 * SIZE(B)
	lfs	f1,  1 * SIZE(B)
	lfs	f2,  2 * SIZE(B)
	lfs	f3,  3 * SIZE(B)

	stfs	f0,  0 * SIZE(BO)
	stfs	f0,  1 * SIZE(BO)
	stfs	f0,  2 * SIZE(BO)
	stfs	f0,  3 * SIZE(BO)

	stfs	f1,  4 * SIZE(BO)
	stfs	f1,  5 * SIZE(BO)
	stfs	f1,  6 * SIZE(BO)
	stfs	f1,  7 * SIZE(BO)

	stfs	f2,  8 * SIZE(BO)
	stfs	f2,  9 * SIZE(BO)
	stfs	f2, 10 * SIZE(BO)
	stfs	f2, 11 * SIZE(BO)

	stfs	f3, 12 * SIZE(BO)
	stfs	f3, 13 * SIZE(BO)
	stfs	f3, 14 * SIZE(BO)
	stfs	f3, 15 * SIZE(BO)

	addi	B,  B,   4 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(62)
	.align 4

LL(65):
	andi.	r0, K,  1
	ble	LL(70)

	lfs	f0,  0 * SIZE(B)
	lfs	f1,  1 * SIZE(B)

	stfs	f0,  0 * SIZE(BO)
	stfs	f0,  1 * SIZE(BO)
	stfs	f0,  2 * SIZE(BO)
	stfs	f0,  3 * SIZE(BO)

	stfs	f1,  4 * SIZE(BO)
	stfs	f1,  5 * SIZE(BO)
	stfs	f1,  6 * SIZE(BO)
	stfs	f1,  7 * SIZE(BO)

	addi	B,  B,   2 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(70):
	mr	CO1, C
	add	CO2, C,  LDC
	add	C,  CO2, LDC

	mr	AO, A
	srawi.	I, M,  4
	ble	LL(80)
	.align 4

LL(71):
	la	BO, BUFFER(SP)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	lvxl	a5, OFFSET_4, AO
	lvxl	a6, OFFSET_5, AO
	lvxl	a7, OFFSET_6, AO
	lvxl	a8, OFFSET_7, AO

	addi	AO, AO, 32 * SIZE
	addi	BO, BO, 16 * SIZE

	vxor	c01, c01, c01
	vxor	c02, c02, c02
	vxor	c03, c03, c03
	vxor	c04, c04, c04
	vxor	c05, c05, c05
	vxor	c06, c06, c06
	vxor	c07, c07, c07
	vxor	c08, c08, c08

	PREFETCH_C1
	PREFETCH_C2

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(75)
	.align 4

LL(72):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	lvxl	a1, OFFSET_0, AO
	vmaddfp	c02, a2, b1, c02
	vmaddfp	c06, a2, b2, c06
	lvxl	a2, OFFSET_1, AO

	vmaddfp	c03, a3, b1, c03
	vmaddfp	c07, a3, b2, c07
	lvxl	a3, OFFSET_2, AO
	vmaddfp	c04, a4, b1, c04
	lvx	b1, OFFSET_0, BO
	vmaddfp	c08, a4, b2, c08
	lvxl	a4, OFFSET_3, AO
	lvx	b2, OFFSET_1, BO

	vmaddfp	c01, a5, b3, c01
	vmaddfp	c05, a5, b4, c05
	lvxl	a5, OFFSET_4, AO
	vmaddfp	c02, a6, b3, c02
	vmaddfp	c06, a6, b4, c06
	lvxl	a6, OFFSET_5, AO

	vmaddfp	c03, a7, b3, c03
	vmaddfp	c07, a7, b4, c07
	lvxl	a7, OFFSET_6, AO
	vmaddfp	c04, a8, b3, c04
	lvx	b3, OFFSET_2, BO
	vmaddfp	c08, a8, b4, c08
	lvxl	a8, OFFSET_7, AO
	lvx	b4, OFFSET_3, BO

	addi	AO, AO, 32 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(72)
	.align 4

LL(75):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP
	addi	AO, AO, -32 * SIZE
	addi	BO, BO, -16 * SIZE
	ble+	LL(78)
	.align 4

LL(76):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c02, a2, b1, c02
	vmaddfp	c06, a2, b2, c06

	vmaddfp	c03, a3, b1, c03
	vmaddfp	c07, a3, b2, c07
	vmaddfp	c04, a4, b1, c04
	vmaddfp	c08, a4, b2, c08

	addi	AO, AO, 16 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(78):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	lvx	a5, OFFSET_3, CO1
	lvx	a6, OFFSET_4, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2
	vmaddfp	c03, alpha, c03, b3
	vmaddfp	c04, alpha, c04, b4

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, c03, PERMRSHIFT
	vperm	a5, c03, c04, PERMRSHIFT
	vperm	a6, c04, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1
	stvx	a5, OFFSET_3, CO1
	stvx	a6, OFFSET_4, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	lvx	a4, OFFSET_2, CO2
	lvx	a5, OFFSET_3, CO2
	lvx	a6, OFFSET_4, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1
	vmaddfp	c06, alpha, c06, b2
	vmaddfp	c07, alpha, c07, b3
	vmaddfp	c08, alpha, c08, b4

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, c06, PERMRSHIFT
	vperm	a4, c06, c07, PERMRSHIFT
	vperm	a5, c07, c08, PERMRSHIFT
	vperm	a6, c08, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2
	stvx	a4, OFFSET_2, CO2
	stvx	a5, OFFSET_3, CO2
	stvx	a6, OFFSET_4, CO2

	addi	CO1, CO1, 16 * SIZE
	addi	CO2, CO2, 16 * SIZE
	addic.	I, I, -1
	bgt+	LL(71)
	.align 4

LL(80):
	andi.	I, M,  8
	la	BO, BUFFER(SP)
	ble	LL(90)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	addi	AO, AO, 16 * SIZE
	addi	BO, BO, 16 * SIZE

	vxor	c01, c01, c01
	vxor	c05, c05, c05
	vxor	c02, c02, c02
	vxor	c06, c06, c06

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(85)
	.align 4

LL(82):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	lvxl	a1, OFFSET_0, AO
	vmaddfp	c02, a2, b1, c02
	lvx	b1, OFFSET_0, BO
	vmaddfp	c06, a2, b2, c06
	lvxl	a2, OFFSET_1, AO
	lvx	b2, OFFSET_1, BO

	vmaddfp	c01, a3, b3, c01
	vmaddfp	c05, a3, b4, c05
	lvxl	a3, OFFSET_2, AO
	vmaddfp	c02, a4, b3, c02
	lvx	b3, OFFSET_2, BO
	vmaddfp	c06, a4, b4, c06
	lvxl	a4, OFFSET_3, AO
	lvx	b4, OFFSET_3, BO

	addi	BO, BO, 16 * SIZE
	addi	AO, AO, 16 * SIZE
	bdnz	LL(82)
	.align 4

LL(85):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO, -16 * SIZE
	addi	BO, BO, -16 * SIZE
	ble+	LL(88)
	.align 4

LL(86):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05
	vmaddfp	c02, a2, b1, c02
	vmaddfp	c06, a2, b2, c06

	addi	AO, AO,  8 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(88):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	lvx	a4, OFFSET_2, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1
	vmaddfp	c06, alpha, c06, b2

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, c06, PERMRSHIFT
	vperm	a4, c06, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2
	stvx	a4, OFFSET_2, CO2

	addi	CO1, CO1, 8 * SIZE
	addi	CO2, CO2, 8 * SIZE
	.align 4

LL(90):
	andi.	I, M,  4
	la	BO, BUFFER(SP)
	ble	LL(100)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	lvx	b3, OFFSET_2, BO
	lvx	b4, OFFSET_3, BO
	
	addi	AO, AO,  8 * SIZE
	addi	BO, BO, 16 * SIZE

	vxor	c01, c01, c01
	vxor	c05, c05, c05
	vxor	c02, c02, c02
	vxor	c06, c06, c06

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(95)
	.align 4

LL(92):
	vmaddfp	c01, a1, b1, c01
	lvx	b1, OFFSET_0, BO
	vmaddfp	c05, a1, b2, c05
	lvxl	a1, OFFSET_0, AO
	lvx	b2, OFFSET_1, BO

	vmaddfp	c02, a2, b3, c02
	lvx	b3, OFFSET_2, BO
	vmaddfp	c06, a2, b4, c06
	lvxl	a2, OFFSET_1, AO
	lvx	b4, OFFSET_3, BO

	addi	AO, AO,  8 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(92)
	.align 4

LL(95):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO,  -8 * SIZE
	addi	BO, BO, -16 * SIZE
	ble+	LL(98)
	.align 4

LL(96):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c05, a1, b2, c05

	addi	AO, AO,  4 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(98):
	vaddfp	c01, c01, c02
	vaddfp	c05, c05, c06

	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1

	lvx	a2, OFFSET_0, CO2
	lvx	a3, OFFSET_1, CO2
	
	lvsl	PERMLSHIFT, 0, CO2
	lvsr	PERMRSHIFT, 0, CO2

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c05, alpha, c05, b1

	vperm	a2, a1,  c05, PERMRSHIFT
	vperm	a3, c05, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO2
	stvx	a3, OFFSET_1, CO2

	addi	CO1, CO1, 4 * SIZE
	addi	CO2, CO2, 4 * SIZE
	.align 4

LL(100):
	andi.	I, M,  2
	la	BO, BUFFER(SP)
	ble	LL(110)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)

	LFD	f10,  0 * SIZE(BO)
	LFD	f11,  4 * SIZE(BO)
	LFD	f12,  8 * SIZE(BO)
	LFD	f13, 12 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0
	fmr	f2,  f0
	fmr	f3,  f0

	fmr	f4,  f0
	fmr	f5,  f0
	fmr	f6,  f0
	fmr	f7,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(105)
	.align 4

LL(102):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f9, f10, f1
	FMADD	f2,  f8, f11, f2
	FMADD	f3,  f9, f11, f3

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)

	FMADD	f4,  f8, f12, f4
	FMADD	f5,  f9, f12, f5
	FMADD	f6,  f8, f13, f6
	FMADD	f7,  f9, f13, f7

	LFD	f8,  4 * SIZE(AO)
	LFD	f9,  5 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	addi	AO, AO,  4 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(102)
	.align 4

LL(105):
	andi.	r0,  K,  1
	lfs	f13,  ALPHA(SP)
	ble	LL(108)
	.align 4

LL(106):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f9, f10, f1
	FMADD	f2,  f8, f11, f2
	FMADD	f3,  f9, f11, f3

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)

	LFD	f10,  8 * SIZE(BO)
	LFD	f11, 12 * SIZE(BO)

	addi	AO, AO,  2 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(108):
	LFD	f8,  0 * SIZE(CO1)
	LFD	f9,  1 * SIZE(CO1)
	LFD	f10, 0 * SIZE(CO2)
	LFD	f11, 1 * SIZE(CO2)

	FADD	f0, f0, f4
	FADD	f1, f1, f5
	FADD	f2, f2, f6
	FADD	f3, f3, f7

	FMADD	f0,  f0, f13, f8
	FMADD	f1,  f1, f13, f9
	FMADD	f2,  f2, f13, f10
	FMADD	f3,  f3, f13, f11

	STFD	f0,  0 * SIZE(CO1)
	STFD	f1,  1 * SIZE(CO1)
	STFD	f2,  0 * SIZE(CO2)
	STFD	f3,  1 * SIZE(CO2)

	addi	CO1, CO1, 2 * SIZE
	addi	CO2, CO2, 2 * SIZE
	.align 4

LL(110):
	andi.	I, M,  1
	la	BO, BUFFER(SP)
	ble	LL(120)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)

	LFD	f10,  0 * SIZE(BO)
	LFD	f11,  4 * SIZE(BO)
	LFD	f12,  8 * SIZE(BO)
	LFD	f13, 12 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0
	fmr	f2,  f0
	fmr	f3,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(115)
	.align 4

LL(112):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f8, f11, f1
	FMADD	f2,  f9, f12, f2
	FMADD	f3,  f9, f13, f3

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)

	LFD	f10, 16 * SIZE(BO)
	LFD	f11, 20 * SIZE(BO)
	LFD	f12, 24 * SIZE(BO)
	LFD	f13, 28 * SIZE(BO)

	addi	AO, AO,  2 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(112)
	.align 4

LL(115):
	andi.	r0,  K,  1
	lfs	f13,  ALPHA(SP)
	ble	LL(118)
	.align 4

LL(116):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f8, f11, f1

	LFD	f8,   1 * SIZE(AO)

	LFD	f10,  8 * SIZE(BO)
	LFD	f11, 12 * SIZE(BO)

	addi	AO, AO,  1 * SIZE
	addi	BO, BO,  8 * SIZE
	.align 4

LL(118):
	LFD	f8,  0 * SIZE(CO1)
	LFD	f9,  0 * SIZE(CO2)

	FADD	f0, f0, f2
	FADD	f1, f1, f3
	
	FMADD	f0,  f0, f13, f8
	FMADD	f1,  f1, f13, f9

	STFD	f0,  0 * SIZE(CO1)
	STFD	f1,  0 * SIZE(CO2)
	.align 4

LL(120):
	andi.	J, N,  1
	ble	LL(999)

	la	BO, BUFFER(SP)
	srawi.	r0, K,  2
	mtspr	CTR, r0
	ble	LL(125)
	.align 4

LL(122):
	lfs	f0,  0 * SIZE(B)
	lfs	f1,  1 * SIZE(B)
	lfs	f2,  2 * SIZE(B)
	lfs	f3,  3 * SIZE(B)

	stfs	f0,  0 * SIZE(BO)
	stfs	f0,  1 * SIZE(BO)
	stfs	f0,  2 * SIZE(BO)
	stfs	f0,  3 * SIZE(BO)

	stfs	f1,  4 * SIZE(BO)
	stfs	f1,  5 * SIZE(BO)
	stfs	f1,  6 * SIZE(BO)
	stfs	f1,  7 * SIZE(BO)

	stfs	f2,  8 * SIZE(BO)
	stfs	f2,  9 * SIZE(BO)
	stfs	f2, 10 * SIZE(BO)
	stfs	f2, 11 * SIZE(BO)

	stfs	f3, 12 * SIZE(BO)
	stfs	f3, 13 * SIZE(BO)
	stfs	f3, 14 * SIZE(BO)
	stfs	f3, 15 * SIZE(BO)

	addi	B,  B,   4 * SIZE
	addi	BO, BO, 16 * SIZE
	bdnz	LL(122)
	.align 4
	
LL(125):
	andi.	r0, K,  3
	mtspr	CTR, r0
	ble	LL(130)
	.align 4

LL(126):
	lfs	f0,  0 * SIZE(B)

	stfs	f0,  0 * SIZE(BO)
	stfs	f0,  1 * SIZE(BO)
	stfs	f0,  2 * SIZE(BO)
	stfs	f0,  3 * SIZE(BO)

	addi	B,  B,   1 * SIZE
	addi	BO, BO,  4 * SIZE
	bdnz	LL(126)
	.align 4

LL(130):
	mr	CO1, C
	mr	AO, A
	srawi.	I, M,  4
	ble	LL(140)
	.align 4

LL(131):
	la	BO, BUFFER(SP)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvxl	a5, OFFSET_4, AO
	lvxl	a6, OFFSET_5, AO
	lvxl	a7, OFFSET_6, AO
	lvxl	a8, OFFSET_7, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	
	addi	AO, AO, 32 * SIZE
	addi	BO, BO,  8 * SIZE

	vxor	c01, c01, c01
	vxor	c02, c02, c02
	vxor	c03, c03, c03
	vxor	c04, c04, c04

	PREFETCH_C1

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(135)
	.align 4

LL(132):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c02, a2, b1, c02
	vmaddfp	c03, a3, b1, c03
	vmaddfp	c04, a4, b1, c04

	vmaddfp	c01, a5, b2, c01
	vmaddfp	c02, a6, b2, c02
	vmaddfp	c03, a7, b2, c03
	vmaddfp	c04, a8, b2, c04

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvxl	a5, OFFSET_4, AO
	lvxl	a6, OFFSET_5, AO
	lvxl	a7, OFFSET_6, AO
	lvxl	a8, OFFSET_7, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO

	addi	AO, AO, 32 * SIZE
	addi	BO, BO,  8 * SIZE
	bdnz	LL(132)
	.align 4

LL(135):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP
	addi	AO, AO, -32 * SIZE
	addi	BO, BO,  -8 * SIZE
	ble+	LL(138)
	.align 4

LL(136):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c02, a2, b1, c02
	vmaddfp	c03, a3, b1, c03
	vmaddfp	c04, a4, b1, c04

	addi	AO, AO, 16 * SIZE
	addi	BO, BO,  4 * SIZE
	.align 4

LL(138):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	lvx	a5, OFFSET_3, CO1
	lvx	a6, OFFSET_4, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	b3, a4, a5, PERMLSHIFT
	vperm	b4, a5, a6, PERMLSHIFT
	vperm	a7, a6, a6, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2
	vmaddfp	c03, alpha, c03, b3
	vmaddfp	c04, alpha, c04, b4

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, c03, PERMRSHIFT
	vperm	a5, c03, c04, PERMRSHIFT
	vperm	a6, c04, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1
	stvx	a5, OFFSET_3, CO1
	stvx	a6, OFFSET_4, CO1

	addi	CO1, CO1, 16 * SIZE
	addic.	I, I, -1
	bgt+	LL(131)
	.align 4

LL(140):
	andi.	I, M,  8
	la	BO, BUFFER(SP)
	ble	LL(150)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO
	lvxl	a3, OFFSET_2, AO
	lvxl	a4, OFFSET_3, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	
	addi	AO, AO, 16 * SIZE
	addi	BO, BO,  8 * SIZE

	vxor	c01, c01, c01
	vxor	c05, c05, c05
	vxor	c02, c02, c02
	vxor	c06, c06, c06

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(145)
	.align 4

LL(142):
	vmaddfp	c01, a1, b1, c01
	lvxl	a1, OFFSET_0, AO
	vmaddfp	c02, a2, b1, c02
	lvxl	a2, OFFSET_1, AO
	lvx	b1, OFFSET_0, BO

	vmaddfp	c01, a3, b2, c01
 	lvxl	a3, OFFSET_2, AO
	vmaddfp	c02, a4, b2, c02
	lvxl	a4, OFFSET_3, AO

	lvx	b2, OFFSET_1, BO

	addi	AO, AO, 16 * SIZE
	addi	BO, BO,  8 * SIZE
	bdnz	LL(142)
	.align 4

LL(145):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO, -16 * SIZE
	addi	BO, BO,  -8 * SIZE
	ble+	LL(148)
	.align 4

LL(146):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c02, a2, b1, c02

	addi	AO, AO,  8 * SIZE
	addi	BO, BO,  4 * SIZE
	.align 4

LL(148):
	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	lvx	a4, OFFSET_2, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	b2, a3, a4, PERMLSHIFT
	vperm	a7, a4, a4, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1
	vmaddfp	c02, alpha, c02, b2

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, c02, PERMRSHIFT
	vperm	a4, c02, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1
	stvx	a4, OFFSET_2, CO1

	addi	CO1, CO1, 8 * SIZE
	.align 4

LL(150):
	andi.	I, M,  4
	la	BO, BUFFER(SP)
	ble	LL(160)

	lvxl	a1, OFFSET_0, AO
	lvxl	a2, OFFSET_1, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO
	
	addi	AO, AO,  8 * SIZE
	addi	BO, BO,  8 * SIZE

	vxor	c01, c01, c01
	vxor	c02, c02, c02

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(155)
	.align 4

LL(152):
	vmaddfp	c01, a1, b1, c01
	vmaddfp	c02, a2, b2, c02

	lvxl	a1, OFFSET_0, AO
 	lvxl	a2, OFFSET_1, AO

	lvx	b1, OFFSET_0, BO
	lvx	b2, OFFSET_1, BO

	addi	AO, AO,  8 * SIZE
	addi	BO, BO,  8 * SIZE
	bdnz	LL(152)
	.align 4

LL(155):
	andi.	r0,  K,  1
	lvx	alpha, 0, SP

	addi	AO, AO, -8 * SIZE
	addi	BO, BO, -8 * SIZE
	ble+	LL(158)
	.align 4

LL(156):
	vmaddfp	c01, a1, b1, c01

	addi	AO, AO,  4 * SIZE
	addi	BO, BO,  4 * SIZE
	.align 4

LL(158):
	vaddfp	c01, c01, c02

	lvx	a2, OFFSET_0, CO1
	lvx	a3, OFFSET_1, CO1
	
	lvsl	PERMLSHIFT, 0, CO1
	lvsr	PERMRSHIFT, 0, CO1

	vperm	a1, a2, a2, PERMLSHIFT
	vperm	b1, a2, a3, PERMLSHIFT
	vperm	a7, a3, a3, PERMLSHIFT

	vmaddfp	c01, alpha, c01, b1

	vperm	a2, a1,  c01, PERMRSHIFT
	vperm	a3, c01, a7,  PERMRSHIFT

	stvx	a2, OFFSET_0, CO1
	stvx	a3, OFFSET_1, CO1

	addi	CO1, CO1, 4 * SIZE
	.align 4

LL(160):
	andi.	I, M,  2
	la	BO, BUFFER(SP)
	ble	LL(170)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)
	LFD	f10,  2 * SIZE(AO)
	LFD	f11,  3 * SIZE(AO)

	LFD	f12,  0 * SIZE(BO)
	LFD	f13,  4 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0
	fmr	f2,  f0
	fmr	f3,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(165)
	.align 4

LL(162):
	FMADD	f0,  f8,  f12, f0
	FMADD	f1,  f9,  f12, f1
	FMADD	f2,  f10, f13, f2
	FMADD	f3,  f11, f13, f3

	LFD	f8,   4 * SIZE(AO)
	LFD	f9,   5 * SIZE(AO)
	LFD	f10,  6 * SIZE(AO)
	LFD	f11,  7 * SIZE(AO)

	LFD	f12,  8 * SIZE(BO)
	LFD	f13, 12 * SIZE(BO)

	addi	AO, AO,  4 * SIZE
	addi	BO, BO,  8 * SIZE
	bdnz	LL(162)
	.align 4

LL(165):
	andi.	r0,  K,  1
	lfs	f13,  ALPHA(SP)
	ble	LL(168)
	.align 4

LL(166):
	FMADD	f0,  f8, f12, f0
	FMADD	f1,  f9, f12, f1

	addi	AO, AO,  2 * SIZE
	addi	BO, BO,  4 * SIZE
	.align 4

LL(168):
	LFD	f8,  0 * SIZE(CO1)
	LFD	f9,  1 * SIZE(CO1)

	FADD	f0, f0, f2
	FADD	f1, f1, f3

	FMADD	f0,  f0, f13, f8
	FMADD	f1,  f1, f13, f9

	STFD	f0,  0 * SIZE(CO1)
	STFD	f1,  1 * SIZE(CO1)

	addi	CO1, CO1, 2 * SIZE
	.align 4

LL(170):
	andi.	I, M,  1
	la	BO, BUFFER(SP)
	ble	LL(999)

	LFD	f8,   0 * SIZE(AO)
	LFD	f9,   1 * SIZE(AO)

	LFD	f10,  0 * SIZE(BO)
	LFD	f11,  4 * SIZE(BO)

	lfs	f0,  FZERO(SP)
 	fmr	f1,  f0

	srawi.	r0,  K,  1
	mtspr	CTR, r0
	ble	LL(175)
	.align 4

LL(172):
	FMADD	f0,  f8, f10, f0
	FMADD	f1,  f9, f11, f1

	LFD	f8,   2 * SIZE(AO)
	LFD	f9,   3 * SIZE(AO)
	LFD	f10,  8 * SIZE(BO)
	LFD	f11, 12 * SIZE(BO)

	addi	AO, AO,  2 * SIZE
	addi	BO, BO,  8 * SIZE
	bdnz	LL(172)
	.align 4

LL(175):
	andi.	r0,  K,  1
	lfs	f13,  ALPHA(SP)
	ble	LL(178)
	.align 4

LL(176):
	FMADD	f0,  f8, f10, f0

	addi	AO, AO,  1 * SIZE
	addi	BO, BO,  4 * SIZE
	.align 4

LL(178):
	LFD	f8,  0 * SIZE(CO1)

	FADD	f0, f0, f1

	FMADD	f0,  f0, f13, f8

	STFD	f0,  0 * SIZE(CO1)
	.align 4
	
LL(999):
	mr	SP, STACK

	li	r0,  0 * 16
	lvx	v20, SP, r0
	li	r0,  1 * 16
	lvx	v21, SP, r0
	li	r0,  2 * 16
	lvx	v22, SP, r0
	li	r0,  3 * 16
	lvx	v23, SP, r0
	li	r0,  4 * 16
	lvx	v24, SP, r0
	li	r0,  5 * 16
	lvx	v25, SP, r0
	li	r0,  6 * 16
	lvx	v26, SP, r0
	li	r0,  7 * 16
	lvx	v27, SP, r0
	li	r0,  8 * 16
	lvx	v28, SP, r0
	li	r0,  9 * 16
	lvx	v29, SP, r0
	li	r0, 10 * 16
	lvx	v30, SP, r0
	li	r0, 11 * 16
	lvx	v31, SP, r0

	mtspr	VRsave, VREG

#ifdef __64BIT__
	ld	r31,  192(SP)
	ld	r30,  200(SP)
	ld	r29,  208(SP)
	ld	r28,  216(SP)
	ld	r27,  224(SP)
	ld	r26,  232(SP)
	ld	r25,  240(SP)
	ld	r24,  248(SP)
	ld	r23,  256(SP)
	ld	r22,  264(SP)
	ld	r21,  272(SP)
	ld	r20,  280(SP)
	ld	r19,  288(SP)
	ld	r18,  296(SP)
	ld	r17,  304(SP)
	ld	r16,  312(SP)
	ld	r15,  320(SP)
	ld	r14,  328(SP)
#else
	lwz	r31,  192(SP)
	lwz	r30,  196(SP)
	lwz	r29,  200(SP)
	lwz	r28,  204(SP)
	lwz	r27,  208(SP)
	lwz	r26,  212(SP)
	lwz	r25,  216(SP)
	lwz	r24,  220(SP)
	lwz	r23,  224(SP)
	lwz	r22,  228(SP)
	lwz	r21,  232(SP)
	lwz	r20,  236(SP)
	lwz	r19,  240(SP)
	lwz	r18,  244(SP)
	lwz	r17,  248(SP)
	lwz	r16,  252(SP)
	lwz	r15,  256(SP)
	lwz	r14,  260(SP)
#endif

	addi	SP, SP, STACKSIZE

	blr

	EPILOGUE
#endif
