/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#define ASSEMBLER
#include "common.h"
 
#define M	ARG1
#define N	ARG2
#define K	ARG3
#define A	ARG4
#define B	ARG5
#define C	ARG6
#define LDC	%r10
	
#define I	%r11
#define J	%r12
#define AO	%r13
#define BO	%r14
#define	CO1	%r15
#define CO2	%rbp
/* rbp is reserved */

#ifdef OPTERON
#define PREFETCH     prefetch
#define PREFETCHSIZE (8 * 2 + 4)
#endif

#define LOCAL_BUFFER_SIZE  GEMM_Q * GEMM_UNROLL_N * COMPSIZE * 16

#define STACKSIZE 64

#ifdef OPTERON	
#define movsd	movlpd
#endif

#define ALPHA	  0(%rsp)
#define OFFSET	 16(%rsp)
#define KK	 24(%rsp)
#define KKK	 32(%rsp)
#define BUFFER	128(%rsp)

#define KERNEL(xx) \
	mulpd	%xmm8, %xmm9 ;\
	addpd	%xmm9, %xmm0 ;\
	movapd	 0 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulpd	%xmm8, %xmm11 ;\
	addpd	%xmm11, %xmm1 ;\
	movapd	 2 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm8, %xmm13 ;\
	mulpd	 6 * SIZE + 2 * (xx) * SIZE(BO), %xmm8 ;\
	addpd	%xmm13, %xmm2 ;\
	movapd	 4 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm8, %xmm3 ;\
	movapd	 8 * SIZE + 1 * (xx) * SIZE(AO), %xmm8 ;\
	mulpd	%xmm10, %xmm9 ;\
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE + 1 * (xx) * SIZE(AO) ;\
	addpd	%xmm9, %xmm4 ;\
	movapd	16 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulpd	%xmm10, %xmm11 ;\
	addpd	%xmm11, %xmm5 ;\
	movapd	10 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm10, %xmm13 ;\
	mulpd	 6 * SIZE + 2 * (xx) * SIZE(BO), %xmm10 ;\
	addpd	%xmm13, %xmm6 ;\
	movapd	12 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm10, %xmm7 ;\
	movapd	10 * SIZE + 1 * (xx) * SIZE(AO), %xmm10 ;\
	mulpd	%xmm12, %xmm15 ;\
	addpd	%xmm15, %xmm0 ;\
	movapd	 8 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulpd	%xmm12, %xmm11 ;\
	addpd	%xmm11, %xmm1 ;\
	movapd	10 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm12, %xmm13 ;\
	mulpd	14 * SIZE + 2 * (xx) * SIZE(BO), %xmm12 ;\
	addpd	%xmm13, %xmm2 ;\
	movapd	12 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm12, %xmm3 ;\
	movapd	12 * SIZE + 1 * (xx) * SIZE(AO), %xmm12 ;\
	mulpd	%xmm14, %xmm15 ;\
	addpd	%xmm15, %xmm4 ;\
	movapd	24 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulpd	%xmm14, %xmm11 ;\
	addpd	%xmm11, %xmm5 ;\
	movapd	18 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm14, %xmm13 ;\
	mulpd	14 * SIZE + 2 * (xx) * SIZE(BO), %xmm14 ;\
	addpd	%xmm13, %xmm6 ;\
	movapd	20 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm14, %xmm7 ;\
	movapd	14 * SIZE + 1 * (xx) * SIZE(AO), %xmm14 ;\
	mulpd	%xmm8, %xmm9 ;\
	addpd	%xmm9, %xmm0 ;\
	movapd	16 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulpd	%xmm8, %xmm11 ;\
	PREFETCH	(PREFETCHSIZE     +  8) * SIZE + 1 * (xx) * SIZE(AO) ;\
	addpd	%xmm11, %xmm1 ;\
	movapd	18 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm8, %xmm13 ;\
	mulpd	22 * SIZE + 2 * (xx) * SIZE(BO), %xmm8 ;\
	addpd	%xmm13, %xmm2 ;\
	movapd	20 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm8, %xmm3 ;\
	movapd	16 * SIZE + 1 * (xx) * SIZE(AO), %xmm8 ;\
	mulpd	%xmm10, %xmm9 ;\
	addpd	%xmm9, %xmm4 ;\
	movapd	32 * SIZE + 2 * (xx) * SIZE(BO), %xmm9 ;\
	mulpd	%xmm10, %xmm11 ;\
	addpd	%xmm11, %xmm5 ;\
	movapd	26 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm10, %xmm13 ;\
	mulpd	22 * SIZE + 2 * (xx) * SIZE(BO), %xmm10 ;\
	addpd	%xmm13, %xmm6 ;\
	movapd	28 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm10, %xmm7 ;\
	movapd	18 * SIZE + 1 * (xx) * SIZE(AO), %xmm10 ;\
	mulpd	%xmm12, %xmm15 ;\
	addpd	%xmm15, %xmm0 ;\
	movapd	24 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulpd	%xmm12, %xmm11 ;\
	addpd	%xmm11, %xmm1 ;\
	movapd	26 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm12, %xmm13 ;\
	mulpd	30 * SIZE + 2 * (xx) * SIZE(BO), %xmm12 ;\
	addpd	%xmm13, %xmm2 ;\
	movapd	28 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm12, %xmm3 ;\
	movapd	20 * SIZE + 1 * (xx) * SIZE(AO), %xmm12 ;\
	mulpd	%xmm14, %xmm15 ;\
	addpd	%xmm15, %xmm4 ;\
	movapd	40 * SIZE + 2 * (xx) * SIZE(BO), %xmm15 ;\
	mulpd	%xmm14, %xmm11 ;\
	addpd	%xmm11, %xmm5 ;\
	movapd	34 * SIZE + 2 * (xx) * SIZE(BO), %xmm11 ;\
	mulpd	%xmm14, %xmm13 ;\
	mulpd	30 * SIZE + 2 * (xx) * SIZE(BO), %xmm14 ;\
	addpd	%xmm13, %xmm6 ;\
	movapd	36 * SIZE + 2 * (xx) * SIZE(BO), %xmm13 ;\
	addpd	%xmm14, %xmm7 ;\
	movapd	22 * SIZE + 1 * (xx) * SIZE(AO), %xmm14
	
	PROLOGUE
	PROFCODE
	
	subq	$STACKSIZE, %rsp
	movq	%rbx,  0(%rsp)
	movq	%rbp,  8(%rsp)
	movq	%r12, 16(%rsp)
	movq	%r13, 24(%rsp)
	movq	%r14, 32(%rsp)
	movq	%r15, 40(%rsp)

	movq	72(%rsp), LDC	# LDC
#ifdef TRMMKERNEL
	movsd	80(%rsp), %xmm4
#endif

	EMMS

	movq	%rsp, %rbx	# save old stack
	subq	$128 + LOCAL_BUFFER_SIZE, %rsp
	andq	$-4096, %rsp	# align stack

	unpcklpd %xmm0, %xmm0
	movapd	 %xmm0, ALPHA

	leaq	(, LDC, SIZE), LDC

#ifdef TRMMKERNEL
	movsd	%xmm4, OFFSET
	movsd	%xmm4, KK
#ifndef LEFT
	negq	KK
#endif	
#endif
	movq	N,  J
	sarq	$2, J		# j = (n >> 2)
	jle	.L40

.L01:
/* Copying to Sub Buffer */
	leaq	16 * SIZE + BUFFER, BO
	movq	C, CO1			# coffset1 = c
	leaq	(C, LDC, 1), CO2	# coffset2 = c + ldc
	
#if defined(TRMMKERNEL) && defined(LEFT)
	movq	OFFSET, %rax
	movq	%rax, KK
#endif	

	movq	K, %rax
	sarq	$2, %rax
	jle	.L03
	.align 8
	
#define RPREFETCHSIZE 48

.L02:
	prefetchnta	 (RPREFETCHSIZE +  0)  * SIZE(B)

	movq	 0 * SIZE(B), %mm0
	movq	%mm0,  -16 * SIZE(BO)
	movq	%mm0,  -15 * SIZE(BO)
	movq	 1 * SIZE(B), %mm1
	movq	%mm1,  -14 * SIZE(BO)
	movq	%mm1,  -13 * SIZE(BO)
	movq	 2 * SIZE(B), %mm2
	movq	%mm2,  -12 * SIZE(BO)
	movq	%mm2,  -11 * SIZE(BO)
	movq	 3 * SIZE(B), %mm3
	movq	%mm3,  -10 * SIZE(BO)
	movq	%mm3,   -9 * SIZE(BO)
	movq	 4 * SIZE(B), %mm4
	movq	%mm4,   -8 * SIZE(BO)
	movq	%mm4,   -7 * SIZE(BO)
	movq	 5 * SIZE(B), %mm5
	movq	%mm5,   -6 * SIZE(BO)
	movq	%mm5,   -5 * SIZE(BO)
	movq	 6 * SIZE(B), %mm6
	movq	%mm6,   -4 * SIZE(BO)
	movq	%mm6,   -3 * SIZE(BO)
	movq	 7 * SIZE(B), %mm7
	movq	%mm7,   -2 * SIZE(BO)
	movq	%mm7,   -1 * SIZE(BO)

	prefetchnta	 (RPREFETCHSIZE +  8)  * SIZE(B)

	movq	 8 * SIZE(B), %mm0
	movq	%mm0,   0 * SIZE(BO)
	movq	%mm0,   1 * SIZE(BO)
	movq	 9 * SIZE(B), %mm1
	movq	%mm1,   2 * SIZE(BO)
	movq	%mm1,   3 * SIZE(BO)
	movq	10 * SIZE(B), %mm2
	movq	%mm2,   4 * SIZE(BO)
	movq	%mm2,   5 * SIZE(BO)
	movq	11 * SIZE(B), %mm3
	movq	%mm3,   6 * SIZE(BO)
	movq	%mm3,   7 * SIZE(BO)
	movq	12 * SIZE(B), %mm4
	movq	%mm4,   8 * SIZE(BO)
	movq	%mm4,   9 * SIZE(BO)
	movq	13 * SIZE(B), %mm5
	movq	%mm5,  10 * SIZE(BO)
	movq	%mm5,  11 * SIZE(BO)
	movq	14 * SIZE(B), %mm6
	movq	%mm6,  12 * SIZE(BO)
	movq	%mm6,  13 * SIZE(BO)
	movq	15 * SIZE(B), %mm7
	movq	%mm7,  14 * SIZE(BO)
	movq	%mm7,  15 * SIZE(BO)

	subq	$-32 * SIZE, BO
	addq	$ 16 * SIZE, B

	decq	%rax
	jne	.L02
	.align 8

.L03:
	movq	K, %rax
	andq	$3, %rax
	BRANCH
	jle	.L10
	.align 8

.L04:
	movq	 0 * SIZE(B), %mm0
	movq	 1 * SIZE(B), %mm1
	movq	 2 * SIZE(B), %mm2
	movq	 3 * SIZE(B), %mm3

	movq	%mm0, -16 * SIZE(BO)
	movq	%mm0, -15 * SIZE(BO)
	movq	%mm1, -14 * SIZE(BO)
	movq	%mm1, -13 * SIZE(BO)
	movq	%mm2, -12 * SIZE(BO)
	movq	%mm2, -11 * SIZE(BO)
	movq	%mm3, -10 * SIZE(BO)
	movq	%mm3,  -9 * SIZE(BO)

	addq	$4 * SIZE, B
	addq	$8 * SIZE, BO
	decq	%rax
	jne	.L04
	.align 8
	
.L10:
	movq	A, AO		# aoffset = a

	movq	M,  I
	sarq	$2, I	# i = (m >> 2)
	jle	.L20
	.align 8

.L11:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movapd	 0 * SIZE(BO), %xmm9
	movapd	 2 * SIZE(BO), %xmm11
	movapd	 4 * SIZE(BO), %xmm13
	movapd	 8 * SIZE(BO), %xmm15

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 2 * SIZE(AO), %xmm10
	pxor	%xmm1, %xmm1
	movapd	 4 * SIZE(AO), %xmm12
	pxor	%xmm2, %xmm2
	movapd	 6 * SIZE(AO), %xmm14
	pxor	%xmm3, %xmm3

	prefetchw      4 * SIZE(CO1)
	pxor	%xmm4, %xmm4
	prefetchw      4 * SIZE(CO2)
	pxor	%xmm5, %xmm5
	prefetchw      4 * SIZE(CO1, LDC, 2)
	pxor	%xmm6, %xmm6
	prefetchw      4 * SIZE(CO2, LDC, 2)
	pxor	%xmm7, %xmm7

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	andq	$-8, %rax
	salq	$4, %rax
	je	.L15

.L1X:	
	KERNEL(16 *  0)		# 0
	KERNEL(16 *  1)		# 1
	cmpq	$64 *  2, %rax
	jle	.L12
	KERNEL(16 *  2)		# 2
	KERNEL(16 *  3)		# 3
	cmpq	$64 *  4, %rax
	jle	.L12
	KERNEL(16 *  4)		# 4
	KERNEL(16 *  5)		# 5
	cmpq	$64 *  6, %rax
	jle	.L12
	KERNEL(16 *  6)		# 6
	KERNEL(16 *  7)		# 7
	cmpq	$64 *  8, %rax
	jle	.L12
	KERNEL(16 *  8)		# 8
	KERNEL(16 *  9)		# 9
	cmpq	$64 * 10, %rax
	jle	.L12
	KERNEL(16 * 10)		# 10
	KERNEL(16 * 11)		# 11
	cmpq	$64 * 12, %rax
	jle	.L12
	KERNEL(16 * 12)		# 12
	KERNEL(16 * 13)		# 13
	cmpq	$64 * 14, %rax
	jle	.L12
	KERNEL(16 * 14)		# 14
	KERNEL(16 * 15)		# 15
	addq	$32 * 8  * SIZE, AO
	addq	$64 * 8  * SIZE, BO
	subq	$128 * 8, %rax
	jg	.L1X

.L12:
	leaq	(AO, %rax, 2), AO	# * 16
	leaq	(BO, %rax, 4), BO	# * 64
	.align 8

.L15:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L19
	.align 8

.L16:
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm0
	movapd	 2 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm1
	movapd	 4 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm2
	movapd	 0 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm3
	movapd	 4 * SIZE(AO), %xmm8
	mulpd	%xmm10, %xmm9
	addpd	%xmm9, %xmm4
	movapd	 2 * SIZE(BO), %xmm9
	mulpd	%xmm10, %xmm9
	addpd	%xmm9, %xmm5
	movapd	 4 * SIZE(BO), %xmm9
	mulpd	%xmm10, %xmm9
	mulpd	 6 * SIZE(BO), %xmm10
	addpd	%xmm9, %xmm6
	movapd	 8 * SIZE(BO), %xmm9
	addpd	%xmm10, %xmm7
	movapd	 6 * SIZE(AO), %xmm10

	addq	$4 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L16
	.align 8

.L19:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8
	movsd	2 * SIZE(CO1), %xmm9
	movhpd	3 * SIZE(CO1), %xmm9

	movsd	0 * SIZE(CO2), %xmm10
	movhpd	1 * SIZE(CO2), %xmm10
	movsd	2 * SIZE(CO2), %xmm11
	movhpd	3 * SIZE(CO2), %xmm11
#endif

	mulpd	%xmm15, %xmm0
	mulpd	%xmm15, %xmm1
	mulpd	%xmm15, %xmm2
	mulpd	%xmm15, %xmm3

	mulpd	%xmm15, %xmm4
	mulpd	%xmm15, %xmm5
	mulpd	%xmm15, %xmm6
	mulpd	%xmm15, %xmm7

#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1, LDC, 2), %xmm12
	movhpd	1 * SIZE(CO1, LDC, 2), %xmm12
	movsd	2 * SIZE(CO1, LDC, 2), %xmm13
	movhpd	3 * SIZE(CO1, LDC, 2), %xmm13

	movsd	0 * SIZE(CO2, LDC, 2), %xmm14
	movhpd	1 * SIZE(CO2, LDC, 2), %xmm14
	movsd	2 * SIZE(CO2, LDC, 2), %xmm15
	movhpd	3 * SIZE(CO2, LDC, 2), %xmm15

	addpd	%xmm8,  %xmm0
	addpd	%xmm9,  %xmm4
	addpd	%xmm10, %xmm1
	addpd	%xmm11, %xmm5
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	movsd	%xmm4, 2 * SIZE(CO1)
	movhpd	%xmm4, 3 * SIZE(CO1)

	movsd	%xmm1, 0 * SIZE(CO2)
	movhpd	%xmm1, 1 * SIZE(CO2)
	movsd	%xmm5, 2 * SIZE(CO2)
	movhpd	%xmm5, 3 * SIZE(CO2)

#ifndef TRMMKERNEL
	addpd	%xmm12, %xmm2
	addpd	%xmm13, %xmm6
	addpd	%xmm14, %xmm3
	addpd	%xmm15, %xmm7
#endif

	movsd	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movhpd	%xmm2, 1 * SIZE(CO1, LDC, 2)
	movsd	%xmm6, 2 * SIZE(CO1, LDC, 2)
	movhpd	%xmm6, 3 * SIZE(CO1, LDC, 2)

	movsd	%xmm3, 0 * SIZE(CO2, LDC, 2)
	movhpd	%xmm3, 1 * SIZE(CO2, LDC, 2)
	movsd	%xmm7, 2 * SIZE(CO2, LDC, 2)
	movhpd	%xmm7, 3 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	addq	$4 * SIZE, CO2		# coffset += 4
	decq	I			# i --
	jg	.L11
	.align 8	

.L20:
	testq	$3, M
	je	.L39

	testq	$2, M
	je	.L30
	.align 8

.L21:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movapd	 8 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movapd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

	movapd	16 * SIZE(BO), %xmm13
	movapd	24 * SIZE(BO), %xmm15

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L25
	.align 8

.L22:
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm0
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	movapd	 2 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm1
	movapd	 4 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm2
	movapd	32 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm3
	movapd	 2 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm11
	addpd	%xmm11, %xmm0
	movapd	10 * SIZE(BO), %xmm11
	mulpd	%xmm8, %xmm11
	addpd	%xmm11, %xmm1
	movapd	12 * SIZE(BO), %xmm11
	mulpd	%xmm8, %xmm11
	mulpd	14 * SIZE(BO), %xmm8
	addpd	%xmm11, %xmm2
	movapd	40 * SIZE(BO), %xmm11
	addpd	%xmm8, %xmm3
	movapd	 4 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm13
	addpd	%xmm13, %xmm0
	movapd	18 * SIZE(BO), %xmm13
	mulpd	%xmm8, %xmm13
	addpd	%xmm13, %xmm1
	movapd	20 * SIZE(BO), %xmm13
	mulpd	%xmm8, %xmm13
	mulpd	22 * SIZE(BO), %xmm8
	addpd	%xmm13, %xmm2
	movapd	48 * SIZE(BO), %xmm13
	addpd	%xmm8, %xmm3
	movapd	 6 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm15
	addpd	%xmm15, %xmm0
	movapd	26 * SIZE(BO), %xmm15
	mulpd	%xmm8, %xmm15
	addpd	%xmm15, %xmm1
	movapd	28 * SIZE(BO), %xmm15
	mulpd	%xmm8, %xmm15
	mulpd	30 * SIZE(BO), %xmm8
	addpd	%xmm15, %xmm2
	movapd	56 * SIZE(BO), %xmm15
	addpd	%xmm8, %xmm3
	movapd	16 * SIZE(AO), %xmm8

	PREFETCH	(PREFETCHSIZE     +  8) * SIZE(AO)
	mulpd	%xmm10, %xmm9
	addpd	%xmm9, %xmm0
	movapd	34 * SIZE(BO), %xmm9
	mulpd	%xmm10, %xmm9
	addpd	%xmm9, %xmm1
	movapd	36 * SIZE(BO), %xmm9
	mulpd	%xmm10, %xmm9
	mulpd	38 * SIZE(BO), %xmm10
	addpd	%xmm9, %xmm2
	movapd	64 * SIZE(BO), %xmm9
	addpd	%xmm10, %xmm3
	movapd	10 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm11
	addpd	%xmm11, %xmm0
	movapd	42 * SIZE(BO), %xmm11
	mulpd	%xmm10, %xmm11
	addpd	%xmm11, %xmm1
	movapd	44 * SIZE(BO), %xmm11
	mulpd	%xmm10, %xmm11
	mulpd	46 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm2
	movapd	72 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm3
	movapd	12 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm13
	addpd	%xmm13, %xmm0
	movapd	50 * SIZE(BO), %xmm13
	mulpd	%xmm10, %xmm13
	addpd	%xmm13, %xmm1
	movapd	52 * SIZE(BO), %xmm13
	mulpd	%xmm10, %xmm13
	mulpd	54 * SIZE(BO), %xmm10
	addpd	%xmm13, %xmm2
	movapd	80 * SIZE(BO), %xmm13
	addpd	%xmm10, %xmm3
	movapd	14 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm15
	addpd	%xmm15, %xmm0
	movapd	58 * SIZE(BO), %xmm15
	mulpd	%xmm10, %xmm15
	addpd	%xmm15, %xmm1
	movapd	60 * SIZE(BO), %xmm15
	mulpd	%xmm10, %xmm15
	mulpd	62 * SIZE(BO), %xmm10
	addpd	%xmm15, %xmm2
	movapd	88 * SIZE(BO), %xmm15
	addpd	%xmm10, %xmm3
	movapd	24 * SIZE(AO), %xmm10

	addq   $16 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L22
	.align 8

.L25:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L29
	.align 8

.L26:
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm0
	movapd	 2 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm1
	movapd	 4 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm2
	movapd	 8 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm3
	movapd	 2 * SIZE(AO), %xmm8

	addq	$2 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L26
	.align 8

.L29:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8
	movsd	0 * SIZE(CO2), %xmm10
	movhpd	1 * SIZE(CO2), %xmm10

	movsd	0 * SIZE(CO1, LDC, 2), %xmm12
	movhpd	1 * SIZE(CO1, LDC, 2), %xmm12
	movsd	0 * SIZE(CO2, LDC, 2), %xmm14
	movhpd	1 * SIZE(CO2, LDC, 2), %xmm14
#endif
	mulpd	%xmm15, %xmm0
	mulpd	%xmm15, %xmm1
	mulpd	%xmm15, %xmm2
	mulpd	%xmm15, %xmm3

#ifndef TRMMKERNEL
	addpd	%xmm8,  %xmm0
	addpd	%xmm10, %xmm1
	addpd	%xmm12, %xmm2
	addpd	%xmm14, %xmm3
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	movsd	%xmm1, 0 * SIZE(CO2)
	movhpd	%xmm1, 1 * SIZE(CO2)
	movsd	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movhpd	%xmm2, 1 * SIZE(CO1, LDC, 2)
	movsd	%xmm3, 0 * SIZE(CO2, LDC, 2)
	movhpd	%xmm3, 1 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif

	addq	$2 * SIZE, CO1		# coffset += 4
	addq	$2 * SIZE, CO2		# coffset += 4
	.align 8	

.L30:
	testq	$1, M
	je	.L39
	.align 8

.L31:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif	

	movsd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movsd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movsd	 8 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movsd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

	movsd	16 * SIZE(BO), %xmm13
	movsd	24 * SIZE(BO), %xmm15

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$4, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L35
	.align 8

.L32:
	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm0
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	movsd	 2 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm1
	movsd	 4 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	mulsd	 6 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm2
	movsd	32 * SIZE(BO), %xmm9
	addsd	%xmm8, %xmm3
	movsd	 1 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm11
	addsd	%xmm11, %xmm0
	movsd	10 * SIZE(BO), %xmm11
	mulsd	%xmm8, %xmm11
	addsd	%xmm11, %xmm1
	movsd	12 * SIZE(BO), %xmm11
	mulsd	%xmm8, %xmm11
	mulsd	14 * SIZE(BO), %xmm8
	addsd	%xmm11, %xmm2
	movsd	40 * SIZE(BO), %xmm11
	addsd	%xmm8, %xmm3
	movsd	 2 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm13
	addsd	%xmm13, %xmm0
	movsd	18 * SIZE(BO), %xmm13
	mulsd	%xmm8, %xmm13
	addsd	%xmm13, %xmm1
	movsd	20 * SIZE(BO), %xmm13
	mulsd	%xmm8, %xmm13
	mulsd	22 * SIZE(BO), %xmm8
	addsd	%xmm13, %xmm2
	movsd	48 * SIZE(BO), %xmm13
	addsd	%xmm8, %xmm3
	movsd	 3 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm15
	addsd	%xmm15, %xmm0
	movsd	26 * SIZE(BO), %xmm15
	mulsd	%xmm8, %xmm15
	addsd	%xmm15, %xmm1
	movsd	28 * SIZE(BO), %xmm15
	mulsd	%xmm8, %xmm15
	mulsd	30 * SIZE(BO), %xmm8
	addsd	%xmm15, %xmm2
	movsd	56 * SIZE(BO), %xmm15
	addsd	%xmm8, %xmm3
	movsd	 4 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm0
	movsd	34 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm1
	movsd	36 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	mulsd	38 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm2
	movsd	64 * SIZE(BO), %xmm9
	addsd	%xmm8, %xmm3
	movsd	 5 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm11
	addsd	%xmm11, %xmm0
	movsd	42 * SIZE(BO), %xmm11
	mulsd	%xmm8, %xmm11
	addsd	%xmm11, %xmm1
	movsd	44 * SIZE(BO), %xmm11
	mulsd	%xmm8, %xmm11
	mulsd	46 * SIZE(BO), %xmm8
	addsd	%xmm11, %xmm2
	movsd	72 * SIZE(BO), %xmm11
	addsd	%xmm8, %xmm3
	movsd	 6 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm13
	addsd	%xmm13, %xmm0
	movsd	50 * SIZE(BO), %xmm13
	mulsd	%xmm8, %xmm13
	addsd	%xmm13, %xmm1
	movsd	52 * SIZE(BO), %xmm13
	mulsd	%xmm8, %xmm13
	mulsd	54 * SIZE(BO), %xmm8
	addsd	%xmm13, %xmm2
	movsd	80 * SIZE(BO), %xmm13
	addsd	%xmm8, %xmm3
	movsd	 7 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm15
	addsd	%xmm15, %xmm0
	movsd	58 * SIZE(BO), %xmm15
	mulsd	%xmm8, %xmm15
	addsd	%xmm15, %xmm1
	movsd	60 * SIZE(BO), %xmm15
	mulsd	%xmm8, %xmm15
	mulsd	62 * SIZE(BO), %xmm8
	addsd	%xmm15, %xmm2
	movsd	88 * SIZE(BO), %xmm15
	addsd	%xmm8, %xmm3
	movsd	 8 * SIZE(AO), %xmm8

	addq   $ 8 * SIZE, AO
	addq   $64 * SIZE, BO
	decq   %rax
	jne    .L32
	.align 8

.L35:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movsd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L38
	.align 8

.L36:
	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm0
	movsd	 2 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	addsd	%xmm9, %xmm1
	movsd	 4 * SIZE(BO), %xmm9
	mulsd	%xmm8, %xmm9
	mulsd	 6 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm2
	movsd	 8 * SIZE(BO), %xmm9
	addsd	%xmm8, %xmm3
	movsd	 1 * SIZE(AO), %xmm8

	addq	$1 * SIZE, AO		# aoffset  += 4
	addq	$8 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L36
	.align 8

.L38:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movsd	0 * SIZE(CO2), %xmm10
	movsd	0 * SIZE(CO1, LDC, 2), %xmm12
	movsd	0 * SIZE(CO2, LDC, 2), %xmm14
#endif

	mulsd	%xmm15, %xmm0
	mulsd	%xmm15, %xmm1
	mulsd	%xmm15, %xmm2
	mulsd	%xmm15, %xmm3

#ifndef TRMMKERNEL
	addsd	%xmm8,  %xmm0
	addsd	%xmm10, %xmm1
	addsd	%xmm12, %xmm2
	addsd	%xmm14, %xmm3
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movsd	%xmm1, 0 * SIZE(CO2)
	movsd	%xmm2, 0 * SIZE(CO1, LDC, 2)
	movsd	%xmm3, 0 * SIZE(CO2, LDC, 2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 8), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$1, KK
#endif
	.align 8	
	
.L39:
#if defined(TRMMKERNEL) && !defined(LEFT)
	addl	$4, KK
#endif

	leaq	(C, LDC, 4), C		# c += 4 * ldc
	decq	J			# j --
	jg	.L01
	.align 8

.L40:
	testq	$3, N
	je	.L999

	testq	$2, N
	je	.L80
	.align 32

.L41:
/* Copying to Sub Buffer */
	leaq	BUFFER, BO
	
#if defined(TRMMKERNEL) && defined(LEFT)
	movq	OFFSET, %rax
	movq	%rax, KK
#endif	

	movq	K, %rax
	sarq	$2, %rax
	jle	.L43
	.align 8
	
.L42:
	prefetch	 56 * SIZE(B)

	movq	 0 * SIZE(B), %mm0
	movq	 1 * SIZE(B), %mm1
	movq	 2 * SIZE(B), %mm2
	movq	 3 * SIZE(B), %mm3
	movq	 4 * SIZE(B), %mm4
	movq	 5 * SIZE(B), %mm5
	movq	 6 * SIZE(B), %mm6
	movq	 7 * SIZE(B), %mm7

	addq	$ 8 * SIZE, B
	addq	$16 * SIZE, BO

	movq	%mm0, -16 * SIZE(BO)
	movq	%mm0, -15 * SIZE(BO)
	movq	%mm1, -14 * SIZE(BO)
	movq	%mm1, -13 * SIZE(BO)
	movq	%mm2, -12 * SIZE(BO)
	movq	%mm2, -11 * SIZE(BO)
	movq	%mm3, -10 * SIZE(BO)
	movq	%mm3,  -9 * SIZE(BO)
	movq	%mm4,  -8 * SIZE(BO)
	movq	%mm4,  -7 * SIZE(BO)
	movq	%mm5,  -6 * SIZE(BO)
	movq	%mm5,  -5 * SIZE(BO)
	movq	%mm6,  -4 * SIZE(BO)
	movq	%mm6,  -3 * SIZE(BO)
	movq	%mm7,  -2 * SIZE(BO)
	movq	%mm7,  -1 * SIZE(BO)

	decq	%rax
	jne	.L42
	.align 8

.L43:
	movq	K, %rax
	andq	$3, %rax
	BRANCH
	jle	.L50
	.align 8

.L44:
	movq	 0 * SIZE(B), %mm0
	movq	 1 * SIZE(B), %mm1

	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  1 * SIZE(BO)
	movq	%mm1,  2 * SIZE(BO)
	movq	%mm1,  3 * SIZE(BO)

	addq	$2 * SIZE, B
	addq	$4 * SIZE, BO
	decq	%rax
	jne	.L44
	.align 8
	
.L50:
	movq	C, CO1			# coffset1 = c
	leaq	(C, LDC, 1), CO2	# coffset2 = c + ldc
	movq	A, AO		# aoffset = a

	movq	M,  I
	sarq	$2, I	# i = (m >> 2)
	jle	.L60
	.align 8

.L51:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movapd	 8 * SIZE(AO), %xmm10
	pxor	%xmm4, %xmm4
	movapd	 8 * SIZE(BO), %xmm11
	pxor	%xmm5, %xmm5

	movapd	16 * SIZE(AO), %xmm12
	movapd	16 * SIZE(BO), %xmm13
	movapd	24 * SIZE(AO), %xmm14
	movapd	24 * SIZE(BO), %xmm15

	prefetchw      4 * SIZE(CO1)
	prefetchw      4 * SIZE(CO2)

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L55
	.align 8

.L52:
	mulpd	%xmm8, %xmm9
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm1
	movapd	 2 * SIZE(AO), %xmm8
	mulpd	%xmm8, %xmm9
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm4
	movapd	 4 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm5
	movapd	 4 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm0
	movapd	 4 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm1
	movapd	 6 * SIZE(AO), %xmm8
	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm4
	movapd	32 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm5
	movapd	32 * SIZE(AO), %xmm8

	PREFETCH	(PREFETCHSIZE     +  8) * SIZE(AO)
	mulpd	%xmm10, %xmm11
	mulpd	10 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm0
	movapd	 8 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm1
	movapd	10 * SIZE(AO), %xmm10
	mulpd	%xmm10, %xmm11
	mulpd	10 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm4
	movapd	12 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm5
	movapd	12 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm11
	mulpd	14 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm0
	movapd	12 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm1
	movapd	14 * SIZE(AO), %xmm10
	mulpd	%xmm10, %xmm11
	mulpd	14 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm4
	movapd	40 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm5
	movapd	40 * SIZE(AO), %xmm10

	PREFETCH	(PREFETCHSIZE     + 16) * SIZE(AO)
	mulpd	%xmm12, %xmm13
	mulpd	18 * SIZE(BO), %xmm12
	addpd	%xmm13, %xmm0
	movapd	16 * SIZE(BO), %xmm13
	addpd	%xmm12, %xmm1
	movapd	18 * SIZE(AO), %xmm12
	mulpd	%xmm12, %xmm13
	mulpd	18 * SIZE(BO), %xmm12
	addpd	%xmm13, %xmm4
	movapd	20 * SIZE(BO), %xmm13
	addpd	%xmm12, %xmm5
	movapd	20 * SIZE(AO), %xmm12

	mulpd	%xmm12, %xmm13
	mulpd	22 * SIZE(BO), %xmm12
	addpd	%xmm13, %xmm0
	movapd	20 * SIZE(BO), %xmm13
	addpd	%xmm12, %xmm1
	movapd	22 * SIZE(AO), %xmm12
	mulpd	%xmm12, %xmm13
	mulpd	22 * SIZE(BO), %xmm12
	addpd	%xmm13, %xmm4
	movapd	48 * SIZE(BO), %xmm13
	addpd	%xmm12, %xmm5
	movapd	48 * SIZE(AO), %xmm12

	PREFETCH	(PREFETCHSIZE     + 24) * SIZE(AO)
	mulpd	%xmm14, %xmm15
	mulpd	26 * SIZE(BO), %xmm14
	addpd	%xmm15, %xmm0
	movapd	24 * SIZE(BO), %xmm15
	addpd	%xmm14, %xmm1
	movapd	26 * SIZE(AO), %xmm14
	mulpd	%xmm14, %xmm15
	mulpd	26 * SIZE(BO), %xmm14
	addpd	%xmm15, %xmm4
	movapd	28 * SIZE(BO), %xmm15
	addpd	%xmm14, %xmm5
	movapd	28 * SIZE(AO), %xmm14

	mulpd	%xmm14, %xmm15
	mulpd	30 * SIZE(BO), %xmm14
	addpd	%xmm15, %xmm0
	movapd	28 * SIZE(BO), %xmm15
	addpd	%xmm14, %xmm1
	movapd	30 * SIZE(AO), %xmm14
	mulpd	%xmm14, %xmm15
	mulpd	30 * SIZE(BO), %xmm14
	addpd	%xmm15, %xmm4
	movapd	56 * SIZE(BO), %xmm15
	addpd	%xmm14, %xmm5
	movapd	56 * SIZE(AO), %xmm14

	addq   $32 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L52
	.align 8

.L55:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L59
	.align 8

.L56:
	movapd	 0 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm0
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm8, %xmm1
	movapd	 2 * SIZE(AO), %xmm8
	movapd	 0 * SIZE(BO), %xmm9
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm4
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm8, %xmm5
	movapd	 4 * SIZE(AO), %xmm8

	addq	$4 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L56
	.align 8

.L59:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8
	movsd	2 * SIZE(CO1), %xmm9
	movhpd	3 * SIZE(CO1), %xmm9
	movsd	0 * SIZE(CO2), %xmm10
	movhpd	1 * SIZE(CO2), %xmm10
	movsd	2 * SIZE(CO2), %xmm11
	movhpd	3 * SIZE(CO2), %xmm11
#endif

	mulpd	%xmm15, %xmm0
	mulpd	%xmm15, %xmm1
	mulpd	%xmm15, %xmm4
	mulpd	%xmm15, %xmm5

#ifndef TRMMKERNEL
	addpd	%xmm8,  %xmm0
	addpd	%xmm9,  %xmm4
	addpd	%xmm10, %xmm1
	addpd	%xmm11, %xmm5
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	movsd	%xmm4, 2 * SIZE(CO1)
	movhpd	%xmm4, 3 * SIZE(CO1)
	movsd	%xmm1, 0 * SIZE(CO2)
	movhpd	%xmm1, 1 * SIZE(CO2)
	movsd	%xmm5, 2 * SIZE(CO2)
	movhpd	%xmm5, 3 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	addq	$4 * SIZE, CO2		# coffset += 4
	decq	I			# i --
	jg	.L51
	.align 8	

.L60:
	testq	$2, M
	je	.L70
	.align 8

.L61:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movapd	 8 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movapd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

	movapd	16 * SIZE(BO), %xmm13
	movapd	24 * SIZE(BO), %xmm15

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L65
	.align 8

.L62:
	mulpd	%xmm8, %xmm9
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm0
	movapd	 4 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm1
	movapd	 2 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm9
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm2
	movapd	32 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm3
	movapd	 4 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm11
	mulpd	10 * SIZE(BO), %xmm8
	addpd	%xmm11, %xmm0
	movapd	12 * SIZE(BO), %xmm11
	addpd	%xmm8, %xmm1
	movapd	 6 * SIZE(AO), %xmm8

	mulpd	%xmm8, %xmm11
	mulpd	14 * SIZE(BO), %xmm8
	addpd	%xmm11, %xmm2
	movapd	40 * SIZE(BO), %xmm11
	addpd	%xmm8, %xmm3
	movapd	16 * SIZE(AO), %xmm8

	PREFETCH	(PREFETCHSIZE     +  8) * SIZE(AO)
	mulpd	%xmm10, %xmm13
	mulpd	18 * SIZE(BO), %xmm10
	addpd	%xmm13, %xmm0
	movapd	20 * SIZE(BO), %xmm13
	addpd	%xmm10, %xmm1
	movapd	10 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm13
	mulpd	22 * SIZE(BO), %xmm10
	addpd	%xmm13, %xmm2
	movapd	48 * SIZE(BO), %xmm13
	addpd	%xmm10, %xmm3
	movapd	12 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm15
	mulpd	26 * SIZE(BO), %xmm10
	addpd	%xmm15, %xmm0
	movapd	28 * SIZE(BO), %xmm15
	addpd	%xmm10, %xmm1
	movapd	14 * SIZE(AO), %xmm10

	mulpd	%xmm10, %xmm15
	mulpd	30 * SIZE(BO), %xmm10
	addpd	%xmm15, %xmm2
	movapd	56 * SIZE(BO), %xmm15
	addpd	%xmm10, %xmm3
	movapd	24 * SIZE(AO), %xmm10

	addq   $16 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L62
	.align 8

.L65:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L69
	.align 8

.L66:
	mulpd	%xmm8, %xmm9
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm0
	movapd	 4 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm1
	movapd	 2 * SIZE(AO), %xmm8

	addq	$2 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L66
	.align 8

.L69:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8
	movsd	0 * SIZE(CO2), %xmm10
	movhpd	1 * SIZE(CO2), %xmm10
#endif

	addpd	%xmm2, %xmm0
	addpd	%xmm3, %xmm1

	mulpd	%xmm15, %xmm0
	mulpd	%xmm15, %xmm1

#ifndef TRMMKERNEL
	addpd	%xmm8,  %xmm0
	addpd	%xmm10, %xmm1
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	movsd	%xmm1, 0 * SIZE(CO2)
	movhpd	%xmm1, 1 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif

	addq	$2 * SIZE, CO1		# coffset += 4
	addq	$2 * SIZE, CO2		# coffset += 4
	.align 8	

.L70:
	testq	$1, M
	je	.L79
	.align 8

.L71:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 4), BO
#endif	

	movsd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movsd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movsd	 4 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movsd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

	movsd	16 * SIZE(BO), %xmm13
	movsd	24 * SIZE(BO), %xmm15

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$2, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L75
	.align 8

.L72:
	mulsd	%xmm8, %xmm9
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	mulsd	 2 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm0
	movsd	 4 * SIZE(BO), %xmm9
	addsd	%xmm8, %xmm1
	movsd	 1 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm9
	mulsd	 6 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm2
	movsd	32 * SIZE(BO), %xmm9
	addsd	%xmm8, %xmm3
	movsd	 2 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm11
	mulsd	10 * SIZE(BO), %xmm8
	addsd	%xmm11, %xmm0
	movsd	12 * SIZE(BO), %xmm11
	addsd	%xmm8, %xmm1
	movsd	 3 * SIZE(AO), %xmm8

	mulsd	%xmm8, %xmm11
	mulsd	14 * SIZE(BO), %xmm8
	addsd	%xmm11, %xmm2
	movsd	40 * SIZE(BO), %xmm11
	addsd	%xmm8, %xmm3
	movsd	 8 * SIZE(AO), %xmm8

	mulsd	%xmm10, %xmm13
	mulsd	18 * SIZE(BO), %xmm10
	addsd	%xmm13, %xmm0
	movsd	20 * SIZE(BO), %xmm13
	addsd	%xmm10, %xmm1
	movsd	 5 * SIZE(AO), %xmm10

	mulsd	%xmm10, %xmm13
	mulsd	22 * SIZE(BO), %xmm10
	addsd	%xmm13, %xmm2
	movsd	48 * SIZE(BO), %xmm13
	addsd	%xmm10, %xmm3
	movsd	 6 * SIZE(AO), %xmm10

	mulsd	%xmm10, %xmm15
	mulsd	26 * SIZE(BO), %xmm10
	addsd	%xmm15, %xmm0
	movsd	28 * SIZE(BO), %xmm15
	addsd	%xmm10, %xmm1
	movsd	 7 * SIZE(AO), %xmm10

	mulsd	%xmm10, %xmm15
	mulsd	30 * SIZE(BO), %xmm10
	addsd	%xmm15, %xmm2
	movsd	56 * SIZE(BO), %xmm15
	addsd	%xmm10, %xmm3
	movsd	12 * SIZE(AO), %xmm10

	addq   $ 8 * SIZE, AO
	addq   $32 * SIZE, BO
	decq   %rax
	jne    .L72
	.align 8

.L75:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movsd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L78
	.align 8

.L76:
	mulsd	%xmm8, %xmm9
	mulsd	 2 * SIZE(BO), %xmm8
	addsd	%xmm9, %xmm0
	addsd	%xmm8, %xmm1
	movsd	 1 * SIZE(AO), %xmm8
	movsd	 4 * SIZE(BO), %xmm9

	addq	$1 * SIZE, AO		# aoffset  += 4
	addq	$4 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L76
	.align 8

.L78:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movsd	0 * SIZE(CO2), %xmm10
#endif

	addsd	%xmm2, %xmm0
	addsd	%xmm3, %xmm1

	mulsd	%xmm15, %xmm0
	mulsd	%xmm15, %xmm1

#ifndef TRMMKERNEL
	addsd	%xmm8,  %xmm0
	addsd	%xmm10, %xmm1
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movsd	%xmm1, 0 * SIZE(CO2)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 4), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$1, KK
#endif
	.align 8	
	
.L79:
#if defined(TRMMKERNEL) && !defined(LEFT)
	addl	$2, KK
#endif
	leaq	(C, LDC, 2), C
	.align 8

.L80:
	testq	$1, N
	je	.L999
	.align 32

.L81:
/* Copying to Sub Buffer */
	leaq	BUFFER, BO
	
#if defined(TRMMKERNEL) && defined(LEFT)
	movq	OFFSET, %rax
	movq	%rax, KK
#endif	

	movq	K, %rax
	sarq	$3, %rax
	jle	.L83
	.align 8
	
.L82:
	prefetch	 56 * SIZE(B)

	movq	 0 * SIZE(B), %mm0
	movq	 1 * SIZE(B), %mm1
	movq	 2 * SIZE(B), %mm2
	movq	 3 * SIZE(B), %mm3
	movq	 4 * SIZE(B), %mm4
	movq	 5 * SIZE(B), %mm5
	movq	 6 * SIZE(B), %mm6
	movq	 7 * SIZE(B), %mm7

	addq	$ 8 * SIZE, B
	addq	$16 * SIZE, BO

	movq	%mm0, -16 * SIZE(BO)
	movq	%mm0, -15 * SIZE(BO)
	movq	%mm1, -14 * SIZE(BO)
	movq	%mm1, -13 * SIZE(BO)
	movq	%mm2, -12 * SIZE(BO)
	movq	%mm2, -11 * SIZE(BO)
	movq	%mm3, -10 * SIZE(BO)
	movq	%mm3,  -9 * SIZE(BO)
	movq	%mm4,  -8 * SIZE(BO)
	movq	%mm4,  -7 * SIZE(BO)
	movq	%mm5,  -6 * SIZE(BO)
	movq	%mm5,  -5 * SIZE(BO)
	movq	%mm6,  -4 * SIZE(BO)
	movq	%mm6,  -3 * SIZE(BO)
	movq	%mm7,  -2 * SIZE(BO)
	movq	%mm7,  -1 * SIZE(BO)

	decq	%rax
	jne	.L82
	.align 8

.L83:
	movq	K, %rax
	andq	$7, %rax
	BRANCH
	jle	.L90
	.align 8

.L84:
	movq	 0 * SIZE(B), %mm0

	movq	%mm0,  0 * SIZE(BO)
	movq	%mm0,  1 * SIZE(BO)

	addq	$1 * SIZE, B
	addq	$2 * SIZE, BO
	decq	%rax
	jne	.L84
	.align 8
	
.L90:
	movq	C, CO1			# coffset1 = c
	movq	A, AO		# aoffset = a

	movq	M,  I
	sarq	$2, I	# i = (m >> 2)
	jle	.L100
	.align 8

.L91:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movapd	 8 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movapd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

	movapd	16 * SIZE(AO), %xmm12
	movapd	24 * SIZE(AO), %xmm14

	prefetchw      4 * SIZE(CO1)
	prefetchw      4 * SIZE(CO2)

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$4, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L95
	.align 8

.L92:
	mulpd	%xmm9, %xmm8
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	mulpd	 2 * SIZE(AO), %xmm9
	addpd	%xmm8, %xmm0
	movapd	 4 * SIZE(AO), %xmm8
	addpd	%xmm9, %xmm1
	movapd	 2 * SIZE(BO), %xmm9
	mulpd	%xmm9, %xmm8
	mulpd	 6 * SIZE(AO), %xmm9
	addpd	%xmm8, %xmm2
	movapd	32 * SIZE(AO), %xmm8
	PREFETCH	(PREFETCHSIZE     +  8) * SIZE(AO)
	addpd	%xmm9, %xmm3
	movapd	 4 * SIZE(BO), %xmm9
	mulpd	%xmm9, %xmm10
	mulpd	10 * SIZE(AO), %xmm9
	addpd	%xmm10, %xmm0
	movapd	12 * SIZE(AO), %xmm10
	addpd	%xmm9, %xmm1
	movapd	 6 * SIZE(BO), %xmm9
	mulpd	%xmm9, %xmm10
	mulpd	14 * SIZE(AO), %xmm9
	addpd	%xmm10, %xmm2
	movapd	40 * SIZE(AO), %xmm10
	PREFETCH	(PREFETCHSIZE     + 16) * SIZE(AO)
	addpd	%xmm9, %xmm3
	movapd	16 * SIZE(BO), %xmm9
	mulpd	%xmm11, %xmm12
	mulpd	18 * SIZE(AO), %xmm11
	addpd	%xmm12, %xmm0
	movapd	20 * SIZE(AO), %xmm12
	addpd	%xmm11, %xmm1
	movapd	10 * SIZE(BO), %xmm11
	mulpd	%xmm11, %xmm12
	mulpd	22 * SIZE(AO), %xmm11
	addpd	%xmm12, %xmm2
	movapd	48 * SIZE(AO), %xmm12
	PREFETCH	(PREFETCHSIZE     + 24) * SIZE(AO)
	addpd	%xmm11, %xmm3
	movapd	12 * SIZE(BO), %xmm11
	mulpd	%xmm11, %xmm14
	mulpd	26 * SIZE(AO), %xmm11
	addpd	%xmm14, %xmm0
	movapd	28 * SIZE(AO), %xmm14
	addpd	%xmm11, %xmm1
	movapd	14 * SIZE(BO), %xmm11
	mulpd	%xmm11, %xmm14
	mulpd	30 * SIZE(AO), %xmm11
	addpd	%xmm14, %xmm2
	movapd	56 * SIZE(AO), %xmm14
	addpd	%xmm11, %xmm3
	movapd	24 * SIZE(BO), %xmm11

	addq   $32 * SIZE, AO
	addq   $16 * SIZE, BO
	decq   %rax
	jne    .L92
	.align 8

.L95:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L99
	.align 8

.L96:
	mulpd	%xmm9, %xmm8
	mulpd	 2 * SIZE(AO), %xmm9
	addpd	%xmm8, %xmm0
	movapd	 4 * SIZE(AO), %xmm8
	addpd	%xmm9, %xmm1
	movapd	 2 * SIZE(BO), %xmm9

	addq	$4 * SIZE, AO		# aoffset  += 4
	addq	$2 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L96
	.align 8

.L99:
#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8
	movsd	2 * SIZE(CO1), %xmm9
	movhpd	3 * SIZE(CO1), %xmm9
#endif

	addpd	%xmm2, %xmm0
	addpd	%xmm3, %xmm1

	mulpd	%xmm15, %xmm0
	mulpd	%xmm15, %xmm1

#ifndef TRMMKERNEL
	addpd	%xmm8,  %xmm0
	addpd	%xmm9,  %xmm1
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	movsd	%xmm1, 2 * SIZE(CO1)
	movhpd	%xmm1, 3 * SIZE(CO1)

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 4), AO
	leaq	(BO, %rax, 2), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$4, KK
#endif

	addq	$4 * SIZE, CO1		# coffset += 4
	decq	I			# i --
	jg	.L91
	.align 8	

.L100:
	testq	$2, M
	je	.L110
	.align 8

.L101:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movapd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movapd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movapd	 8 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movapd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$2, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L105
	.align 8

.L102:
	mulpd	%xmm8, %xmm9
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	movapd	 2 * SIZE(AO), %xmm8
	mulpd	 2 * SIZE(BO), %xmm8
	addpd	%xmm9, %xmm0
	movapd	16 * SIZE(BO), %xmm9
	addpd	%xmm8, %xmm1
	movapd	 4 * SIZE(AO), %xmm8
	mulpd	 4 * SIZE(BO), %xmm8
	addpd	%xmm8, %xmm2
	movapd	 6 * SIZE(AO), %xmm8
	mulpd	 6 * SIZE(BO), %xmm8
	addpd	%xmm8, %xmm3
	movapd	16 * SIZE(AO), %xmm8
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	mulpd	%xmm10, %xmm11
	movapd	10 * SIZE(AO), %xmm10
	mulpd	10 * SIZE(BO), %xmm10
	addpd	%xmm11, %xmm0
	movapd	24 * SIZE(BO), %xmm11
	addpd	%xmm10, %xmm1
	movapd	12 * SIZE(AO), %xmm10
	mulpd	12 * SIZE(BO), %xmm10
	addpd	%xmm10, %xmm2
	movapd	14 * SIZE(AO), %xmm10
	mulpd	14 * SIZE(BO), %xmm10
	addpd	%xmm10, %xmm3
	movapd	24 * SIZE(AO), %xmm10

	addq   $16 * SIZE, AO
	addq   $16 * SIZE, BO
	decq   %rax
	jne    .L102
	.align 8

.L105:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movapd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L109
	.align 8

.L106:
	mulpd	%xmm8, %xmm9
	addpd	%xmm9, %xmm0
	movapd	 2 * SIZE(AO), %xmm8
	movapd	 2 * SIZE(BO), %xmm9

	addq	$2 * SIZE, AO		# aoffset  += 4
	addq	$2 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L106
	.align 8

.L109:
	addpd	%xmm1, %xmm0
	addpd	%xmm3, %xmm2
	addpd	%xmm2, %xmm0

	mulpd	%xmm15, %xmm0

#ifndef TRMMKERNEL
	movsd	0 * SIZE(CO1), %xmm8
	movhpd	1 * SIZE(CO1), %xmm8

	addpd	%xmm8,  %xmm0
#endif

	movsd	%xmm0, 0 * SIZE(CO1)
	movhpd	%xmm0, 1 * SIZE(CO1)
	addq	$2 * SIZE, CO1		# coffset += 4

#if (defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
    (defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))
	movq	K, %rax
	subq	KKK, %rax
	leaq	(,%rax, SIZE), %rax
	leaq	(AO, %rax, 2), AO
	leaq	(BO, %rax, 2), BO
#endif

#if defined(TRMMKERNEL) && defined(LEFT)
	addq	$2, KK
#endif
	.align 8	

.L110:
	testq	$1, M
	je	.L999
	.align 8

.L111:
#if !defined(TRMMKERNEL) || \
	(defined(TRMMKERNEL) &&  defined(LEFT) &&  defined(TRANSA)) || \
	(defined(TRMMKERNEL) && !defined(LEFT) && !defined(TRANSA))

	leaq	BUFFER, BO
#else
	leaq	BUFFER, BO
	movq	KK, %rax
	leaq	(, %rax, SIZE), %rax
	leaq	(AO, %rax, 1), AO
	leaq	(BO, %rax, 2), BO
#endif	

	movsd	 0 * SIZE(AO), %xmm8
	pxor	%xmm0, %xmm0
	movsd	 0 * SIZE(BO), %xmm9
	pxor	%xmm1, %xmm1
	movsd	 4 * SIZE(AO), %xmm10
	pxor	%xmm2, %xmm2
	movsd	 8 * SIZE(BO), %xmm11
	pxor	%xmm3, %xmm3

#ifndef TRMMKERNEL
	movq	K, %rax
#elif (defined(LEFT) && !defined(TRANSA)) || (!defined(LEFT) && defined(TRANSA))
	movq	K, %rax
	subq	KK, %rax
	movq	%rax, KKK	
#else
	movq	KK, %rax
#ifdef LEFT
	addq	$1, %rax
#else
	addq	$1, %rax
#endif
	movq	%rax, KKK
#endif
	sarq	$3, %rax
	je	.L115
	.align 8

.L112:
	mulsd	%xmm8, %xmm9
	PREFETCH	(PREFETCHSIZE     +  0) * SIZE(AO)
	movsd	 1 * SIZE(AO), %xmm8
	addsd	%xmm9, %xmm0
	movsd	16 * SIZE(BO), %xmm9
	mulsd	 2 * SIZE(BO), %xmm8
	addsd	%xmm8, %xmm1
	movsd	 2 * SIZE(AO), %xmm8
	mulsd	 4 * SIZE(BO), %xmm8
	addsd	%xmm8, %xmm2
	movsd	 3 * SIZE(AO), %xmm8
	mulsd	 6 * SIZE(BO), %xmm8
	addsd	%xmm8, %xmm3
	movsd	 8 * SIZE(AO), %xmm8
	mulsd	%xmm10, %xmm11
	movsd	 5 * SIZE(AO), %xmm10
	addsd	%xmm11, %xmm0
	movsd	24 * SIZE(BO), %xmm11
	mulsd	10 * SIZE(BO), %xmm10
	addsd	%xmm10, %xmm1
	movsd	 6 * SIZE(AO), %xmm10
	mulsd	12 * SIZE(BO), %xmm10
	addsd	%xmm10, %xmm2
	movsd	 7 * SIZE(AO), %xmm10
	mulsd	14 * SIZE(BO), %xmm10
	addsd	%xmm10, %xmm3
	movsd	12 * SIZE(AO), %xmm10

	addq   $ 8 * SIZE, AO
	addq   $16 * SIZE, BO
	decq   %rax
	jne    .L112
	.align 8

.L115:
#ifndef TRMMKERNEL
	movq	K, %rax
#else
	movq	KKK, %rax
#endif
	movsd	ALPHA, %xmm15
	andq	$7, %rax		# if (k & 1)
	BRANCH
	je .L118
	.align 8

.L116:
	mulsd	%xmm8, %xmm9
	movsd	 1 * SIZE(AO), %xmm8
	addsd	%xmm9, %xmm0
	movsd	 2 * SIZE(BO), %xmm9

	addq	$1 * SIZE, AO		# aoffset  += 4
	addq	$2 * SIZE, BO		# boffset1 += 8
	decq	%rax
	jg	.L116
	.align 8

.L118:
	addsd	%xmm2, %xmm0
	addsd	%xmm3, %xmm1
	addsd	%xmm1, %xmm0

	mulsd	%xmm15, %xmm0
#ifndef TRMMKERNEL
	addsd	0 * SIZE(CO1), %xmm0
#endif
	movsd	%xmm0, 0 * SIZE(CO1)
	.align 8
	
.L999:
	movq	%rbx, %rsp

	EMMS

	movq	  0(%rsp), %rbx
	movq	  8(%rsp), %rbp
	movq	 16(%rsp), %r12
	movq	 24(%rsp), %r13
	movq	 32(%rsp), %r14
	movq	 40(%rsp), %r15
	addq	$STACKSIZE, %rsp
	ret

	EPILOGUE
