/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifndef CONJ
#define GEMM_KERNEL	GEMM_KERNEL_R
#define GEMM_KERNEL_B0	GEMM_KERNEL_R_B0
#else
#define GEMM_KERNEL	GEMM_KERNEL_L
#define GEMM_KERNEL_B0	GEMM_KERNEL_L_B0
#endif

void CNAME(long m, long n, long k, FLOAT alpha_r, FLOAT alpha_i,
	   FLOAT *a, FLOAT *b, FLOAT *c, long ldc, long offset, FLOAT *subbuffer, int flag){

  long i, j;
  long maxmn = MAX(GEMM_UNROLL_M, GEMM_UNROLL_N);
  long loop;

  if (offset > 0) {
#ifdef LOWER
    GEMM_KERNEL(m, offset, k,
		alpha_r, alpha_i,
		a, b, c, ldc); 
    
#endif
    b += offset * k   * COMPSIZE;
    c += offset * ldc * COMPSIZE;
    n -= offset;
    offset = 0;
  }

  if (n <= 0) return;

  if (n > m + offset) {
#ifndef LOWER
    GEMM_KERNEL(m, n - m - offset, k,
		alpha_r, alpha_i,
		a, 
		b + (m + offset) * k   * COMPSIZE,
		c + (m + offset) * ldc * COMPSIZE, ldc); 
#endif
    n = m + offset;
  }

  if (offset < 0) {
#ifndef LOWER
    GEMM_KERNEL(-offset, n, k,
		  alpha_r,
		  alpha_i,
		  a, b, c, ldc); 
#endif
    a -= offset * k   * COMPSIZE;
    c -= offset       * COMPSIZE;
    m += offset;
    offset = 0;
  }

  if (m <= 0) return;

  if (m > n - offset) {
#ifdef LOWER
    GEMM_KERNEL(m - n + offset, n, k,
		  alpha_r,
		  alpha_i,
		  a + (n - offset) * k * COMPSIZE,
		  b,
		  c + (n - offset)     * COMPSIZE, ldc); 
#endif
    m = n + offset;
  }


  for (loop = 0; loop < n; loop += maxmn) {
      
    int mm, nn;
    
    mm = (loop & ~(maxmn - 1));
    nn = MIN(maxmn, n - loop);
    
#ifndef LOWER
    GEMM_KERNEL(mm, nn, k,
		  alpha_r,
#ifdef COMPLEX
		  alpha_i,
#endif
		  a, b + loop * k * COMPSIZE, c + loop * ldc * COMPSIZE, ldc); 
#endif
    
    if (flag) {
#ifndef BETAZERO_IMPLEMENTED
      GEMM_BETA(nn, nn, 0, ZERO, 
#ifdef COMPLEX
		ZERO,
#endif
		NULL, 0, NULL, 0, subbuffer, nn);
      
      GEMM_KERNEL(nn, nn, k,
		    alpha_r,
#ifdef COMPLEX
		    alpha_i,
#endif
		    a + loop * k * COMPSIZE, b + loop * k * COMPSIZE, subbuffer, nn); 
#else
      GEMM_KERNEL_B0(nn, nn, k,
		    alpha_r,
#ifdef COMPLEX
		    alpha_i,
#endif
		    a + loop * k * COMPSIZE, b + loop * k * COMPSIZE, subbuffer, nn); 
#endif


#ifndef LOWER
      
      for (j = 0; j < nn; j ++) {
	for (i = 0; i <= j; i ++) {
	  c[(i + loop + (j + loop) * ldc) * 2 + 0] += 
	    subbuffer[(i + j * nn) * 2 + 0] + subbuffer[(j + i * nn) * 2 + 0];
	  if (i != j) {
	    c[(i + loop + (j + loop) * ldc) * 2 + 1] += 
	      subbuffer[(i + j * nn) * 2 + 1] - subbuffer[(j + i * nn) * 2 + 1];
	  } else {
	    c[(i + loop + (j + loop) * ldc) * 2 + 1] = ZERO;
	  }
	}
      }
#else
      for (j = 0; j < nn; j ++) {
	for (i = j; i < nn; i ++) {
	  c[(i + loop + (j + loop) * ldc) * 2 + 0] +=
	    subbuffer[(i + j * nn) * 2 + 0] + subbuffer[(j + i * nn) * 2 + 0];
	  if (i != j) {
	    c[(i + loop + (j + loop) * ldc) * 2 + 1] +=
	      subbuffer[(i + j * nn) * 2 + 1] - subbuffer[(j + i * nn) * 2 + 1];
	  } else {
	    c[(i + loop + (j + loop) * ldc) * 2 + 1]  = ZERO;
	  }
	}
      }
#endif
    }
    
#ifdef LOWER
    GEMM_KERNEL(m - mm - nn, nn, k,
		  alpha_r,
#ifdef COMPLEX
		  alpha_i,
#endif
		  a + (mm + nn) * k * COMPSIZE, b + loop * k * COMPSIZE, 
		  c + (mm + nn + loop * ldc) * COMPSIZE, ldc); 
#endif
  }

    return;
}
