/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef COMPLEX
#ifndef CONJ
#ifndef DOUBLE
#define SYRK_KERNEL_U cherk_kernel_UN
#define SYRK_KERNEL_L cherk_kernel_LN
#else
#define SYRK_KERNEL_U zherk_kernel_UN
#define SYRK_KERNEL_L zherk_kernel_LN
#endif
#define GEMM_KERNEL GEMM_KERNEL_R
#else
#ifndef DOUBLE
#define SYRK_KERNEL_U cherk_kernel_UC
#define SYRK_KERNEL_L cherk_kernel_LC
#else
#define SYRK_KERNEL_U zherk_kernel_UC
#define SYRK_KERNEL_L zherk_kernel_LC
#endif
#define GEMM_KERNEL GEMM_KERNEL_L
#endif
#endif

int CNAME(long dummy1, long n, long k, FLOAT alpha_r, FLOAT alpha_i,
	  FLOAT *a, long lda, 
	  FLOAT *dummy2, long dummy3,
	  FLOAT *c, long ldc, FLOAT *buffer, long from, long to){

  long ls, is, js;
  long min_l, min_i, min_j;

#ifndef PPC440
  FLOAT *sa = (FLOAT *)((void *)buffer + GEMM_BUFFER_A_OFFSET);
  FLOAT *sb = (FLOAT *)((void *)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_BUFFER_A_OFFSET * COMPSIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
  FLOAT *subbuffer = (FLOAT *)((void *)buffer + BUFFER_LIMIT);
#else
  extern FLOAT *sa, *sb;
  FLOAT *subbuffer = malloc(GEMM_P * GEMM_Q * COMPSIZE * SIZE);
#endif

  for(ls = 0; ls < k; ls += GEMM_Q){
    min_l = k - ls;
    if (min_l > GEMM_Q) min_l = GEMM_Q;
    
    for(js = from; js < to; js += GEMM_R){
      min_j = to - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
#ifndef TRANS
      GEMM_OTCOPY(min_l, min_j, a + (js + ls * lda) * COMPSIZE, lda, sb);
#else
      GEMM_ONCOPY(min_l, min_j, a + (ls + js * lda) * COMPSIZE, lda, sb);
#endif
      
      for(is = 0; is < n; is += GEMM_P){
	min_i = n - is;
	if (min_i > GEMM_P) min_i = GEMM_P;
	
#ifndef TRANS
	GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
	GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
	
#ifndef LOWER
	if (js > is + min_i) {
	  GEMM_KERNEL(min_i, min_j, min_l,
			alpha_r, ZERO,
			sa, sb, c + (is + js * ldc) * COMPSIZE, ldc); 
	} else
	  if ( is > js + min_j) {
	    /* do nothing */
	  } else {
	    SYRK_KERNEL_U(min_i, min_j, min_l, alpha_r,
			  sa, sb, 
			  c + (is + js * ldc) * COMPSIZE, ldc,
			  is - js, subbuffer);
	  }
#else

	if (js > is + min_i) {
	  /* do nothing */
	} else
	  if (is > js + min_j) {
	    GEMM_KERNEL(min_i, min_j, min_l,
			  alpha_r, ZERO,
			  sa, sb, c + (is + js * ldc) * COMPSIZE, ldc); 
	  } else {
	    SYRK_KERNEL_L(min_i, min_j, min_l, alpha_r, 
			  sa, sb, 
			  c + (is + js * ldc) * COMPSIZE, ldc,
			  is - js, subbuffer);
	  }
#endif

      } /* end of ls */
    } /* end of is */
  } /* end of js */
  
#ifdef PPC440
  free(subbuffer);
#endif

  return 0;
}
