/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

#ifdef SYMM_P1
#define ZSYMM_P1	((SYMM_P1) / 2)
#endif

#if ((ZSYMM_P1 * ZSYMM_P1 * SIZE * 2) > BUFFER_OFFSET_LIMIT)
#error "Parameter is too Large."
#endif

int CNAME(long dummy1, long n, long k, FLOAT alpha_r, FLOAT alpha_i,
	 FLOAT *a, long lda, FLOAT *b, long ldb, FLOAT *c, long ldc, FLOAT *buffer){

  long i, j, is, ilength;
  FLOAT *subbuffer = buffer + SUB_BUFFER_OFFSET;
  
  for (is = 0; is < n; is += ZSYMM_P1){

    ilength = MIN(ZSYMM_P1, n - is);
    
#ifndef LOWER
    if (is > 0){
#ifndef TRANS
      GEMM_NT(is, ilength, k, alpha_r, alpha_i,
	       a               , lda,
	       b + is * 2, ldb, 
	       c + is * ldc * 2, ldc, buffer);
      GEMM_NT(is, ilength, k, alpha_r, alpha_i,
	       b               , ldb,
	       a + is * 2, lda, 
	       c + is * ldc * 2, ldc, buffer);
#else
      GEMM_TN(is, ilength, k, alpha_r, alpha_i,
	       a               , lda,
	       b + is * ldb * 2, ldb, 
	       c + is * ldc * 2, ldc, buffer);
      GEMM_TN(is, ilength, k, alpha_r, alpha_i,
	       b               , ldb,
	       a + is * lda * 2, lda, 
	       c + is * ldc * 2, ldc, buffer);
#endif
    }
#endif
    
    GEMM_BETA(ilength, ilength, 0, ZERO, ZERO, NULL, 0, NULL, 0, subbuffer, ilength);
    
#ifndef TRANS
    GEMM_NT(ilength, ilength, k, alpha_r, alpha_i,
	     a + is * 2, lda,
	     b + is * 2, ldb, 
	     subbuffer, ilength, buffer);
#else
    GEMM_TN(ilength, ilength, k, alpha_r, alpha_i,
	     a + is * lda * 2, lda,
	     b + is * ldb * 2, ldb, 
	     subbuffer, ilength, buffer);
#endif
    
    for (j = 0; j < ilength; j++){
#ifndef LOWER
      for (i = 0; i < j; i++){
#else
      for (i = j + 1; i < ilength; i++){
#endif
	c[(i + is + (j + is) * ldc) * 2 + 0]
	  +=  subbuffer[(i + j * ilength) * 2 + 0] 
	    + subbuffer[(j + i * ilength) * 2 + 0];
	c[(i + is + (j + is) * ldc) * 2 + 1]
	  +=  subbuffer[(i + j * ilength) * 2 + 1] 
	    + subbuffer[(j + i * ilength) * 2 + 1];
	
      }
      c[(j + is + (j + is) * ldc) * 2 + 0] 
	+= 2 * subbuffer[(j + j * ilength) * 2 + 0];
      c[(j + is + (j + is) * ldc) * 2 + 1] 
	+= 2 * subbuffer[(j + j * ilength) * 2 + 1];
    }
      
      
#ifdef LOWER
      if (n - is - ZSYMM_P1> 0){
#ifndef TRANS
	GEMM_NT(n - is - ZSYMM_P1, ZSYMM_P1, k, alpha_r, alpha_i, 
		 a + (is + ZSYMM_P1)           * 2, lda,
		 b +  is                 * 2, ldb, 
		 c + (is + ZSYMM_P1 + is *ldc) * 2, ldc,
		 buffer);
	GEMM_NT(n - is - ZSYMM_P1, ZSYMM_P1, k, alpha_r, alpha_i, 
		 b + (is + ZSYMM_P1)           * 2, ldb,
		 a + is                  * 2, lda, 
		 c + (is + ZSYMM_P1 + is *ldc) * 2, ldc,
		 buffer);
#else
	GEMM_TN(n - is - ZSYMM_P1, ZSYMM_P1, k, alpha_r, alpha_i, 
		 a + (is + ZSYMM_P1) * lda * 2, lda,
		 b +  is       * ldb * 2, ldb, 
		 c + (is + ZSYMM_P1 + is *ldc) * 2, ldc, 
		 buffer);
	GEMM_TN(n - is - ZSYMM_P1, ZSYMM_P1, k, alpha_r, alpha_i, 
		 b + (is + ZSYMM_P1) * ldb * 2, ldb,
		 a +  is       * lda * 2, lda, 
		 c + (is + ZSYMM_P1 + is *ldc) * 2, ldc, 
		 buffer);
#endif
      }
#endif
    }
    return 0;
  }
