/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

static inline void inner_syrk(long n, long k, FLOAT alpha_r, FLOAT alpha_i,
		       FLOAT *a, long lda,
		       FLOAT *c, long ldc, FLOAT *buffer){

  long ks, klength;
  long i, is;
  
  for (ks = 0; ks < k; ks += SYRK_Q1){
    klength = MIN(SYRK_Q1, k - ks);
    for (is = 0; is < n; is += SYRK_P1){

#ifndef LOWER
      if (is > 0){
#ifndef TRANS
	GEMM_NT(is, MIN(n - is, SYRK_P1), klength, alpha_r, alpha_i,
		a       + ks * lda  * 2, lda,
		a + (is + ks * lda) * 2, lda, 
		c       + is * ldc  * 2, ldc, buffer);
#else
	GEMM_TN(is, MIN(n - is, SYRK_P1), klength, alpha_r, alpha_i,
		a +  ks             * 2, lda,
		a + (ks + is * lda) * 2, lda, 
		c       + is * ldc  * 2, ldc, buffer);
#endif
      }
#endif

      for (i = is; i < MIN(is + SYRK_P1, n); i++){
#if   !defined(LOWER) && !defined(TRANS)
	ZGEMV_N(i - is + 1,  klength, 0, alpha_r, alpha_i,
	       a + (is + ks * lda) * 2, lda,
	       a + (i +  ks * lda) * 2, lda,
	       c + (is + i  * ldc) * 2, 1, buffer);
#elif !defined(LOWER) &&  defined(TRANS)
	ZGEMV_T(klength, i - is + 1, 0, alpha_r, alpha_i,
	       a + (ks + is * lda) * 2, lda,
	       a + (ks + i  * lda) * 2, 1,
	       c + (is + i  * ldc) * 2, 1, buffer);
#elif  defined(LOWER) && !defined(TRANS)
	ZGEMV_N(MIN(is + SYRK_P1, n) - i, klength, 0, alpha_r, alpha_i,
	       a + (i + ks * lda) * 2, lda,
	       a + (i + ks * lda) * 2, lda,
	       c + (i + i  * ldc) * 2, 1, buffer);
#else
	ZGEMV_T(klength, MIN(is + SYRK_P1, n) - i, 0, alpha_r, alpha_i,
	       a + (ks + i * lda) * 2, lda,
	       a + (ks + i * lda) * 2, 1,
	       c + (i  + i * ldc) * 2, 1, buffer);
#endif
      }
      
#ifdef LOWER
      if (n - is - SYRK_P1> 0){
#ifndef TRANS
	GEMM_NT(n - is - SYRK_P1, SYRK_P1, klength, alpha_r, alpha_i, 
		     a + (is + SYRK_P1 + ks * lda) * 2, lda,
		     a + (is      + ks * lda) * 2, lda, 
		     c + (is + SYRK_P1 + is * ldc) * 2, ldc,
		     buffer);
#else
	GEMM_TN(n - is - SYRK_P1, SYRK_P1, klength, alpha_r, alpha_i, 
		     a + ((is + SYRK_P1) * lda + ks) * 2, lda,
		     a + (is * lda + ks) * 2,        lda, 
		     c + (is + SYRK_P1 + is *ldc) * 2,    ldc, 
		     buffer);
#endif
      }
#endif

    }
  }
}

int CNAME(long dummy1, long n, long k, FLOAT alpha_r, FLOAT alpha_i,
	 FLOAT *a, long lda, 
	 FLOAT *dummy2, long dummy3,
	 FLOAT *c, long ldc, FLOAT *buffer){

#ifdef SYRK_Q2

  long ks, klength;
  long is;
  
  for (ks = 0; ks < k; ks += SYRK_Q2){
    klength = MIN(SYRK_Q2, k - ks);
    for (is = 0; is < n; is += SYRK_P2){

#ifndef LOWER
      if (is > 0){
#ifndef TRANS
	GEMM_NT(is, MIN(n - is, SYRK_P2), klength, alpha_r, alpha_i,
		a       + ks * lda  * 2, lda,
		a + (is + ks * lda) * 2, lda, 
		c       + is * ldc  * 2, ldc, buffer);
#else
	GEMM_TN(is, MIN(n - is, SYRK_P2), klength, alpha_r, alpha_i,
		a +  ks             * 2, lda,
		a + (ks + is * lda) * 2, lda, 
		c       + is * ldc  * 2, ldc, buffer);
#endif
      }
#endif

#ifndef TRANS
      inner_syrk(MIN(SYRK_P2, n - is), MIN(SYRK_Q2, k - ks), alpha_r, alpha_i, 
		 a + (is + ks * lda) * 2, lda, 
		 c + (is + is * ldc) * 2, ldc, buffer);
#else
      inner_syrk(MIN(SYRK_P2, n - is), MIN(SYRK_Q2, k - ks), alpha_r, alpha_i, 
		 a + (is * lda + ks) * 2, lda,
		 c + (is + is * ldc) * 2, ldc, buffer);
#endif

#ifdef LOWER
      if (n - is - SYRK_P2> 0){
#ifndef TRANS
	GEMM_NT  (n - is - SYRK_P2, SYRK_P2, klength, alpha_r, alpha_i, 
		     a + (is + SYRK_P2 + ks * lda) * 2, lda,
		     a + (is + ks * lda) * 2,      lda, 
		     c + (is + SYRK_P2 + is *ldc) * 2,  ldc,
		     buffer);
#else
	GEMM_TN  (n - is - SYRK_P2, SYRK_P2, klength, alpha_r, alpha_i, 
		     a + ((is + SYRK_P2) * lda + ks) * 2, lda,
		     a + (is * lda + ks) * 2,        lda, 
		     c + (is + SYRK_P2 + is *ldc) * 2,    ldc, 
		     buffer);
#endif
      }
#endif

    }
  }

#else
  long is;
  
    for (is = 0; is < n; is += SYRK_P2){

#ifndef LOWER
      if (is > 0){
#ifndef TRANS
	GEMM_NT(is, MIN(n - is, SYRK_P2), k, alpha_r, alpha_i,
		a               , lda,
		a + is       * 2, lda, 
		c + is * ldc * 2, ldc, buffer);
#else
	GEMM_TN(is, MIN(n - is, SYRK_P2), k, alpha_r, alpha_i,
		a               , lda,
		a + is * lda * 2, lda, 
		c + is * ldc * 2, ldc, buffer);
#endif
      }
#endif

#ifndef TRANS
      inner_syrk(MIN(SYRK_P2, n - is), k, alpha_r, alpha_i, 
		 a + is * 2, lda, 
		 c + (is + is * ldc) * 2, ldc, buffer);
#else
      inner_syrk(MIN(SYRK_P2, n - is), k, alpha_r, alpha_i, 
		 a + is * lda * 2, lda,
		 c + (is + is * ldc) * 2, ldc, buffer);
#endif

#ifdef LOWER
      if (n - is - SYRK_P2> 0){
#ifndef TRANS
	GEMM_NT  (n - is - SYRK_P2, SYRK_P2, k, alpha_r, alpha_i, 
		     a + (is + SYRK_P2) * 2, lda,
		     a + (is     ) * 2, lda, 
		     c + (is + SYRK_P2 + is *ldc) * 2,  ldc,
		     buffer);
#else
	GEMM_TN  (n - is - SYRK_P2, SYRK_P2, k, alpha_r, alpha_i, 
		     a + ((is + SYRK_P2) * lda) * 2, lda,
		     a + (is * lda       ) * 2, lda, 
		     c + (is + SYRK_P2 + is *ldc) * 2,    ldc, 
		     buffer);
#endif
      }
#endif

    }
#endif

  return 0;
}
