/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

const static FLOAT dp1 = 1.;

#if 0
#undef GEMM_P
#undef GEMM_Q
#undef GEMM_R

#define GEMM_P  128
#define GEMM_Q 1024
#define GEMM_R 2000
#endif

#ifdef CONJ
#define GEMM_KERNEL   GEMM_KERNEL_L
#define TRMM_KERNEL_N TRMM_KERNEL_LR
#define TRMM_KERNEL_T TRMM_KERNEL_LC
#else
#define GEMM_KERNEL   GEMM_KERNEL_N
#define TRMM_KERNEL_N TRMM_KERNEL_LN
#define TRMM_KERNEL_T TRMM_KERNEL_LT
#endif


int CNAME(long m, long n, long dummy1,
	  FLOAT dummyalpha1, 
#ifdef COMPLEX
	  FLOAT dummyalpha2,
#endif
	  FLOAT *a, long lda,
	  FLOAT *dummy3, long dummy4, 
	  FLOAT *b, long ldb, FLOAT *buffer){

  long ls, is, js;
  long min_l, min_i, min_j;
  long offset;
#if (!defined(TRANSA) && !defined(UPPER)) || (defined(TRANSA) && defined(UPPER))
  int start_ls;
#endif

#ifndef PPC440
  FLOAT *sa, *sb;
  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)((long)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
#else
  extern FLOAT *sa, *sb;
#endif


#if (!defined(TRANSA) && defined(UPPER)) || (defined(TRANSA) && !defined(UPPER))
  for(ls = 0; ls < m; ls += GEMM_Q){
#else
  start_ls = 0;
  while (start_ls < m) start_ls += GEMM_Q;
  start_ls -= GEMM_Q;

  for(ls = start_ls; ls >= 0; ls -= GEMM_Q){
#endif

    min_l = m - ls;
    if (min_l > GEMM_Q) min_l = GEMM_Q;
    
    for(js = 0; js < n; js += GEMM_R){
      min_j = n - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
      GEMM_ONCOPY(min_l, min_j, b + (ls + js * ldb) * COMPSIZE, ldb, sb);
      
      for(is = 0; is < m; is += GEMM_P){
	min_i = m - is;
	if (min_i > GEMM_P) min_i = GEMM_P;

#if (defined(UPPER) && !defined(TRANSA)) || (!defined(UPPER) && defined(TRANSA))
	if (ls + GEMM_Q > is){
	  if (is + GEMM_P < ls) {

#ifndef TRANSA
	    GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
	    GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
	    GEMM_KERNEL(min_i, min_j, min_l, dp1, 
#ifdef COMPLEX
			  ZERO,
#endif
			  sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	  } else {
#ifndef TRANSA
	    TRMM_IUTCOPY(min_l, min_i, a, lda, ls, is, sa);
#else
	    TRMM_ILNCOPY(min_l, min_i, a, lda, ls, is, sa);
#endif

	    offset = is - ls;

	    if (min_l > offset) {
	      FLOAT *aa = sa;
	      FLOAT *bb = sb;
	      FLOAT *cc = b + (is + js * ldb) * COMPSIZE;
	      long mm = min_i;

	      if (offset < 0) {
		GEMM_KERNEL(-offset, min_j, min_l, dp1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb); 

		aa -= offset * min_l * COMPSIZE;
		mm += offset;
		cc -= offset * COMPSIZE;
		offset = 0;
	      }

#if 0
	      GEMM_BETA(MIN(mm, min_l - offset), min_j, 0, ZERO, 
#ifdef COMPLEX
			ZERO,
#endif
			NULL, 0, NULL, 0, cc, ldb);
#endif

	      TRMM_KERNEL_N(MIN(mm, min_l - offset), min_j, min_l, dp1,
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb, offset); 
	      
	    }
	  }
	}
#else
	if (is + GEMM_P > ls){
	  if (ls + GEMM_Q < is) {
#ifndef TRANSA
	    GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
	    GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
	    GEMM_KERNEL(min_i, min_j, min_l, dp1, 
#ifdef COMPLEX
			  ZERO,
#endif
			  sa, sb, b + (is + js * ldb) * COMPSIZE, ldb); 
	  } else {
#ifndef TRANSA
	    TRMM_ILTCOPY(min_l, min_i, a, lda, ls, is, sa);
#else
	    TRMM_IUNCOPY(min_l, min_i, a, lda, ls, is, sa);
#endif

	    offset = is - ls;

	    if (min_i > -offset) {
	      FLOAT *aa = sa;
	      FLOAT *bb = sb;
	      FLOAT *cc = b + (is + js * ldb) * COMPSIZE;
	      long mm = min_i;

	      if (offset < 0) {
		aa -= offset * min_l * COMPSIZE;
		cc -= offset * COMPSIZE;
		mm += offset;
		offset = 0;
	      }

	      if (mm > min_l - offset) {
		GEMM_KERNEL(mm - min_l + offset, min_j, min_l, dp1, 
#ifdef COMPLEX
			    ZERO,
#endif
			    aa + min_l * (min_l - offset) * COMPSIZE,
			    bb,
			    cc + (min_l - offset) * COMPSIZE, ldb); 
		
		mm = min_l - offset;
	      }
	
#if 0
	      GEMM_BETA(mm, min_j, 0, ZERO, 
#ifdef COMPLEX
			ZERO,
#endif
			NULL, 0, NULL, 0, cc, ldb);
#endif

	      TRMM_KERNEL_T(mm, min_j, min_l, dp1,
#ifdef COMPLEX
			    ZERO,
#endif
			    aa, bb, cc, ldb, offset); 

	    }

	  }
	}

#endif
	  
      } /* end of ls */
    } /* end of is */
  } /* end of js */

  return 0;
}
